
#include <iostream>
#include <vector>

#include <pcl/point_cloud.h>

namespace rgbd
{
using std::cout;
using std::endl;


template <typename PointT1, typename PointT2>
visualization_msgs::Marker cloud_correspondence_marker(
		pcl::PointCloud<PointT1> const& cloud0,
		pcl::PointCloud<PointT2> const& cloud1,
		const std::vector<std::pair<unsigned int, unsigned int> >& index_pairs,
		float line_width, float red, float green, float blue)
{
	visualization_msgs::Marker marker;
	marker.header = cloud0.header;

	marker.id = 0;
	marker.type = visualization_msgs::Marker::LINE_LIST;
	marker.color.r = red;
	marker.color.g = green;
	marker.color.b = blue;
	marker.color.a = 1.0;
	marker.scale.x = line_width;
	//marker.scale.y = 1; // ignored for line lists
	//marker.scale.z = 1; // ignored for line lists
	marker.points.clear();
	for (unsigned int i = 0; i < index_pairs.size(); i++) {
		unsigned int source_index = index_pairs[i].first;
		unsigned int target_index = index_pairs[i].second;
		// source
		geometry_msgs::Point source_point;
		source_point.x = cloud0.points[source_index].x;
		source_point.y = cloud0.points[source_index].y;
		source_point.z = cloud0.points[source_index].z;
		marker.points.push_back(source_point);
		// target
		geometry_msgs::Point target_point;
		target_point.x = cloud1.points[target_index].x;
		target_point.y = cloud1.points[target_index].y;
		target_point.z = cloud1.points[target_index].z;
		marker.points.push_back(target_point);
	}
	return marker;
}

template <typename PointT1, typename PointT2>
visualization_msgs::Marker cloud_correspondence_marker_with_colors(
		pcl::PointCloud<PointT1> const& cloud0,
		pcl::PointCloud<PointT2> const& cloud1,
		const std::vector<std::pair<unsigned int, unsigned int> >& index_pairs,
		const std::vector<rgbd::eigen::Vector3f> & color_vector,
		float line_width)
{
	assert(index_pairs.size() == color_vector.size());

	visualization_msgs::Marker marker;
	marker.header = cloud0.header;

	marker.id = 0;
	marker.type = visualization_msgs::Marker::LINE_LIST;
	marker.color.r = 1;
	marker.color.g = 1;
	marker.color.b = 1;
	marker.color.a = 1.0;
	marker.scale.x = line_width;
	//marker.scale.y = 1; // ignored for line lists
	//marker.scale.z = 1; // ignored for line lists
	marker.points.clear();
	marker.colors.clear();
	for (unsigned int i = 0; i < index_pairs.size(); i++) {
		unsigned int source_index = index_pairs[i].first;
		unsigned int target_index = index_pairs[i].second;
		// source
		geometry_msgs::Point source_point;
		source_point.x = cloud0.points[source_index].x;
		source_point.y = cloud0.points[source_index].y;
		source_point.z = cloud0.points[source_index].z;
		marker.points.push_back(source_point);
		// target
		geometry_msgs::Point target_point;
		target_point.x = cloud1.points[target_index].x;
		target_point.y = cloud1.points[target_index].y;
		target_point.z = cloud1.points[target_index].z;
		marker.points.push_back(target_point);
		// color
		std_msgs::ColorRGBA color;
		color.r = color_vector[i][0];
		color.g = color_vector[i][1];
		color.b = color_vector[i][2];
		marker.colors.push_back(color);
		marker.colors.push_back(color);
	}
	return marker;
}



template <typename PointT>
visualization_msgs::Marker cloud_normal_marker(
		pcl::PointCloud<PointT> const& cloud,
		const std::vector<unsigned int>& indices,
		float line_length, float line_width, float red, float green, float blue)
{
	visualization_msgs::Marker marker;
	marker.header = cloud.header;

	marker.id = 0;
	marker.type = visualization_msgs::Marker::LINE_LIST;
	marker.scale.x = line_width;
	//marker.scale.y = 1;
	//marker.scale.z = 1;
	marker.color.r = red;
	marker.color.g = green;
	marker.color.b = blue;
	marker.color.a = 1.0;

	marker.points.clear();
	unsigned int point_count = indices.size();
	if (point_count == 0) point_count = cloud.points.size();
	for (unsigned int i = 0; i < point_count; i++)
	{
		unsigned int index = indices.size() > 0 ? indices[i] : i;

		const PointT& cloud_point = cloud.points[index];

		if (isnan(cloud_point.x) || isnan(cloud_point.y) || isnan(cloud_point.z)) continue;
		if (isnan(cloud_point.normal[0]) || isnan(cloud_point.normal[1]) || isnan(cloud_point.normal[2])) continue;

		geometry_msgs::Point base_point;
		base_point.x = cloud_point.x;
		base_point.y = cloud_point.y;
		base_point.z = cloud_point.z;
		marker.points.push_back(base_point);

		geometry_msgs::Point target_point;
		target_point.x = cloud_point.x + cloud_point.normal[0] * line_length;
		target_point.y = cloud_point.y + cloud_point.normal[1] * line_length;
		target_point.z = cloud_point.z + cloud_point.normal[2] * line_length;
		marker.points.push_back(target_point);
	}
	return marker;
}

template <typename PointT>
visualization_msgs::Marker cloud_marker(
		pcl::PointCloud<PointT> const& cloud,
		const std::vector<unsigned int>& indices,
		float size, float red, float green, float blue)
{
	visualization_msgs::Marker marker;
	marker.header = cloud.header;

	marker.id = 0;
	marker.type = visualization_msgs::Marker::POINTS;
	marker.scale.x = marker.scale.y = marker.scale.z = size;
	marker.color.r = red;
	marker.color.g = green;
	marker.color.b = blue;
	marker.color.a = 1.0;

	marker.points.clear();
	unsigned int point_count = indices.size();
	if (point_count == 0) point_count = cloud.points.size();
	for (unsigned int i = 0; i < point_count; i++)
	{
		unsigned int index = indices.size() > 0 ? indices[i] : i;

		const PointT& cloud_point = cloud.points[index];

		geometry_msgs::Point base_point;
		base_point.x = cloud_point.x;
		base_point.y = cloud_point.y;
		base_point.z = cloud_point.z;
		marker.points.push_back(base_point);
	}
	return marker;
}

} //namespace
