/*
 * cloudNormals: computing and retrieving normals
 *
 * Evan Herbst
 * 3 / 24 / 10
 */

#include <cassert>
#include <cmath> //max(), fabs()
#include <iostream>
#include <algorithm>
#include <iterator>
#include <functional>
#include <utility>
#include <boost/unordered_map.hpp>
#include <boost/optional.hpp>
#include <boost/make_shared.hpp>
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>
#include <pcl/features/normal_3d.h>
#include <pcl/features/impl/normal_3d.hpp>
#include <pcl/kdtree/kdtree_flann.h>
#include <ros/package.h>
#include "rgbd_util/timer.h"
#include "rgbd_util/mathUtils.h" //sqr()
#include "rgbd_util/parallelism.h" //partitionEvenly()
#include "rgbd_util/primesensorUtils.h"
#include "rgbd_util/eigen/QR" //SelfAdjointEigenSolver
#include "primesensor_tools/primeSense_quantization_v5.h"
#include "pcl_rgbd/pointTypes.h"
#include "pcl_rgbd/cloudUtils.h"
#include "pcl_rgbd/cloudSearchTrees.h"
#include "pcl_rgbd/cloudSampling.h"

namespace rgbd
{
using std::vector;
using std::cout;
using std::endl;
using std::pair;
using boost::unordered_map;
using boost::optional;
using rgbd::eigen::Vector3d;
using rgbd::eigen::Vector4d;

#if ENABLE_OPENCL
int exec_kernel( float input_points_data[640][640], const geometry_msgs::Point32 &viewpoint, cl_float4 *output_data, char *fname, char *kname, int k , double max_z);
char * load_program_source(const char *filename);
int setup_GPU( char *fname, char *kname, cl_device_id &devices, cl_program &program, cl_context &context, cl_command_queue &cmd_queue, cl_kernel &kernel, cl_mem &output_mem, cl_float4 *output_data, size_t output_data_buffer_size, int k, double max_z  );
void shutdown_GPU( cl_device_id devices, cl_program program, cl_context context, cl_command_queue cmd_queue, cl_kernel kernel, cl_mem output_mem );
#endif

namespace normalsAux
{

template <typename PointNT>
void computePointPrincipalCurvatures(const pcl::PointCloud<PointNT> &normals, int p_idx, const std::vector<int> &indices,
      float &pcx, float &pcy, float &pcz, float &pc1, float &pc2)
{
  EIGEN_ALIGN16 rgbd::eigen::Matrix3f I = rgbd::eigen::Matrix3f::Identity ();
  rgbd::eigen::Vector3f n_idx (normals.points[p_idx].normal[0], normals.points[p_idx].normal[1], normals.points[p_idx].normal[2]);
  EIGEN_ALIGN16 rgbd::eigen::Matrix3f M = I - n_idx * n_idx.transpose ();    // projection matrix (into tangent plane)

  // Project normals into the tangent plane
  rgbd::eigen::Vector3f normal;
  std::vector<rgbd::eigen::Vector3f> projected_normals_(indices.size ());
  rgbd::eigen::Vector3f xyz_centroid_;
  xyz_centroid_.setZero ();
  for (size_t idx = 0; idx < indices.size(); ++idx)
  {
    normal[0] = normals.points[indices[idx]].normal[0];
    normal[1] = normals.points[indices[idx]].normal[1];
    normal[2] = normals.points[indices[idx]].normal[2];

    projected_normals_[idx] = M * normal;
    xyz_centroid_ += projected_normals_[idx];
  }

  // Estimate the XYZ centroid
  xyz_centroid_ /= indices.size ();

  // Initialize to 0
  EIGEN_ALIGN16 rgbd::eigen::Matrix3f covariance_matrix_;
  covariance_matrix_.setZero ();

  double demean_xy, demean_xz, demean_yz;
  // For each point in the cloud
  for (size_t idx = 0; idx < indices.size (); ++idx)
  {
	  rgbd::eigen::Vector3f demean_ = projected_normals_[idx] - xyz_centroid_;

    demean_xy = demean_[0] * demean_[1];
    demean_xz = demean_[0] * demean_[2];
    demean_yz = demean_[1] * demean_[2];

    covariance_matrix_(0, 0) += demean_[0] * demean_[0];
    covariance_matrix_(0, 1) += demean_xy;
    covariance_matrix_(0, 2) += demean_xz;

    covariance_matrix_(1, 0) += demean_xy;
    covariance_matrix_(1, 1) += demean_[1] * demean_[1];
    covariance_matrix_(1, 2) += demean_yz;

    covariance_matrix_(2, 0) += demean_xz;
    covariance_matrix_(2, 1) += demean_yz;
    covariance_matrix_(2, 2) += demean_[2] * demean_[2];
  }

  // Extract the eigenvalues and eigenvectors
  //rgbd::eigen::SelfAdjointEigenSolver<rgbd::eigen::Matrix3f> ei_symm (covariance_matrix_);
  //eigenvalues_  = ei_symm.eigenvalues ();
  //eigenvectors_ = ei_symm.eigenvectors ();
  EIGEN_ALIGN16 rgbd::eigen::Matrix3f eigenvectors_;
  rgbd::eigen::Vector3f eigenvalues_;
  pcl::eigen33 (covariance_matrix_, eigenvectors_, eigenvalues_);

  pcx = eigenvectors_ (0, 2);
  pcy = eigenvectors_ (1, 2);
  pcz = eigenvectors_ (2, 2);
  pc1 = eigenvalues_ (2);
  pc2 = eigenvalues_ (1);
}

} //namespace

template <typename PointT>
void computePrincipalCurvaturesThreadMain(const pcl::PointCloud<PointT>& cloud, const boost::shared_ptr<kdtree2>& kdtree, const float sqrRadius, std::vector<float>& pc1, std::vector<float>& pc2, const uint32_t firstIndex, const uint32_t lastIndex)
{
	for(uint32_t i = firstIndex; i <= lastIndex; i++)
	{
		vector<float> qpt(3);
		qpt[0] = cloud.points[i].x;
		qpt[1] = cloud.points[i].y;
		qpt[2] = cloud.points[i].z;
		kdtree2_result_vector result;
		kdtree->r_nearest(qpt, sqrRadius, result);

		vector<int> indices(result.size());
		for(size_t j = 0; j < result.size(); j++) indices[j] = result[j].idx;
		float pcx, pcy, pcz, _pc1, _pc2;
		normalsAux::computePointPrincipalCurvatures<PointT>(cloud, i, indices, pcx, pcy, pcz, _pc1, _pc2);
		pc1[i] = _pc1;
		pc2[i] = _pc2;
	}
}

/*
 * first and second principal curvatures, computed in parallel (unlike pcl as of 20110131)
 *
 * pc1, pc2 will be allocated
 *
 * pre: cloud has normals
 */
template <typename PointT>
void computePrincipalCurvatures(const pcl::PointCloud<PointT>& cloud, const float nbrhoodRadius, std::vector<float>& pc1, std::vector<float>& pc2, const bool parallelize)
{
	pc1.resize(cloud.points.size());
	pc2.resize(cloud.points.size());
	const boost::shared_ptr<kdtree2> kdtree = createKDTree2(cloud);
	if(parallelize)
	{
		const uint32_t numThreads = getSuggestedThreadCount();
		std::vector<boost::shared_ptr<rgbd::thread> > threads(numThreads);
		const vector<unsigned int> indices = partitionEvenly(cloud.points.size(), numThreads);
		for(uint32_t i = 0; i < numThreads; i++)
		{
			const std::function<void ()> f = rgbd::bind(computePrincipalCurvaturesThreadMain<PointT>, rgbd::cref(cloud), rgbd::cref(kdtree), sqr(nbrhoodRadius), rgbd::ref(pc1), rgbd::ref(pc2), indices[i], indices[i + 1] - 1);
			threads[i] = boost::shared_ptr<rgbd::thread>(new rgbd::thread(f));
		}
		for(uint32_t i = 0; i < numThreads; i++) threads[i]->join();
	}
	else
	{
		computePrincipalCurvaturesThreadMain<PointT>(cloud, kdtree, sqr(nbrhoodRadius), pc1, pc2, 0, cloud.points.size() - 1);
	}
}

template <typename PointT>
bool hasValidNormal(const PointT& pt)
{
	return pt.normal[0] != 0 || pt.normal[1] != 0 || pt.normal[2] != 0;
}
template <typename PointT>
void markNormalInvalid(PointT& pt)
{
	pt.normal[0] = pt.normal[1] = pt.normal[2] = 0;
}

/*
 * auxiliary to computePointCloudNormals()
 *
 * fill in the last four channels of points
 */
template <typename PointT>
void computePointCloudNormalsThreadMain(pcl::PointCloud<PointT> &points, unsigned int k, const geometry_msgs::Point32& viewpoint, const boost::shared_ptr<kdtree2>& kdtree, const unsigned int firstPt, const unsigned int lastPt)
{
	/*
	 * preallocate
	 */
	vector<float> qpt(3);
	vector<int> nn_indices(k);

	pcl::NormalEstimation<PointT, PointT> ester;
	for(unsigned int i = firstPt; i <= lastPt; i++)
	{
		qpt[0] = points.points[i].x;
		qpt[1] = points.points[i].y;
		qpt[2] = points.points[i].z;
		kdtree2_result_vector result;
		kdtree->n_nearest(qpt, k, result);
		for(unsigned int j = 0; j < k; j++) nn_indices[j] = result[j].idx;

		rgbd::eigen::Vector4f plane_parameters;                     // Compute the point normals (nx, ny, nz), surface curvature estimates (c)
		float curvature;
		ester.computePointNormal (points, nn_indices, plane_parameters, curvature);
		// cturtle likes this:

		pcl::flipNormalTowardsViewpoint<PointT> (points.points[i], viewpoint.x, viewpoint.y, viewpoint.z, plane_parameters);

		for(unsigned int j = 0; j < 3; j++) points.points[i].normal[j] = plane_parameters[j];
		points.points[i].curvature = curvature;
	}
}

template <typename PointT>
void computePointCloudNormals (pcl::PointCloud<PointT> &points, unsigned int k, const geometry_msgs::Point32& viewpoint)
{
	rgbd::timer t;
	boost::shared_ptr<kdtree2> kdtree = rgbd::createKDTree2(points);
	t.stop("create k-d tree");

	const unsigned int numThreads = getSuggestedThreadCount();
	vector<boost::shared_ptr<rgbd::thread> > threads(numThreads);
	const vector<unsigned int>& indices = partitionEvenly(points.points.size(), numThreads);
	t.restart();
	for(unsigned int i = 0; i < numThreads; i++)
	{
		const std::function<void ()> f = rgbd::bind(computePointCloudNormalsThreadMain<PointT>, rgbd::ref(points), k, rgbd::cref(viewpoint), rgbd::cref(kdtree), indices[i], indices[i + 1] - 1);
		threads[i] = boost::shared_ptr<rgbd::thread>(new rgbd::thread(f));
	}
	for(unsigned int i = 0; i < numThreads; i++) threads[i]->join();
	t.stop("compute normals");
}

/*
 * EVH, for debugging
 */
void checkFloat(const float f);

/*
 * pre: points has width, height set and points allocated
 */
template <typename PointT>
void computeOrganizedPointCloudNormalsThreadMain(pcl::PointCloud<PointT> &points, const pcl::PointCloud<PointT> &surface,
		boost::multi_array<bool, 2>& surfaceValidity, int k, unsigned int downsample_factor, double max_z,
		const geometry_msgs::Point32 & viewpoint, const unsigned int firstTgtRow, const unsigned int lastTgtRow)
{
    pcl::NormalEstimation<PointT, PointT> ester;
    for(unsigned int i = firstTgtRow; i < lastTgtRow; i++) //indices into target cloud
   	 for(unsigned int j = 0; j < points.width; j++)
		 {
   		 const unsigned int index = i * points.width + j;

   		 //indices into source cloud
   		 const unsigned int si = downsample_factor * i, sj = downsample_factor * j;
   		 const unsigned int sindex = si * surface.width + sj;

			//ignore points specifically marked as invalid
			if(!surfaceValidity[sj][si]) markNormalInvalid(points.points[index]);
			else
			{
				// Get all point neighbors in a local window
				int l = 0;
				std::vector<int> nn_indices((k + k + 1) * (k + k + 1));
				for(int y = std::max(0, (int)si - k); y <= std::min((int)surface.height - 1, (int)si + k); y++) //indices into surface
					for(int x = std::max(0, (int)sj - k); x <= std::min((int)surface.width - 1, (int)sj + k); x++)
					{
						unsigned int sindex2 = y * surface.width + x;
						// If the difference in Z (depth) between the query point and the current neighbor is smaller than max_z
						if(max_z != -1)
						{
							if(fabs(surface.points[sindex].z - surface.points[sindex2].z) < max_z)
								nn_indices[l++] = sindex2;
						}
						else nn_indices[l++] = sindex2;
					}

				if(l < 4){
					markNormalInvalid(points.points[index]); //too few neighbors for a reliable estimate
					surfaceValidity[sj][si] = false;
				}
				else
				{
					nn_indices.resize (l);
					// Compute the point normals (nx, ny, nz), surface curvature estimates (c)
					rgbd::eigen::Vector4f plane_parameters;
					float curvature;
					ester.computePointNormal (surface, nn_indices, plane_parameters, curvature);

					/*
					 * sometimes pcl gives nans; haven't looked into why -- EVH 20101203
					 */
					bool invalid = false;
					for(unsigned int m = 0; m < 4; m++)
						if(isinf(plane_parameters[m]) || isnan(plane_parameters[m]))
						{
							invalid = true;
							break;
						}
					if(!invalid && (isinf(curvature) || isnan(curvature))) invalid = true;

					if(!invalid)
					{
						pcl::flipNormalTowardsViewpoint (surface.points[sindex], viewpoint.x, viewpoint.y, viewpoint.z, plane_parameters);
						points.points[index] = surface.points[sindex];
						points.points[index].normal[0] = plane_parameters[0];
						points.points[index].normal[1] = plane_parameters[1];
						points.points[index].normal[2] = plane_parameters[2];
						points.points[index].curvature = curvature;
					}
					else
					{
						markNormalInvalid(points.points[index]);
						surfaceValidity[sj][si] = false;
					}
				}
			}
		 }
}

/*
 * compute normals and associated covariances of n_x, n_y, offset params in Bayesian linear regression
 *
 * pre: points has width, height set and points allocated
 */
template <typename PointT>
void computeOrganizedPointCloudNormalsWithStdevsThreadMain(pcl::PointCloud<PointT> &points, vector<rgbd::eigen::Matrix3f>& stdevs, const pcl::PointCloud<PointT> &surface,
	boost::multi_array<bool, 2>& surfaceValidity, const normals2dParams& params, const unsigned int firstTgtRow, const unsigned int lastTgtRow)
{
	const int k = params.windowHalfwidth;
	geometry_msgs::Point32 viewpoint;
	viewpoint.x = viewpoint.y = viewpoint.z = 0;

    pcl::NormalEstimation<PointT, PointT> ester;
    for(unsigned int i = firstTgtRow; i < lastTgtRow; i++) //indices into target cloud
   	 for(unsigned int j = 0; j < points.width; j++)
		 {
   		 const unsigned int index = i * points.width + j;

   		 //indices into source cloud
   		 const unsigned int si = params.downsampleFactor * i, sj = params.downsampleFactor * j;
   		 const unsigned int sindex = si * surface.width + sj;

			//ignore points specifically marked as invalid
			if(!surfaceValidity[sj][si]) markNormalInvalid(points.points[index]);
			else
			{
				// Get all point neighbors in a local window
				int l = 0;
				int minGoodX = INT_MAX, maxGoodX = INT_MIN, minGoodY = INT_MAX, maxGoodY = INT_MIN; //bounds on img coords
				std::vector<int> nn_indices((k + k + 1) * (k + k + 1));
				for(int y = std::max(0, (int)si - k); y <= std::min((int)surface.height - 1, (int)si + k); y++) //indices into surface
					for(int x = std::max(0, (int)sj - k); x <= std::min((int)surface.width - 1, (int)sj + k); x++)
					{
#define ADD_POINT() \
	{\
		nn_indices[l++] = sindex2;\
		if(x < minGoodX) minGoodX = x;\
		if(x > maxGoodX) maxGoodX = x;\
		if(y < minGoodY) minGoodY = y;\
		if(y > maxGoodY) maxGoodY = y;\
	}
						const unsigned int sindex2 = y * surface.width + x;
						// If the difference in Z (depth) between the query point and the current neighbor is smaller than max_z
						if(params.maxZDiff != -1)
						{
							if(fabs(surface.points[sindex].z - surface.points[sindex2].z) < params.maxZDiff * primesensor::stereoErrorRatio(surface.points[sindex].z))
								ADD_POINT();
						}
						else ADD_POINT();
#undef ADD_POINT
					}

				if(l < 6 //too few neighbors for a reliable estimate
					|| maxGoodX == minGoodX || maxGoodY == minGoodY) //all selected points in same img row or col; unreliable (in particular screws up Bayes lin reg)
				{
					markNormalInvalid(points.points[index]);
					surfaceValidity[sj][si] = false;
				}
				else
				{
					nn_indices.resize (l);

					rgbd::eigen::Vector3f nbrhoodMean(0, 0, 0);
					for(unsigned int k = 0; k < l; k++) nbrhoodMean += rgbd::ptX2eigen<rgbd::eigen::Vector3f>(surface.points[nn_indices[k]]);
					nbrhoodMean /= l;

//					for(unsigned int k = 0; k < l; k++) cout << rgbd::ptX2eigen<rgbd::eigen::Vector3f>(surface.points[nn_indices[k]]).transpose() << endl;

					/*
					 * also tried combining a vMF prior and Laplace likelihood to get a vMF posterior; doesn't seem to work -- EVH 20101123
					 */
#if 0 //Bayesian linear regression with z as the target (uncertainty info seems good but MAP estimates are way off -- ?)
					/*
					 * Bayesian linear regression of the form z = n_x x + n_y y + offset; get uncertainty of the params n_x, n_y, offset
					 */
					rgbd::eigen::MatrixXf Phi(l, 3);
					rgbd::eigen::VectorXf t(l);
					for(unsigned int k = 0; k < l; k++)
					{
						Phi(k, 0) = surface.points[nn_indices[k]].x;
						Phi(k, 1) = surface.points[nn_indices[k]].y;
						Phi(k, 2) = 1;
						t(k) = surface.points[nn_indices[k]].z;
					}
					const float beta = 1 / primesensor::stereoError(nbrhoodMean.z());
					assert(beta > 0);
#ifdef DEBUGGING
					rgbd::eigen::MatrixXf C = Phi.transpose() * Phi;
					/*
					 * C is of the form A^T A, so should be sym pos semidef
					 */
					if(C.determinant() <= 0)
					{
						cout << "Phi = " << endl << Phi << endl;
						cout << "C = " << endl << C << endl;
						cout << "Cdet " << C.determinant() << endl;
						getchar();
					}
#endif
					const rgbd::eigen::MatrixXf Sigma = (beta * Phi.transpose() * Phi).inverse();
#ifdef DEBUGGING
					if(Sigma.determinant() < 1e-16)
					{
						cout << "C = " << endl << C << endl;
						cout << "Sdet " << Sigma.determinant() << endl;
					}
#endif
					const rgbd::eigen::VectorXf mu = Sigma * beta * (Phi.transpose() * t);
					cout << "mu " << mu.transpose() << endl;
					getchar();

					rgbd::eigen::Vector3f normal(mu[0], mu[1], -1);
					normal.normalize();
					points.points[index] = surface.points[sindex];
					points.points[index].normal[0] = normal[0];
					points.points[index].normal[1] = normal[1];
					points.points[index].normal[2] = normal[2];
					points.points[index].curvature = -1; //flag for not computed

					stdevs[index] = Sigma;
#elif 1 //combine Bingham prior and Bingham likelihood to get a Bingham posterior
					/*
					 * combine a Bingham prior and a Bingham likelihood to get a Bingham posterior
					 *
					 * info on Bingham dist. from
					 * C. Bingham, "An Antipodally Symmetric Distribution on the Sphere", Annals of Statistics, 1974
					 */
					rgbd::eigen::Matrix3f A0 = 1e-4 * rgbd::eigen::Matrix3f::Identity(); //precision
					rgbd::eigen::Matrix3f A = -A0;
					for(unsigned int k = 0; k < l; k++)
					{
						rgbd::eigen::Vector3f x1 = (rgbd::ptX2eigen<rgbd::eigen::Vector3f>(surface.points[nn_indices[k]]) - nbrhoodMean).normalized();
//						cout << "x1 " << x1 << endl;
						rgbd::eigen::Vector3f x2 = (x1 + rgbd::eigen::Vector3f(0, .1, 0)).normalized();
						x2 = (x2 - x2.dot(x1) * x1).normalized();
						if(x2.dot(x1) > .98)
						{
							x2 = (x1 + rgbd::eigen::Vector3f(.1, 0, 0)).normalized();
							x2 = (x2 - x2.dot(x1) * x1).normalized();
						}
						rgbd::eigen::Vector3f x3 = x1.cross(x2);
						rgbd::eigen::Matrix3f A_; //covariance
						A_.row(0) = x1;
						A_.row(1) = x2;
						A_.row(2) = x3;
//						cout << "axes" << endl << A_ << endl;
						A_ = A_ * rgbd::eigen::DiagonalMatrix<float, 3, 3>(rgbd::eigen::Vector3f(1e-1, 1e0, 1e0)) * A_.transpose();
//						cout << "A_" << endl << A_ << endl;
						A_ = A_.inverse().eval();
//						cout << "A_-1" << endl << A_ << endl;
//						cout << endl;
						A -= A_;
					}
//					cout << "A" << endl << A << endl;
					rgbd::eigen::Matrix3f S = -A.inverse(); //posterior covariance
//					cout << "post S" << endl << S << endl;

					rgbd::eigen::SelfAdjointEigenSolver<rgbd::eigen::Matrix3f> eig(S);
					const rgbd::eigen::Matrix3f evecs = eig.eigenvectors();
					rgbd::eigen::Vector3f normal = evecs.col(2).normalized();
					if(normal.z() > 0) normal *= -1;
//					cout << "normal " << normal.transpose() << endl;
					points.points[index] = surface.points[sindex];
					points.points[index].normal[0] = normal[0];
					points.points[index].normal[1] = normal[1];
					points.points[index].normal[2] = normal[2];
					points.points[index].curvature = -1; //flag for not computed

					stdevs[index] = S;

//					getchar();
#endif
				}
			}
		 }
}

template <typename PointT>
void computeOrganizedPointCloudNormals (
    		pcl::PointCloud<PointT> &points, const pcl::PointCloud<PointT> &surface,
    		boost::multi_array<bool, 2>& surfaceValidity,
    		unsigned int k, unsigned int downsample_factor, double max_z,
    		const geometry_msgs::Point32 &viewpoint, const bool multithread)
{
	// Reduce by a factor of N
	points.width = lrint (ceil (surface.width / (double)downsample_factor));
	points.height = lrint (ceil (surface.height / (double)downsample_factor));
	points.points.resize(points.width * points.height);

	if(multithread)
	{
		const unsigned int numThreads = getSuggestedThreadCount();
		vector<boost::shared_ptr<rgbd::thread> > threads(numThreads);
		const vector<unsigned int> rowIndices = partitionEvenly(points.height, numThreads);
		for(unsigned int i = 0; i < numThreads; i++)
		{
			const std::function<void ()> f = rgbd::bind(computeOrganizedPointCloudNormalsThreadMain<PointT>, rgbd::ref(points), rgbd::cref(surface), rgbd::ref(surfaceValidity), k, downsample_factor, max_z, rgbd::cref(viewpoint), rowIndices[i], rowIndices[i + 1]);
			threads[i] = boost::shared_ptr<rgbd::thread>(new rgbd::thread(f));
		}
		for(unsigned int i = 0; i < numThreads; i++) threads[i]->join();
	}
	else
	{
		computeOrganizedPointCloudNormalsThreadMain(points, surface, surfaceValidity, k, downsample_factor, max_z, viewpoint, 0, points.height);
	}
}

template <typename PointT>
void computeOrganizedPointCloudNormalsWithStdevs(
    pcl::PointCloud<PointT> &points, std::vector<rgbd::eigen::Matrix3f>& stdevs, const pcl::PointCloud<PointT> &surface, boost::multi_array<bool, 2>& surfaceValidity, const normals2dParams& params)
{
	// Reduce by a factor of N
	points.width = lrint (ceil (surface.width / (double)params.downsampleFactor));
	points.height = lrint (ceil (surface.height / (double)params.downsampleFactor));
	points.points.resize(points.width * points.height);

	const unsigned int numThreads = getSuggestedThreadCount();
	vector<boost::shared_ptr<rgbd::thread> > threads(numThreads);
	const vector<unsigned int> rowIndices = partitionEvenly(points.height, numThreads);
	for(unsigned int i = 0; i < numThreads; i++)
		threads[i] = boost::shared_ptr<rgbd::thread>(new rgbd::thread(rgbd::bind(computeOrganizedPointCloudNormalsWithStdevsThreadMain<PointT>, rgbd::ref(points), rgbd::ref(stdevs), rgbd::cref(surface), rgbd::ref(surfaceValidity), rgbd::cref(params), rowIndices[i], rowIndices[i + 1])));
	for(unsigned int i = 0; i < numThreads; i++) threads[i]->join();
}

/**
 * Upsamples normals for an organized cloud that only has normals where
 * x%normals_downsample = 0 and y%normals_downsample = 0
 *
 * @param organizedCloud should have height, width set
 * @param targetValidityGrid should have been allocated
 * @param normals_downsample
 */
template <typename PointT>
void interpolateNormals(pcl::PointCloud<PointT> &organizedCloud, boost::multi_array<bool, 2>& targetValidityGrid, unsigned int normals_downsample)
{
	assert(isOrganizedCloud(organizedCloud));

	const unsigned int inc = normals_downsample;
	if(inc == 1) return;
	assert(inc > 0); //we divide by it later

	/*
	* compute the first x- and y-indices for which {x,y}_floor + inc is oob
	*/
	const unsigned int xRes = organizedCloud.width, yRes = organizedCloud.height;
	unsigned int xlast = xRes - (xRes % inc), ylast = yRes - (yRes % inc);
	if(xlast == xRes) xlast -= inc;
	if(ylast == yRes) ylast -= inc;

	for(unsigned int index = 0; index<organizedCloud.points.size(); index++)
	{
		const unsigned int x = index%xRes;
		const unsigned int y = index/xRes;

		//don't bother setting a normal for a point we don't care about
		if(!targetValidityGrid[x][y])
			continue;

		const unsigned int x_resid = x%inc;
		const unsigned int y_resid = y%inc;

		//normal already set for these
		if(x_resid == 0 && y_resid ==0)
			continue;

		const unsigned int x_floor = x - x_resid;
		const unsigned int y_floor = y - y_resid;

		/*
		 * avoid indexing out of bounds
		 */
		const unsigned int xinc = (x >= xlast) ? 0 : inc,
			yinc = (y >= ylast) ? 0 : inc;

		if(targetValidityGrid[x_floor][y_floor] &&
			targetValidityGrid[x_floor+xinc][y_floor] &&
			targetValidityGrid[x_floor][y_floor+yinc] &&
			targetValidityGrid[x_floor+xinc][y_floor+yinc])
		{
			//interpolate normal
			float x_r = ((float)x_resid)/inc; //in [0, 1)
			float y_r = ((float)y_resid)/inc;

			float coefCC = x_r*y_r;
			float coefCF = x_r*(1-y_r);
			float coefFC = (1-x_r)*y_r;
			float coefFF = (1-x_r)*(1-y_r);

			rgbd::eigen::Vector3f normalSum =
				coefFF * getNormal(organizedCloud,x_floor,y_floor) +
				coefFC * getNormal(organizedCloud,x_floor,y_floor+yinc) +
				coefCF * getNormal(organizedCloud,x_floor+xinc,y_floor) +
				coefCC * getNormal(organizedCloud,x_floor+xinc,y_floor+yinc);
			if(normalSum.squaredNorm() < 1e-6) //this does happen in practice -- EVH
			{
				targetValidityGrid[x][y] = false;
			}
			else
			{
				normalSum.normalize();

				float curvatureSum =
					coefFF * getCurvature(organizedCloud,x_floor,y_floor) +
					coefFC * getCurvature(organizedCloud,x_floor,y_floor+yinc) +
					coefCF * getCurvature(organizedCloud,x_floor+xinc,y_floor) +
					coefCC * getCurvature(organizedCloud,x_floor+xinc,y_floor+yinc);

				rgbd::eigen2ptNormal(organizedCloud.points[index], normalSum);
				organizedCloud.points[index].curvature = curvatureSum;
			}
		}
		else
		{
				//allow the normal to be set if there are for instance 3 valid
				//neighbors with similar depth values -- MSK
			targetValidityGrid[x][y] = false;
		}
	}
}

template <typename PointT>
    void setOrganizedNormals(
    		pcl::PointCloud<PointT> &organizedCloud,
    		boost::multi_array<bool, 2>& targetValidityGrid,
    		unsigned int normals_window_size,
    		unsigned int normals_downsample,
    		float normals_max_z_diff, 
		bool use_GPU_for_normals, const bool multithread)
    {
		assert(isOrganizedCloud(organizedCloud));

//    	rgbd::timer t;

    	//organized normals
    	geometry_msgs::Point32 origin;
    	pcl::PointCloud<PointT> withNormals;

#if ENABLE_OPENCL
	static int first_time = 1;
	static bool using_GPU_for_normals = false;

	ROS_INFO("cloudNormals: opencl enabled: about to computeOrganizedPointCloudNormals");
	if ( first_time == 1 )
	{
		first_time = 0;

		if ( use_GPU_for_normals == true )
		{
			bool ret = GPUcomputeOrganizedPointCloudNormals(withNormals, organizedCloud,normals_window_size,normals_downsample, organizedCloud.width,organizedCloud.height,normals_max_z_diff,origin);
			if ( ret == false )
			{
				ROS_INFO("cloudNormals: configured to use GPU for normals calculation, but can't execute kernel. Falling back to CPU computation.");
			}
			else
			{
				using_GPU_for_normals = true;
			}
		}
		else
		{
			//computeOrganizedPointCloudNormals(withNormals, organizedCloud,normals_window_size,normals_downsample, normals_max_z_diff,origin);
			computeOrganizedPointCloudNormals(withNormals, organizedCloud, targetValidityGrid, normals_window_size,normals_downsample, normals_max_z_diff,origin);
		}
	}
	else
	{
		if ( using_GPU_for_normals == true )
		{
			bool ret = GPUcomputeOrganizedPointCloudNormals(withNormals, organizedCloud,normals_window_size,normals_downsample, organizedCloud.width,organizedCloud.height,normals_max_z_diff,origin);
			if ( ret == false )
			{
				ROS_INFO("cloudNormals: configured to use GPU for normals calculation, but can't execute kernel. Falling back to CPU computation.");
				using_GPU_for_normals = false;
			}
		}
		else
		{
			//computeOrganizedPointCloudNormals(withNormals, organizedCloud,normals_window_size,normals_downsample, normals_max_z_diff,origin);
			computeOrganizedPointCloudNormals(withNormals, organizedCloud, targetValidityGrid, normals_window_size,normals_downsample, normals_max_z_diff,origin);
		}
	}
#else
	computeOrganizedPointCloudNormals(withNormals, organizedCloud, targetValidityGrid, normals_window_size,normals_downsample, normals_max_z_diff,origin, multithread);
#endif

    	//put the normals into organizedCloud because withNormals is the wrong dimensions

#if ENABLE_OPENCL
    	int smIncr = 1;

    	unsigned int smX, smY, lgX, lgY, lgIndex;
      for(unsigned int smIndex=0; smIndex<withNormals.points.size(); smIndex += smIncr) {
			if ( using_GPU_for_normals == true )
			{
				// GPU computation calculates all 307200 points, so we need to simulate any
				// requested downsampling here
				smX = smIndex%(withNormals.width*normals_downsample);
				smY = smIndex/(withNormals.width*normals_downsample);
				if ( smX % normals_downsample != 0 ) continue;
				if ( smY % normals_downsample != 0 ) continue;
				lgX = smX;
				lgY = smY;
				lgIndex = smIndex;
			}
			else
			{
				smX = smIndex%withNormals.width;
				smY = smIndex/withNormals.width;
				lgX = smX*normals_downsample;
				lgY = smY*normals_downsample;
				lgIndex = lgX + lgY*organizedCloud.width;
			}

		//ROS_INFO("smX,Y=%d,%d smIndex=%d lgX,Y=%d,%d lgIndex=%d", smX, smY, smIndex, lgX, lgY, lgIndex );
#else
		for(unsigned int smY = 0; smY < withNormals.height; smY++)
			for(unsigned int smX = 0; smX < withNormals.width; smX++)
			{
				const unsigned int smIndex = smY * withNormals.width + smX;
				unsigned int lgX = smX*normals_downsample;
				unsigned int lgY = smY*normals_downsample;
				unsigned int lgIndex = lgX + lgY*organizedCloud.width;
#endif
				if(targetValidityGrid[lgX][lgY])
				{
					organizedCloud.points[lgIndex].normal[0] = withNormals.points[smIndex].normal[0];
					organizedCloud.points[lgIndex].normal[1] = withNormals.points[smIndex].normal[1];
					organizedCloud.points[lgIndex].normal[2] = withNormals.points[smIndex].normal[2];
					organizedCloud.points[lgIndex].curvature = withNormals.points[smIndex].curvature;
				}
			}

    	interpolateNormals(organizedCloud, targetValidityGrid, normals_downsample);

//    	t.stop("setOrganizedNormals");
    }

#ifdef UNUSED
/*
 * auxiliary to compute2DNormalsAdaptively()
 */
namespace aux
{

vector<int> getNeighborIndices(const unsigned int x, const unsigned int y, const unsigned int radius, const boost::multi_array<bool, 2>& validityMap);

template <typename PointT>
Vector4d computeBestFitPlane(const pcl::PointCloud<PointT>& organizedCloud, const unsigned int ptIndex, const vector<int>& indices)
{
	geometry_msgs::Point32 viewpoint;
	viewpoint.x = viewpoint.y = viewpoint.z = 0;
	pcl::NormalEstimation<PointT, PointT> ester;
	rgbd::eigen::Vector4d plane_parameters;
	double curvature;
	ester.computePointNormal (organizedCloud, indices, plane_parameters, curvature);
	ester.flipNormalTowardsViewpoint (plane_parameters, organizedCloud.points[ptIndex], viewpoint);
	return plane_parameters;
}

/*
 * the higher the score, the better the plane fits the neighborhood
 */
template <typename PointT>
double computePlanarityScore(const Vector4d& plane, const pcl::PointCloud<PointT>& organizedCloud, const vector<int>& indices, const vector<double>& minDepthRatios, const vector<double>& maxDepthRatios)
{
	const Vector3d normal(plane[0], plane[1], plane[2]);
	const double offset = plane[3];

	double score = 0;
	for(unsigned int i = 0; i < indices.size(); i++)
	{
		const unsigned int index = indices[i];
		/*
		 * points at the near and far ends of the original pt's depth bin
		 */
		const Vector3d closePt(organizedCloud.points[index].x.x(), organizedCloud.points[index].x.y(), organizedCloud.points[index].x.z() * minDepthRatios[index]),
			farPt(organizedCloud.points[index].x.x(), organizedCloud.points[index].x.y(), organizedCloud.points[index].x.z() * maxDepthRatios[index]);
		const double closeOffset = -normal.dot(closePt), farOffset = -normal.dot(farPt);
		const double minOffset = std::min(closeOffset, farOffset), maxOffset = std::max(closeOffset, farOffset);
		/*
		 * penalize distance of the plane from the pt's depth bin
		 */
		if(offset < minOffset) score += std::max(-1.0, 1 - (minOffset - offset) / (maxOffset - minOffset) / 2);
		else if(offset > maxOffset) score += std::max(-1.0, 1 - (offset - maxOffset) / (maxOffset - minOffset) / 2);
		else score += 1;
	}
	return score / indices.size()
		* log(indices.size()); //prefer large neighborhoods
}

/*
 * work around boost::bind() only accepting 10 arguments
 */
struct adaptive2dNormalsOptions
{
	unsigned int minNbrhoodRadius, maxNbrhoodRadius;
	unsigned int minNumNbrs;
};

} //namespace

template <typename PointT>
void compute2DNormalsAdaptivelyThreadMain(pcl::PointCloud<PointT>& organizedCloud, const boost::multi_array<bool, 2>& validityMap, const vector<double>& minDepthRatios, const vector<double>& maxDepthRatios, unordered_map<unsigned int, unsigned int>& bestRadiusHistogram, const aux::adaptive2dNormalsOptions& opts, const unsigned int firstXIndex, const unsigned int lastXIndex)
{
	const unsigned int xSize = validityMap.size(), ySize = validityMap[0].size();

	for(unsigned int i = opts.minNbrhoodRadius; i <= opts.maxNbrhoodRadius; i++) bestRadiusHistogram[i] = 0;

	for(unsigned int x = firstXIndex; x <= lastXIndex; x++)
		for(unsigned int y = 0; y < ySize; y++)
			if(validityMap[x][y])
			{
				vector<int> bestNbrIndices;
				Vector4d bestPlane;
				double bestPlaneFitness;
				optional<unsigned int> bestRadius;
				for(unsigned int r = opts.minNbrhoodRadius; r <= opts.maxNbrhoodRadius; r++) //TODO sometimes avoid running at all nbrhood sizes?
				{
				//	cout << "r = " << r << endl;
					const vector<int> nbrIndices = aux::getNeighborIndices(x, y, r, validityMap);
					if(nbrIndices.size() >= opts.minNumNbrs)
					{
						const Vector4d plane = aux::computeBestFitPlane(organizedCloud, y * xSize + x, nbrIndices);
//						cout << " plane = " << plane.transpose() << endl;
						const double planeFitness = aux::computePlanarityScore(plane, organizedCloud, nbrIndices, minDepthRatios, maxDepthRatios);
					//	cout << " planarity = " << planeFitness << endl;
						if(!bestRadius || planeFitness > bestPlaneFitness)
						{
							bestRadius = r;
							bestNbrIndices = nbrIndices;
							bestPlane = plane;
							bestPlaneFitness = planeFitness;
						}
					}
				}
				assert(bestRadius); //if there aren't enough points even in a large nbrhood, we'll need to throw away the point, probably
				//getchar();
				bestRadiusHistogram[*bestRadius]++;

				organizedCloud.points[y * xSize + x].normal = bestPlane.head<3>();
			}
}

template <typename PointT>
void compute2DNormalsAdaptively(pcl::PointCloud<PointT>& organizedCloud, const boost::multi_array<bool, 2>& validityMap)
{
	assert(rgbd::isOrganizedCloud(organizedCloud));
	const unsigned int xSize = validityMap.shape()[0], ySize = validityMap.shape()[1];

	//TODO make params
	const unsigned int maxNbrhoodRadius = 12, minNbrhoodRadius = 2; //halfwidth of square
	const unsigned int minNumNbrs = 15;

	/*
	 * compute sufficient statistics for plane fitting
	 */
	vector<double> minDepthRatios(organizedCloud.points.size()), maxDepthRatios(organizedCloud.points.size());
	for(unsigned int x = 0; x < xSize; x++)
		for(unsigned int y = 0; y < ySize; y++)
			if(validityMap[x][y])
			{
				const unsigned int index = y * xSize + x;
				/*
				 * pt depths (in mm)
				 */
				const unsigned int depth = organizedCloud.points[index].z * 1000;
				/*
				 * min and max depths that have the same depth bin
				 */
				const pair<const unsigned short*, const unsigned short*> depthsForDepthIndex = std::equal_range(indexByDepth, indexByDepth + INDEX_BY_DEPTH_ARRAY_SIZE, indexByDepth[depth]);
				const unsigned int minDepthForDepthIndex = std::distance(indexByDepth, depthsForDepthIndex.first),
					maxDepthForDepthIndex = std::distance(indexByDepth, depthsForDepthIndex.second) - 1; //will be slightly off if distance() returns end of array

				minDepthRatios[index] = (double)minDepthForDepthIndex / depth;
				maxDepthRatios[index] = (double)maxDepthForDepthIndex / depth;
			}

	unordered_map<unsigned int, unsigned int> bestRadiusHistogram; //radius -> # points selecting it as best
	for(unsigned int i = minNbrhoodRadius; i <= maxNbrhoodRadius; i++) bestRadiusHistogram[i] = 0;

	const unsigned int numThreads = getSuggestedThreadCount();
	vector<boost::shared_ptr<rgbd::thread> > threads(numThreads);
	vector<unordered_map<unsigned int, unsigned int> > threadBestRadiusHistograms(numThreads);
	const vector<unsigned int>& indices = partitionEvenly(xSize, numThreads);
	aux::adaptive2dNormalsOptions opts;
	opts.minNbrhoodRadius = minNbrhoodRadius;
	opts.maxNbrhoodRadius = maxNbrhoodRadius;
	opts.minNumNbrs = minNumNbrs;
	for(unsigned int i = 0; i < numThreads; i++)
		threads[i] = boost::make_shared<rgbd::thread>(rgbd::bind(compute2DNormalsAdaptivelyThreadMain<PointT>, rgbd::ref(organizedCloud), rgbd::cref(validityMap), rgbd::cref(minDepthRatios), rgbd::cref(maxDepthRatios), rgbd::ref(threadBestRadiusHistograms[i]), opts, indices[i], indices[i + 1] - 1));
	for(unsigned int i = 0; i < numThreads; i++)
	{
		Peter: note that this is broken (join should be in its own for loop) but commented out so I am not fixing it
		threads[i]->join();
		for(unsigned int j = minNbrhoodRadius; j <= maxNbrhoodRadius; j++) bestRadiusHistogram[j] += threadBestRadiusHistograms[i][j];
	}
	cout << "best-radius histogram:" << endl;
	for(unsigned int i = minNbrhoodRadius; i <= maxNbrhoodRadius; i++) cout << ' ' << i << ": " << bestRadiusHistogram[i] << endl;
	cout << endl;
}
#endif

template <typename PointT>
void computeNormals2d(pcl::PointCloud<PointT>& cloud, const unsigned int windowSize, const unsigned int downsampleFactor, const double maxDepthDifference, const unsigned int width, const unsigned int height, bool use_GPU_for_normals )
{
	assert(!isOrganizedCloud(cloud)); //this function may actually work for organized clouds but obviously isn't intended for them -- EVH 20100812
	rgbd::timer t;

	// get an organized cloud and the associated validity_grid
	pcl::PointCloud<PointT> organized_cloud;
	organized_cloud.width = width;
	organized_cloud.height = height;
	boost::multi_array<bool, 2> validity_grid;
	computeOrganizedPointCloud<PointT>(cloud, organized_cloud, validity_grid);
	for(unsigned int i = 0; i < organized_cloud.points.size(); i++) markNormalInvalid(organized_cloud.points[i]); //if the cloud already had normals, clear them

	setOrganizedNormals<PointT>(organized_cloud, validity_grid, windowSize, downsampleFactor, maxDepthDifference, use_GPU_for_normals);

	// get only valid points out of organized cloud
	std::vector<unsigned int> valid_indices;
	getValidIndices(validity_grid, valid_indices);
	rgbd::downsampleFromIndices(organized_cloud, cloud, valid_indices);

	t.stop("computeNormals2d total");
}

template <typename PointT>
void computeNormals2dWithStdevs(pcl::PointCloud<PointT>& cloud, std::vector<rgbd::eigen::Matrix3f>& stdevs, const normals2dParams& params)
{
	assert(!isOrganizedCloud(cloud)); //this function may actually work for organized clouds but obviously isn't intended for them -- EVH 20100912
	rgbd::timer t;

	// get an organized cloud and the associated validity_grid
	pcl::PointCloud<PointT> organized_cloud;
	organized_cloud.width = 640;
	organized_cloud.height = 480;
	boost::multi_array<bool, 2> validity_grid;
	rgbd::computeOrganizedPointCloud<PointT>(cloud, organized_cloud, validity_grid);
	const boost::multi_array<bool, 2> validityBeforeNormals = validity_grid;
	for(unsigned int i = 0; i < organized_cloud.points.size(); i++) markNormalInvalid(organized_cloud.points[i]); //if the cloud already had normals, clear them

	computeOrganizedPointCloudNormalsWithStdevs(organized_cloud, stdevs, organized_cloud, validity_grid, params);

	// get only valid points out of organized cloud
	std::vector<unsigned int> valid_indices;
	rgbd::getValidIndices(validityBeforeNormals, valid_indices);
	rgbd::downsampleFromIndices(organized_cloud, cloud, valid_indices);
	std::vector<rgbd::eigen::Matrix3f> validStdevs(valid_indices.size());
	copySelected(stdevs, valid_indices, validStdevs.begin());
	stdevs = validStdevs;

	t.stop("do computeNormals2dWithStdevs");
}

template <typename PointT>
rgbd::eigen::Vector3f getNormal(pcl::PointCloud<PointT> &organizedCloud, unsigned int x, unsigned int y)
{
	assert(isOrganizedCloud(organizedCloud));
	const PointT& pt = organizedCloud.points[y * organizedCloud.width + x];
	return ptNormal2eigen<rgbd::eigen::Vector3f>(pt);
}

template <typename PointT>
float getCurvature(pcl::PointCloud<PointT> &organizedCloud, unsigned int x, unsigned int y)
{
	assert(isOrganizedCloud(organizedCloud));
	return organizedCloud.points[y * organizedCloud.width + x].curvature;
}

template <typename PointT, typename PointVectorT>
void pointCloudNormalsToEigenNormals(const pcl::PointCloud<PointT>& point_cloud, PointVectorT& normals)
{
	BOOST_STATIC_ASSERT((boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector3f> >::value || boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector4f, rgbd::eigen::aligned_allocator<rgbd::eigen::Vector4f> > >::value));\
	typedef typename PointVectorT::value_type TgtPointT;

	normals.resize(point_cloud.points.size());
	for(unsigned int i = 0; i < point_cloud.points.size(); i++) normals[i] = ptNormal2eigen<TgtPointT>(point_cloud.points[i]);
}

template <typename PointT>
void computeOrganizedPointCloud(const pcl::PointCloud<PointT>& cloud, pcl::PointCloud<PointT>& organizedCloud, boost::multi_array<bool, 2>&validity)
{
	assert(isOrganizedCloud(organizedCloud));
	organizedCloud.header = cloud.header;
	validity.resize(boost::extents[organizedCloud.width][organizedCloud.height]);
	std::fill(validity.data(), validity.data() + validity.num_elements(), false);
	organizedCloud.points.resize(organizedCloud.width*organizedCloud.height);
	
	for(unsigned int i=0; i<cloud.points.size(); i++){
		unsigned int x = cloud.points[i].imgX;
		unsigned int y = cloud.points[i].imgY;
		unsigned int index = x + organizedCloud.width*y;
		validity[x][y] = true;
		organizedCloud.points[index] = cloud.points[i];
	}
}

/*
 * unorganizedCloud will be allocated; if it has img{X,Y} fields they'll be set
 *
 * all fields will be copied
 */
template <typename PointT>
void organized2unorganizedCloud(const pcl::PointCloud<PointT>& organizedCloud, const boost::multi_array<bool, 2>&validity, pcl::PointCloud<PointT>& unorganizedCloud)
{
	uint32_t numValidPts = 0;
	for(const bool* i = validity.data(); i != validity.data() + validity.num_elements(); i++)
		if(*i)
			numValidPts++;

	unorganizedCloud.points.resize(numValidPts);
	unorganizedCloud.width = numValidPts;
	unorganizedCloud.height = 1;

	size_t l = 0;
	const bool* v = validity.data();
	for(uint32_t x = 0; x < validity.shape()[0]; x++)
		for(uint32_t y = 0; y < validity.shape()[1]; y++, v++)
			if(*v)
			{
				unorganizedCloud.points[l] = organizedCloud.at(x, y);
				rgbd::setImgCoords<PointT>(unorganizedCloud.points[l], x, y);
				l++;
			}
}

template <typename PointT>
void getValidityGrid(const pcl::PointCloud<PointT> & organized_cloud, boost::multi_array<bool, 2>& validity_grid )
{
	assert(isOrganizedCloud(organized_cloud));

	validity_grid.resize(boost::extents[organized_cloud.width][organized_cloud.height]);
	for (unsigned int y = 0; y < organized_cloud.height; y++) {
		for (unsigned int x = 0; x < organized_cloud.width; x++) {
			float depth_value = organized_cloud.points[y * organized_cloud.width + x].z;
			// depth_value must be positive
			validity_grid[x][y] = (depth_value > 0.0);
		}
	}
}

template <typename PointT>
void pointcloud_normal_marker (pcl::PointCloud<PointT> const& cloud, visualization_msgs::Marker& marker, const std::string& marker_namespace, float line_length, float line_width, float red, float green, float blue)
{
	marker.header = cloud.header;

	// these form the identifier
	marker.ns = marker_namespace;
	marker.id = 0;

	marker.type = visualization_msgs::Marker::LINE_LIST;
	marker.action = visualization_msgs::Marker::ADD;
	marker.pose.position.x = 0;
	marker.pose.position.y = 0;
	marker.pose.position.z = 0;
	marker.pose.orientation.x = 0.0;
	marker.pose.orientation.y = 0.0;
	marker.pose.orientation.z = 0.0;
	marker.pose.orientation.w = 1.0;
	marker.scale.x = line_width;
	marker.scale.y = 1;
	marker.scale.z = 1;
	marker.color.a = 1.0;
	marker.color.r = red;
	marker.color.g = green;
	marker.color.b = blue;

	unsigned int point_count = cloud.points.size();

	marker.points.resize (2 * point_count);
	for (unsigned int i = 0; i < point_count; i++)
	{
		marker.points[2*i].x = cloud.points[i].x;
		marker.points[2*i].y = cloud.points[i].y;
		marker.points[2*i].z = cloud.points[i].z;
		marker.points[2*i+1].x = cloud.points[i].x + cloud.points[i].normal[0] * line_length;
		marker.points[2*i+1].y = cloud.points[i].y + cloud.points[i].normal[1] * line_length;
		marker.points[2*i+1].z = cloud.points[i].z + cloud.points[i].normal[2] * line_length;
	}
}

template <typename PointT>
bool computeOrganizedNormalForIndex(
		pcl::PointCloud<PointT> & organized_cloud,
		unsigned int index,
		int k,
		int min_neighbors,
		double max_z,
		const geometry_msgs::Point32 & viewpoint)
{
	assert(isOrganizedCloud(organized_cloud));

	// as opposed to a validity grid, check if depth of indexed point is > 0
	if (organized_cloud.points[index].z <= 0) {
		markNormalInvalid(organized_cloud.points[index]);
		return false;
	}

	unsigned int row = index / organized_cloud.width;
	unsigned int col = index % organized_cloud.width;
	// Get all point neighbors in a k x k window
	int l = 0;
	std::vector<int> nn_indices((k + k + 1) * (k + k + 1));
	for(int y = std::max(0, (int)row - k); y <= std::min((int)organized_cloud.height - 1, (int)row + k); y++)
		for(int x = std::max(0, (int)col - k); x <= std::min((int)organized_cloud.width - 1, (int)col + k); x++)
		{
			unsigned int neighbor_index = y * organized_cloud.width + x;
			// treat depths <= 0 as invalid
			if (organized_cloud.points[neighbor_index].z <= 0) continue;
			// If max_z is not set (<=0) or the depth difference is less than max_z, this is a valid neighbor
			if(max_z <= 0 || fabs(organized_cloud.points[neighbor_index].z - organized_cloud.points[neighbor_index].z) < max_z)
				nn_indices[l++] = neighbor_index;
		}

	// if not enough valid neighbors
	if(l < std::max(min_neighbors, 4)){
		markNormalInvalid(organized_cloud.points[index]);
		return false;
	}

	nn_indices.resize (l);
	// Compute the point normals (nx, ny, nz), surface curvature estimates (c)
	rgbd::eigen::Vector4f plane_parameters;
	float curvature;

	// perhaps slow to create a new one of these every time??
	// This normal estimation can do range-based nearest neighbor as well
    pcl::NormalEstimation<PointT, PointT> normal_estimation;
    normal_estimation.computePointNormal (organized_cloud, nn_indices, plane_parameters, curvature);
    pcl::flipNormalTowardsViewpoint (organized_cloud.points[index], viewpoint.x, viewpoint.y, viewpoint.z, plane_parameters);

	organized_cloud.points[index].normal[0] = plane_parameters[0];
	organized_cloud.points[index].normal[1] = plane_parameters[1];
	organized_cloud.points[index].normal[2] = plane_parameters[2];
	organized_cloud.points[index].curvature = curvature;
	return true;
}

template <typename PointT>
void computeOrganizedNormalsForIndices(
		pcl::PointCloud<PointT> & organized_cloud,
		const std::vector<unsigned int> & indices,
		std::vector<bool> & index_has_valid_normal,
		int k,
		int min_neighbors,
		double max_z,
		const geometry_msgs::Point32 & viewpoint)
{
	assert(isOrganizedCloud(organized_cloud));
	index_has_valid_normal.resize(indices.size());
    for (unsigned int i = 0; i < indices.size(); i++) {
    	unsigned int index = indices[i];
    	index_has_valid_normal[i] = computeOrganizedNormalForIndex(organized_cloud, index, k, min_neighbors, max_z, viewpoint);
    }
}



#if ENABLE_OPENCL
	/*
	 * Do the normals computation on the GPU.
	 */
	template <typename PointT>
	bool GPUcomputeOrganizedPointCloudNormals (
    		pcl::PointCloud<PointT> &points, const pcl::PointCloud<PointT> &surface,
    		int k, int downsample_factor, int width, int height, double max_z,
    		const geometry_msgs::Point32 &viewpoint)
	{

		ROS_INFO ("GPUcomputeOrganizedPointCloudNormals: point size: %d downsample: %d k: %d max_z: %f", surface.points.size(), downsample_factor , k, max_z );
		rgbd::timer gputimer;

		// get the point data into a form that we can convert into a opencl image
		assert(width <= 640 && height <= 640);
		float new_input[640][640];
		// surface points is an array organized row major, i.e., row 0, row 1 ... row 479
		for ( int i = 0 ; i < (int)surface.points.size(); i++ )
		{
			int x = surface.points[i].imgX;
			int y = surface.points[i].imgY;
			float z = surface.points[i].z;
			// something odd happens when i convert this to a cl_image, such that if I
			// do what seems reasonable, (i.e, new_input[x][y] = z) then i have to
			// use read_image(y,x) to get the correct data. ?!? More painfully, this
			// memory access pattern leads to a performance hit of 20 ms over
			// read_image(x,y). So I will therefore build new_input as a square matrix
			// and swap the rows and columns, so I can use the faster parameters
			//new_input[ x ][ y ] = z;
			new_input[ y ][ x ] = z;
			//ROS_INFO("in: x,y=%d,%d z=%f i=%d sp.z=%f (but x,y are swapped in actual input)", x, y, z, i, surface.points[i].z );
		}

		// allocate memory for the kernel output. FIXME: implication for static allocation of output_mem?
		const int num_points = 640 * 640;
		cl_float4 *output = ( cl_float4 *)calloc( num_points , sizeof(cl_float4) );
		if ( output == NULL )
		{
			ROS_INFO("could not allocate memory for output");
			return false;
		}

		// execute the kernel
		//if ( exec_kernel(new_input, new_viewpoint, output, "normals.cl", "compute_normals", k , max_z ) < CL_SUCCESS )
		if ( exec_kernel(new_input, viewpoint, output, "normals.cl", "compute_normals", k , max_z ) < CL_SUCCESS )
		{
			ROS_INFO("exec kernel failed");
    			free(output);
			return false;
		}

		// go grab the kernel output and store the returns normals data in the point array
		points = surface;
		for ( int i = 0; i < (int)surface.points.size(); i++ )
		{
			// figure out where in the output buffer I can find the normals data for this point
			int x = surface.points[i].imgX;
			int y = surface.points[i].imgY;
			// data is returned in row major order
			int offset = ( y * 640 ) + x;

			// filter out points where the calculations weren't made -- either
			// because the z value was invalid (9999), or because there weren't enough
			// neighbors with valid z values (8888)
			cl_float4 *pp = (cl_float4 *)(output + offset);

			if ( pp->s[0] == 9999 || pp->s[0] == 8888 )
			{
				//ROS_INFO("gpu-dropped: i=%d offset=%d x,y,z=%d,%d,%f pp=%f,%f,%f,%f", i, offset, x, y, surface.points[i].z, pp->s[0], pp->s[1], pp->s[2], pp->s[3] );
				continue;
			}

			points.points[i].normal[0] = pp->s[0];
			points.points[i].normal[1] = pp->s[1];
			points.points[i].normal[2] = pp->s[2];
			points.points[i].curvature = pp->s[3];

			//ROS_INFO("gpu: i=%d offset=%d x,y,z=%d,%d,%f pp=%f,%f,%f,%f", i, offset, x, y, surface.points[i].z, pp->s[0], pp->s[1], pp->s[2], pp->s[3] );
		}

		free(output);
		gputimer.stop("polly: organized normals (kernel wrapper)");
		return true;
	}
#endif

} //namespace
