/*
 * cloudSampling: selecting points to throw away
 *
 * Evan Herbst
 * 6 / 22 / 10
 */

#include <numeric> //accumulate()
#include <boost/concept/assert.hpp>
#include <boost/concept_check.hpp>
#include <boost/range.hpp>
#include <boost/range/concepts.hpp>
#include "rgbd_util/mathUtils.h" //copySelected()
#include "pcl_rgbd/pointTypes.h" //ptX2eigen

namespace rgbd
{

template <typename PointT>
std::vector<unsigned int> getRadiusDownsamplingIndices(const pcl::PointCloud<PointT>& cloud, rgbd::eigen::Vector3f center,float radius,bool radiusIsMax)
{
	std::vector<unsigned int> acceptedIndices;
	float squaredRadius = radius*radius;
	for(unsigned int i=0; i<cloud.points.size(); i++){
		const rgbd::eigen::Vector3f pt(cloud.points[i].x, cloud.points[i].y, cloud.points[i].z);
		if(((pt - center).squaredNorm() < squaredRadius) == radiusIsMax)
			acceptedIndices.push_back(i);
	}

	return acceptedIndices;
}

template <typename PointT>
std::vector<unsigned int> getImageRadiusDownsamplingIndices(const pcl::PointCloud<PointT>& cloud, int source_width, int source_height, float radius_ratio)
{
	const rgbd::eigen::Vector2f center(source_width / 2.0, source_height / 2.0);

	const float max_dist_squared = center.squaredNorm();
	const float cutoff_dist_squared = radius_ratio * radius_ratio * max_dist_squared;

	std::vector<unsigned int> acceptedIndices;
	for (unsigned int i = 0; i < cloud.points.size(); i++) {
		const rgbd::eigen::Vector2f pointImage((float)cloud.points[i].imgX, (float)cloud.points[i].imgY);
		if ((pointImage - center).squaredNorm() <= cutoff_dist_squared) {
			acceptedIndices.push_back(i);
		}
	}
	return acceptedIndices;
}

template <typename PointT>
std::vector<unsigned int> getPixelSpaceDownsamplingIndices(const pcl::PointCloud<PointT>& cloud, const unsigned int downsampleFactor)
{
	std::vector<unsigned int> result;
	for(unsigned int i = 0; i < cloud.points.size(); i++)
		if(cloud.points[i].imgX % downsampleFactor == 0 && cloud.points[i].imgY % downsampleFactor == 0)
			result.push_back(i);
	return result;
}

template <typename PointT>
std::vector<unsigned int> getPixelSpaceDownsamplingIndices(const pcl::PointCloud<PointT>& cloud, std::vector<unsigned int> indices, const unsigned int downsampleFactor)
{
	std::vector<unsigned int> result;
	for(unsigned int i = 0; i < indices.size(); i++)
		if(cloud.points[indices[i]].imgX % downsampleFactor == 0 && cloud.points[indices[i]].imgY % downsampleFactor == 0)
			result.push_back(indices[i]);
	return result;
}

template <typename PointT>
std::vector<unsigned int> getWellSeparatedPointsIndices(const pcl::PointCloud<PointT>& cloud, const float minDist, const unsigned int maxSamplesOnePt)
{
	std::vector<unsigned int> indices;
	while(true)
	{
		unsigned int j = 0;
		for(; j < maxSamplesOnePt; j++)
		{
			const unsigned int pt = rand() % cloud.points.size();
			bool tooClose = false;
			for(std::vector<unsigned int>::const_iterator k = indices.begin(); k != indices.end(); k++)
				if(pt == *k || (rgbd::ptX2eigen<rgbd::eigen::Vector3f>(cloud.points[pt]) - rgbd::ptX2eigen<rgbd::eigen::Vector3f>(cloud.points[*k])).squaredNorm() < sqr(minDist))
				{
					tooClose = true;
					break;
				}
			if(!tooClose)
			{
				indices.push_back(pt);
				break;
			}
		}
		if(j == maxSamplesOnePt) break;
	}
	return indices;
}

template <typename PointT>
std::vector<unsigned int> getWellSeparatedPointsIndices(const pcl::PointCloud<PointT>& cloud, const std::vector<double>& ptWeights, const float minDist, const unsigned int maxSamplesOnePt = 100)
{
	assert(ptWeights.size() == cloud.points.size());
	std::vector<double> probs = ptWeights;
	const double sum = std::accumulate(probs.begin(), probs.end(), 0.0);
	for(unsigned int i = 0; i < probs.size(); i++) probs[i] /= sum;

	std::vector<unsigned int> indices;
	while(true)
	{
		unsigned int j = 0;
		for(; j < maxSamplesOnePt; j++)
		{
			const unsigned int pt = sampleDiscreteDistribution(probs);
			bool tooClose = false;
			for(std::vector<unsigned int>::const_iterator k = indices.begin(); k != indices.end(); k++)
				if(pt == *k || (rgbd::ptX2eigen<rgbd::eigen::Vector3f>(cloud.points[pt]) - rgbd::ptX2eigen<rgbd::eigen::Vector3f>(cloud.points[*k])).squaredNorm() < sqr(minDist))
				{
					tooClose = true;
					break;
				}
			if(!tooClose)
			{
				indices.push_back(pt);
				break;
			}
		}
		if(j == maxSamplesOnePt) break;
	}
	return indices;
}

template <typename PointT, typename PointT2, typename IndexRangeT>
void downsampleFromIndices(const pcl::PointCloud<PointT>& cloud, pcl::PointCloud<PointT2>& downsampledCloud, const IndexRangeT& acceptedIndices)
{
	downsampledCloud.points.resize(acceptedIndices.size());
	copySelected(cloud.points, acceptedIndices, downsampledCloud.points.begin());
	downsampledCloud.width = downsampledCloud.points.size();
	downsampledCloud.height = 1;
	downsampledCloud.header = cloud.header;
}

} //namespace
