/*
 * cloudSearchTrees: search trees over point clouds (using 3-d position and possibly other things)
 *
 * Evan Herbst
 * 6 / 16 / 10
 */

#include <cmath>
#include <vector>
#include <boost/multi_array.hpp>
#include <pcl/kdtree/kdtree_flann.h>
#include <geometry_msgs/Point32.h>
#include "pcl_rgbd/pointTypes.h"
#include "pcl_rgbd/cloudSearchTrees.h"

namespace rgbd
{

/*
 * efficient alternative to libANN
 */
template <typename PointT>
boost::shared_ptr<kdtree2> createKDTree2(const pcl::PointCloud<PointT>& cloud)
{
	boost::multi_array<float, 2> data(boost::extents[cloud.points.size()][3]);
	float* j = data.data();
	for (unsigned int i = 0; i < cloud.points.size(); i++)
	{
		*j++ = cloud.points[i].x;
		*j++ = cloud.points[i].y;
		*j++ = cloud.points[i].z;
	}
	boost::shared_ptr<kdtree2> tree(new kdtree2(data, true/* rearrange input pts */)); //if we rearrange input pts the tree doesn't need our multi_array to stay in scope (it has a copy)
	tree->sort_results = true;
	return tree;
}
/*
 * apply xform to all points in cloud
 */
template <typename PointT>
boost::shared_ptr<kdtree2> createKDTree2(const pcl::PointCloud<PointT>& cloud, const rgbd::eigen::Affine3f& xform)
{
	boost::multi_array<float, 2> data(boost::extents[cloud.points.size()][3]);
	float* j = data.data();
	for(unsigned int i = 0; i < cloud.points.size(); i++)
	{
		const rgbd::eigen::Vector4f x = xform * rgbd::ptX2eigen<rgbd::eigen::Vector4f>(cloud.points[i]);
		*j++ = x.x();
		*j++ = x.y();
		*j++ = x.z();
	}
	boost::shared_ptr<kdtree2> tree(new kdtree2(data, true/* rearrange input pts */)); //if we rearrange input pts the tree doesn't need our multi_array to stay in scope (it has a copy)
	tree->sort_results = true;
	return tree;
}

template <typename PointT>
typename pcl::KdTree<PointT>::Ptr createKDTree(const typename pcl::PointCloud<PointT>::Ptr& cloudPtr)
{
	typename pcl::KdTreeFLANN<PointT>::Ptr tree(new typename pcl::KdTreeFLANN<PointT>);
	tree->setInputCloud(cloudPtr);
	return tree;
}

template <typename PointT>
std::tuple<std::vector<int>, std::vector<float>> searchKDTree(pcl::KdTree<PointT>& tree, const kdtreeNbrhoodSpec& nspec, const PointT& qpt)
{
	std::tuple<std::vector<int>, std::vector<float>> result;
	std::vector<int>& indices = std::get<0>(result);
	std::vector<float>& dists = std::get<1>(result);
	if(nspec.numNbrs < UINT_MAX)
	{
		tree.nearestKSearch(qpt, nspec.numNbrs, indices, dists);
		//TODO additionally bound with radius
	}
	else
	{
		tree.radiusSearch(qpt, sqrt(nspec.sqrNbrhoodRadius), indices, dists);
		//TODO additionally bound with nbr count
	}
	return result;
}

} //namespace
