/*
 * cloudUncertainty: uncertainty modeling of single clouds
 *
 * Evan Herbst
 * 1 / 9 / 11
 */

#include <cmath> //rint()
#include <algorithm> //lower_bound()
#include <iterator> //distance()
#include <stdexcept>
#include "rgbd_util/mathUtils.h"
#include "primesensor_tools/primeSense_quantization_v5.h"

namespace rgbd
{

/*
 * return a covariance matrix for the position of each point in the cloud independently
 *
 * dthetaX, dthetaY: angle differences (in rad) between adjacent pixels in x and y
 * xySigmaInPixels: covariance in x and y directions is this multiplier x the distance between pixels
 * zSigmaInDepthBins: covariance in depth direction is this multipler x the size of the depth bin containing the point
 *
 * points with z <= 0 will have invalid entries
 */
template <typename PointT>
std::vector<rgbd::eigen::Matrix3d> getIndependentPointCovariances(const pcl::PointCloud<PointT>& cloud, const double dthetaX, const double dthetaY, const double xySigmaInPixels, const double zSigmaInDepthBins)
{
	std::vector<rgbd::eigen::Matrix3d> covs(cloud.points.size());
	for(unsigned int i = 0; i < cloud.points.size(); i++)
	{
		const rgbd::eigen::Vector3d pt(cloud.points[i].x, cloud.points[i].y, cloud.points[i].z);
		if(pt.z() > 0) //true for a valid pt
		{
			const double depth = pt.z(); //want pt.norm() but then can't index indexByDepth using depth since it might be more than the max sensor reading

			/*
			 * compute covariance in a frame in which it's diagonal (ie where depth dir is an axis)
			 */
			rgbd::eigen::Matrix3d& cov = covs[i];
			cov.setZero();
			//sigma_depth: find cur depth index; take harmonic mean of dists to front of prev depth's index set and back of next depth's index set (TODO parameterize # bins to cover)
			const unsigned int depthIndex = indexByDepth[(unsigned int)rint(depth * 1000)];
			const unsigned int minDepthForCov = std::distance(indexByDepth, std::lower_bound(indexByDepth, indexByDepth + INDEX_BY_DEPTH_ARRAY_SIZE, depthIndex - 1)), //fails if we hit beginning of depthByIndex
				maxDepthForCov = std::distance(indexByDepth, std::upper_bound(indexByDepth, indexByDepth + INDEX_BY_DEPTH_ARRAY_SIZE, depthIndex + 1)); //fails if we hit end of depthByIndex
			cov(2, 2) = sqr((2 * zSigmaInDepthBins) * 2 / (1.0 / std::max(depth - .001 * (double)minDepthForCov, 3e-4) + 1.0 / std::max(.001 * (double)maxDepthForCov - depth, 3e-4))); //avoid singularity and remember the min sensable difference is 1e-3 (meters)
			//sigma_x, sigma_y: use small-angle approximation
			cov(0, 0) = sqr(xySigmaInPixels * dthetaX * depth);
			cov(1, 1) = sqr(xySigmaInPixels * dthetaY * depth);

			/*
			 * convert to world (standard) frame (will get nondiag mtx)
			 */
			const rgbd::eigen::Vector3d unitBasis1 = pt.normalized();
			rgbd::eigen::Vector3d vec2(1, 0, 0); //should be lin indep from basis1
			if(vec2.dot(unitBasis1) > .9) vec2 = rgbd::eigen::Vector3d(0, 1, 0);
			const rgbd::eigen::Vector3d unitBasis2 = (vec2 - vec2.dot(unitBasis1) * unitBasis1).normalized();
			const rgbd::eigen::Vector3d unitBasis3 = unitBasis1.cross(unitBasis2);
			rgbd::eigen::Matrix3d frameXform;
			frameXform.col(0) = unitBasis2;
			frameXform.col(1) = unitBasis3;
			frameXform.col(2) = unitBasis1;
			cov = frameXform.transpose() * cov * frameXform;
		}
		else covs[i] = rgbd::eigen::Matrix3d::Constant(0.0 / 0/* nan */);
	}
	return covs;
}

} //namespace
