/*
 * cloudUtils: general-purpose functions on PointClouds
 *
 * Evan Herbst
 * 2 / 18 / 10
 */

#include <iostream>
#include <sensor_msgs/point_cloud_conversion.h>
#include <pcl/ros/conversions.h>
#include "pcl_rgbd/pointTypes.h"
#include <boost/make_shared.hpp>
#include <sensor_msgs/image_encodings.h>

namespace rgbd
{
using std::cout;
using std::endl;

template <typename T> struct pointCloud2datatypeCode {};
template <> struct pointCloud2datatypeCode<int8_t> {static const uint8_t value = sensor_msgs::PointField::INT8;};
template <> struct pointCloud2datatypeCode<uint8_t> {static const uint8_t value = sensor_msgs::PointField::UINT8;};
template <> struct pointCloud2datatypeCode<int16_t> {static const uint8_t value = sensor_msgs::PointField::INT16;};
template <> struct pointCloud2datatypeCode<uint16_t> {static const uint8_t value = sensor_msgs::PointField::UINT16;};
template <> struct pointCloud2datatypeCode<int32_t> {static const uint8_t value = sensor_msgs::PointField::INT32;};
template <> struct pointCloud2datatypeCode<uint32_t> {static const uint8_t value = sensor_msgs::PointField::UINT32;};
template <> struct pointCloud2datatypeCode<float> {static const uint8_t value = sensor_msgs::PointField::FLOAT32;};
template <> struct pointCloud2datatypeCode<double> {static const uint8_t value = sensor_msgs::PointField::FLOAT64;};

template <typename PointT, typename PointVectorT>
void convertPointCloudToVector (const pcl::PointCloud<PointT>& point_cloud, PointVectorT& point_vector)
{
	point_vector.resize(point_cloud.points.size());
	for(unsigned int i=0; i<point_cloud.points.size(); i++) point_vector[i] = ptX2eigen<typename PointVectorT::value_type>(point_cloud.points[i]);
}

template <typename PointT, typename PointVectorT>
void convertVectorToPointCloud (const PointVectorT& point_vector, pcl::PointCloud<PointT>& point_cloud)
{
	point_cloud.points.resize(point_vector.size());
	for(unsigned int i=0; i<point_vector.size(); i++) eigen2ptX(point_cloud.points[i], point_vector[i]);
	point_cloud.width = point_cloud.points.size();
	point_cloud.height = 1;
}

template <typename PointT>
void transform_point_cloud(rgbd::eigen::Affine3f t, const pcl::PointCloud<PointT>& point_cloud, pcl::PointCloud<PointT>& result, const bool transformNormals)
{
	result = point_cloud;
	for(unsigned int i=0; i < point_cloud.points.size(); i++) xformPtPosSafe(result.points[i], t, point_cloud.points[i]);

	if(transformNormals) for(unsigned int i=0; i < point_cloud.points.size(); i++) xformPtNormalSafe(result.points[i], t, point_cloud.points[i]);
}

template <typename PointT>
void transform_point_cloud(rgbd::eigen::Affine3f t, const pcl::PointCloud<PointT>& point_cloud, pcl::PointCloud<PointT>& result)
{
	result = point_cloud;
	for(unsigned int i=0; i < point_cloud.points.size(); i++) xformPtPosSafe(result.points[i], t, point_cloud.points[i]);
}

template <typename PointT>
void transform_point_cloud_in_place(rgbd::eigen::Affine3f t, pcl::PointCloud<PointT>& point_cloud, const bool transformNormals)
{
	for(unsigned int i=0; i < point_cloud.points.size(); i++) xformPtPos(point_cloud.points[i], t, point_cloud.points[i]);

	if(transformNormals) for(unsigned int i=0; i < point_cloud.points.size(); i++) xformPtNormal(point_cloud.points[i], t, point_cloud.points[i]);
}

template <typename PointT>
void transform_point_cloud_in_place(rgbd::eigen::Affine3f t, pcl::PointCloud<PointT>& point_cloud)
{
	for(unsigned int i=0; i < point_cloud.points.size(); i++) xformPtPos(point_cloud.points[i], t, point_cloud.points[i]);
}

template <typename PointT>
void mergeInPlace(pcl::PointCloud<PointT> const& src, pcl::PointCloud<PointT>& dest)
{
	dest.points.insert(dest.points.end(), src.points.begin(), src.points.end());
	dest.width = dest.points.size();
	dest.height = 1;
}

template <typename SrcT, typename DestT>
void convertFieldType(const std::string& fieldName, const sensor_msgs::PointCloud2& src, sensor_msgs::PointCloud2& dest)
{
	const int fieldIndex = sensor_msgs::getPointCloud2FieldIndex(src, fieldName);
	assert(fieldIndex >= 0);
	assert(src.fields[fieldIndex].datatype == pointCloud2datatypeCode<SrcT>::value);

	dest.header = src.header;
	dest.width = src.width;
	dest.height = src.height;
	dest.is_bigendian = src.is_bigendian;
	dest.is_dense = src.is_dense;
	dest.fields = src.fields;
	dest.fields[fieldIndex].datatype = pointCloud2datatypeCode<DestT>::value;
	const int offsetDelta = sizeof(DestT) - sizeof(SrcT);
	for(unsigned int i = fieldIndex + 1; i < dest.fields.size(); i++) dest.fields[i].offset += offsetDelta;

	dest.point_step = src.point_step + offsetDelta;
	dest.row_step = dest.width * dest.point_step;
	dest.data.resize(dest.row_step * dest.height);
	const uint8_t* srcRow, *srcPt;
	uint8_t* destRow, *destPt;
	for(unsigned int i = 0; i < dest.height; i++)
	{
		srcRow = &src.data[i * src.row_step];
		destRow = &dest.data[i * dest.row_step];
		for(unsigned int j = 0; j < dest.width; j++)
		{
			srcPt = srcRow + j * src.point_step;
			destPt = destRow + j * dest.point_step;
			if(fieldIndex > 0) memcpy(destPt, srcPt, src.fields[fieldIndex].offset);
			const SrcT s = *reinterpret_cast<const SrcT*>(srcPt + src.fields[fieldIndex].offset);
			DestT& d = *reinterpret_cast<DestT*>(destPt + dest.fields[fieldIndex].offset);
			d = static_cast<DestT>(s);
			if(fieldIndex < (int)src.fields.size() - 1) memcpy(destPt + dest.fields[fieldIndex + 1].offset, srcPt + src.fields[fieldIndex + 1].offset, src.point_step - src.fields[fieldIndex + 1].offset);
		}
	}
}

template <typename PointT>
rgbd_msgs::DepthMapPtr convertKinectCloudToDepthMap(const pcl::PointCloud<PointT> & cloud, float focal_distance)
{
	assert(isOrganizedCloud(cloud));
	pcl::PointCloud<PointT> & cloud_without_const = const_cast<pcl::PointCloud<PointT> &> (cloud);

	// Peter:
	// is_dense now means "no NaN values"
	// we should use isCloudOrganized everywhere (height > 1)
	// Not sure whether this matters in this function...can't see why it would
	//cloud_without_const.is_dense = true; // This is a weird problem. Point cloud has is_dense=false eventhough it is dense.

	rgbd_msgs::DepthMapPtr result = boost::make_shared<rgbd_msgs::DepthMap>();

	result->header = cloud.header;
	result->no_sample_value = -1;
	result->shadow_value = -2;
	result->height = cloud.height;
	result->width = cloud.width;
	result->focal_distance = focal_distance;
	result->float_data.resize(result->width * result->height);

	for (unsigned int row = 0; row < cloud.height; row++) {
		for (unsigned int col = 0; col < cloud.width; col++) {
			float z = cloud_without_const.at(col, row).z;
			float& to_write = result->float_data[row * result->width + col];
			if (std::isnan(z)) {
				to_write = result->no_sample_value;
			}
			else {
				to_write = z;
			}
		}
	}

	return result;
}

template <typename PointT>
sensor_msgs::ImagePtr convertKinectCloudToImage(const pcl::PointCloud<PointT> & cloud)
{
	assert(isOrganizedCloud(cloud));
	pcl::PointCloud<PointT> & cloud_without_const = const_cast<pcl::PointCloud<PointT> &> (cloud);

	// use opencv to create the image
	cv::Mat result_cv(cloud.height, cloud.width, CV_8UC3);

	for (unsigned int row = 0; row < cloud.height; row++) {
		for (unsigned int col = 0; col < cloud.width; col++) {
			float rgb = cloud_without_const.at(col, row).rgb;
			boost::array<unsigned char, 3> rgb_unpacked = rgbd::unpackRGB<unsigned char>(rgb);
			// put in in bgr8 (what opencv likes)
			result_cv.at<cv::Vec3b>(row, col)[0] = rgb_unpacked[2];
			result_cv.at<cv::Vec3b>(row, col)[1] = rgb_unpacked[1];
			result_cv.at<cv::Vec3b>(row, col)[2] = rgb_unpacked[0];
		}
	}

	cv_bridge::CvImage cv_bridge_image;
	cv_bridge_image.image = result_cv;
	cv_bridge_image.encoding = sensor_msgs::image_encodings::BGR8;
	cv_bridge_image.header = cloud.header;

	return cv_bridge_image.toImageMsg();
}

template <typename PointT>
sensor_msgs::ImagePtr convertKinectCloudToDepthImage(const pcl::PointCloud<PointT> & cloud)
{
	assert(isOrganizedCloud(cloud));
	pcl::PointCloud<PointT> & cloud_without_const = const_cast<pcl::PointCloud<PointT> &> (cloud);

	// use opencv to create the image
	cv::Mat result_cv(cloud.height, cloud.width, CV_32FC1);

	for (unsigned int row = 0; row < cloud.height; row++) {
		for (unsigned int col = 0; col < cloud.width; col++) {
			result_cv.at<float>(row,col) = cloud_without_const.at(col, row).z;
		}
	}

	cv_bridge::CvImage cv_bridge_image;
	cv_bridge_image.image = result_cv;
	cv_bridge_image.encoding = sensor_msgs::image_encodings::TYPE_32FC1;
	cv_bridge_image.header = cloud.header;

	return cv_bridge_image.toImageMsg();
}

template <typename PointT>
typename pcl::PointCloud<PointT>::Ptr convertKinectCloudType(
		const sensor_msgs::PointCloud2 & kinect_cloud)
{
	pcl::PointCloud<pcl::PointXYZRGB>::Ptr cloud_xyzrgb(new pcl::PointCloud<pcl::PointXYZRGB>());
	pcl::fromROSMsg(kinect_cloud,*cloud_xyzrgb);
	return rgbd::convertKinectCloudType<PointT>(*cloud_xyzrgb);
}

template <typename PointT>
typename pcl::PointCloud<PointT>::Ptr convertKinectCloudType(
		const pcl::PointCloud<pcl::PointXYZRGB> & kinect_cloud)
{
	assert(isOrganizedCloud(kinect_cloud));
	typename pcl::PointCloud<PointT>::Ptr out_cloud(new pcl::PointCloud<PointT>());

	PointT pt;
	for(unsigned int x=0; x<kinect_cloud.width; x++){
		for(unsigned int y=0; y<kinect_cloud.height; y++){
			unsigned int i = x+y*kinect_cloud.width;
			pcl::PointXYZRGB const& kinect_pt = kinect_cloud.points[i];
			if(pcl::hasValidXYZ(kinect_pt)){
				pt.x = kinect_pt.x; pt.y = kinect_pt.y; pt.z = kinect_pt.z; pt.rgb = kinect_pt.rgb;
				pt.imgX = x; pt.imgY = y;
				out_cloud->points.push_back(pt);
			}
		}
	}

	out_cloud->header = kinect_cloud.header;
	out_cloud->is_dense = false;
	out_cloud->width = out_cloud->points.size();
	out_cloud->height = 1;

	return out_cloud;
}

template <typename PointT>
bool isOrganizedCloud(const pcl::PointCloud<PointT> & cloud)
{
	// whatever the currently popular check for organized is...
	return (cloud.height > 1);
}

template <typename PointT>
void getValidIndicesForOrganizedCloud(
		const pcl::PointCloud<PointT> &organized_cloud,
		float max_depth,
		std::vector<unsigned int> &valid_indices)
{
	valid_indices.clear();
	valid_indices.reserve(organized_cloud.points.size());
	for (unsigned int i = 0; i < organized_cloud.points.size(); i++) {
		if (!isnan(organized_cloud.points[i].z) && organized_cloud.points[i].z > 0 && (max_depth <= 0 || organized_cloud.points[i].z <= max_depth)) {
			valid_indices.push_back(i);
		}
	}
}

template <typename PointT>
void getValidIndicesXYForOrganizedCloud(
		const pcl::PointCloud<PointT> &organized_cloud,
		float max_depth,
		std::vector<std::pair<unsigned int, unsigned int> > &valid_indices)
{
	valid_indices.clear();
	valid_indices.reserve(organized_cloud.points.size());
	for (unsigned int x = 0; x < organized_cloud.width; x++) {
			for (unsigned int y = 0; y < organized_cloud.height; y++) {
				if (!isnan(organized_cloud.at(x,y).z) && organized_cloud.at(x,y).z > 0 && (max_depth <= 0 || organized_cloud.at(x,y).z <= max_depth)) {
					valid_indices.push_back(std::make_pair(x,y));
				}
			}
	}
}

/*
 * set all points to the given color
 */
template <typename PointT>
void recolor(pcl::PointCloud<PointT>& cloud, const rgbd::eigen::Vector3f& rgb)
{
	const float _rgb = rgbd::packRGB(rgb);
	for(unsigned int i = 0; i < cloud.points.size(); i++) cloud.points[i].rgb = _rgb;
}

template <typename PointT>
void getMeshForOrganizedCloud(
		const pcl::PointCloud<PointT> &organized_cloud,
		float max_edge_length,
		std::vector<unsigned int> &triangles)
{
	assert(isOrganizedCloud(organized_cloud));

	triangles.clear();
	triangles.reserve(organized_cloud.points.size() * 2);

	float max_edge_length_squared = max_edge_length * max_edge_length;

	// note that "-1" means we don't need to bounds check in the loop (always look down and right)
	for (int x = 0; x < (int)organized_cloud.width-1; x++) {
		for (int y = 0; y < (int)organized_cloud.height-1; y++) {
			int pi00 = y * organized_cloud.width + x;
			const PointT & p00 = organized_cloud.points[pi00];
			int pi10 = y * organized_cloud.width + (x+1);
			const PointT & p10 = organized_cloud.points[pi10];
			int pi01 = (y+1) * organized_cloud.width + x;
			const PointT & p01 = organized_cloud.points[pi01];
			int pi11 = (y+1) * organized_cloud.width + (x+1);
			const PointT & p11 = organized_cloud.points[pi11];
			// first always go UL to LR (adapt later...)
			// right now skip faces when any point is missing...could sometimes add 1 of the 2 triangles
			// also...I think we could eliminate this entirely and use the later length checks (which must evaluate to true)
			if (isnan(p00.z) || isnan(p01.z) || isnan(p10.z) || isnan(p11.z)) continue;

			// check edges of "upper right" triangle
			if ( 	(p00.getVector3fMap() - p11.getVector3fMap()).squaredNorm() <= max_edge_length_squared ||
					(p11.getVector3fMap() - p10.getVector3fMap()).squaredNorm() <= max_edge_length_squared ||
					(p10.getVector3fMap() - p00.getVector3fMap()).squaredNorm() <= max_edge_length_squared) {
				triangles.push_back(pi00);
				triangles.push_back(pi11);
				triangles.push_back(pi10);
			}

			// check edges of "lower right" triangle
			if ( 	(p00.getVector3fMap() - p01.getVector3fMap()).squaredNorm() <= max_edge_length_squared ||
					(p01.getVector3fMap() - p11.getVector3fMap()).squaredNorm() <= max_edge_length_squared ||
					(p11.getVector3fMap() - p00.getVector3fMap()).squaredNorm() <= max_edge_length_squared) {
				triangles.push_back(pi00);
				triangles.push_back(pi01);
				triangles.push_back(pi11);
			}
		}
	}
}

} //namespace
