/*
 * depth_to_cloud_lib
 */

#include <exception>
#include <cv_bridge/cv_bridge.h>
#include "rgbd_util/eigen/Geometry"
#include "rgbd_util/imageSampling.h"
#include "rgbd_util/primesensorUtils.h"
#include "rgbd_depthmaps/decompressDepth.h"
#include "pcl_rgbd/pointTypes.h"

/*
 * throw on any error
 */
template <typename PointT>
void depth_to_cloud(
		const rgbd::depthmapPtrs& dmptrs,
		const rgbd_msgs::DepthMap& depth_msg,
		bool include_xy_channels,
		bool publish_all_points,
		pcl::PointCloud<PointT>& cloud,
		const rgbd::CameraParams& depthCamParams)
{
	cloud.header = depth_msg.header;
	// need to clear it out
	cloud.points.clear();

	//presize the points vector
	unsigned int numPoints;
	if(publish_all_points) numPoints = depth_msg.width * depth_msg.height;
	else
	{
		numPoints = 0;
		for(int row=0, i = 0; row < depth_msg.height; row++)
		  for(int col=0; col < depth_msg.width; col++, i++)
		  {
				const float depth = dmptrs.dm[i];
				if(depth > 0) numPoints++;
		  }
	}
	cloud.points.resize(numPoints);

	for(int row=0, i = 0, j = 0; row < depth_msg.height; row++) {
	  for(int col=0; col < depth_msg.width; col++, j++) {
			const float depth = dmptrs.dm[j];
			if(depth > 0 || publish_all_points)
			{
				rgbd::eigen::Vector3f pos;
				pos.x() = (((float)col - depthCamParams.centerX) * depth / depthCamParams.focalLength);
				pos.y() = (((float)row - depthCamParams.centerY) * depth / depthCamParams.focalLength);
				pos.z() = depth;

				PointT& p = cloud.points[i];
				p.x = pos.x();
				p.y = pos.y();
				p.z = pos.z();
				p.rgb = 0; //black
				if(include_xy_channels) rgbd::setImgCoords<PointT>(p, col, row);

				i++;
			}
	  }
	}

	if (publish_all_points) {
		cloud.width = depth_msg.width;
		cloud.height = depth_msg.height;
	}
	else {
		cloud.width = cloud.points.size();
		cloud.height = 1;
	}
}

template <typename PointT>
bool depth_to_cloud(
		const rgbd_msgs::DepthMap& depth_msg,
		bool include_xy_channels,
		bool publish_all_points,
		pcl::PointCloud<PointT>& cloud,
		const rgbd::CameraParams& depthCamParams)
{
	rgbd::depthmapPtrs dmptrs;
	rgbd::getUncompressedDepthMapPtrs(depth_msg, dmptrs);

	depth_to_cloud(dmptrs, depth_msg, include_xy_channels, publish_all_points, cloud, depthCamParams);

	return true;
}

template <typename PointT>
bool image_and_depth_to_cloud(
		const sensor_msgs::Image& image_msg,
		const rgbd_msgs::DepthMap& depth_msg,
		const rgbd::eigen::Affine3f& depth2imgXform,
		const rgbd::CameraParams& camParams,
		bool include_xy_channels,
		bool publish_all_points,
		pcl::PointCloud<PointT>& cloud,
		const rgbd::CameraParams& depthCamParams)
{
	rgbd::depthmapPtrs dmptrs;
	rgbd::getUncompressedDepthMapPtrs(depth_msg, dmptrs);

	//create uncolored cloud
	depth_to_cloud(dmptrs, depth_msg, include_xy_channels, publish_all_points, cloud, depthCamParams);

	//add colors
	for(int row=0, i = 0, j = 0; row < depth_msg.height; row++) {
	  for(int col=0; col < depth_msg.width; col++, j++) {
			const float depth = dmptrs.dm[j];
			if(depth > 0 || publish_all_points)
			{
				PointT& p = cloud.points[i];

				const rgbd::eigen::Vector3f pos(p.x, p.y, p.z);
				const rgbd::eigen::Vector3f posInCam = depth2imgXform * pos;
				const float imgX = posInCam.z() <= 0 ? 0 : (camParams.centerX + camParams.focalLength * posInCam.x() / posInCam.z()),
						imgY = posInCam.z() <= 0 ? 0 : (camParams.centerY + camParams.focalLength * posInCam.y() / posInCam.z());
				const rgbd::eigen::Vector3f imgSample = (imgX >= 0 && imgX < image_msg.width - 1 && imgY >= 0 && imgY < image_msg.height - 1)
																		? interpolateImagePixel(image_msg, imgX, imgY)
																		: rgbd::eigen::Vector3f::Zero();

				p.rgb = rgbd::packRGB(imgSample);
				i++;
			}
	  }
	}

	return true;
}

template <typename PointT>
bool image_and_depth_to_cloud(
		const sensor_msgs::Image& image_msg,
		const rgbd_msgs::DepthMap& depth_msg,
		bool include_xy_channels,
		bool publish_all_points,
		pcl::PointCloud<PointT>& cloud,
		const rgbd::CameraParams& depthCamParams)
{
	if(image_msg.width != depth_msg.width || image_msg.height != depth_msg.height)
		throw std::invalid_argument("image and depth must have same size");

	rgbd::depthmapPtrs dmptrs;
	rgbd::getUncompressedDepthMapPtrs(depth_msg, dmptrs);

	//create uncolored cloud
	depth_to_cloud(dmptrs, depth_msg, include_xy_channels, publish_all_points, cloud, depthCamParams);

	//add colors
	for(int row=0, i = 0, j = 0; row < depth_msg.height; row++) {
	  for(int col=0; col < depth_msg.width; col++, j++) {
			const float depth = dmptrs.dm[j];
			if(depth > 0 || publish_all_points)
			{
				PointT& p = cloud.points[i];

				const rgbd::eigen::Vector3f pos(p.x, p.y, p.z);
				const boost::array<uint8_t, 3> pixval = sampleAtPixel(image_msg, col, row);

				p.rgb = rgbd::packRGB(pixval[0], pixval[1], pixval[2]);
				i++;
			}
	  }
	}

	return true;
}

template <typename PointT>
bool image_and_depth_to_cloud(
		const cv::Mat& image,
		const rgbd_msgs::DepthMap& depth_msg,
		bool include_xy_channels,
		bool publish_all_points,
		pcl::PointCloud<PointT>& cloud,
		const rgbd::CameraParams& depthCamParams)
{
	if(image.cols != depth_msg.width || image.rows != depth_msg.height)
		throw std::invalid_argument("image and depth must have same size");

	rgbd::depthmapPtrs dmptrs;
	rgbd::getUncompressedDepthMapPtrs(depth_msg, dmptrs);

	//create uncolored cloud
	depth_to_cloud(dmptrs, depth_msg, include_xy_channels, publish_all_points, cloud, depthCamParams);

	//add colors
	for(int row=0, i = 0, j = 0; row < depth_msg.height; row++) {
	  for(int col=0; col < depth_msg.width; col++, j++) {
			const float depth = dmptrs.dm[j];
			if(depth > 0 || publish_all_points)
			{
				PointT& p = cloud.points[i];

				const rgbd::eigen::Vector3f pos(p.x, p.y, p.z);
				const cv::Vec3b pixval = image.at<cv::Vec3b>(row, col);

				p.rgb = rgbd::packRGB(pixval[2], pixval[1], pixval[0]);
				i++;
			}
	  }
	}

	return true;
}
