/*
 * istreamReader: generically read from a stream of text or binary format
 *
 * Evan Herbst
 * 2 / 9 / 10
 */

#include <array>
#include <boost/detail/endian.hpp>

namespace ply
{

template <typename T>
void istreamReader::readText(T& t) const
{
	in >> t;
}

template <typename T>
void istreamReader::readLittleEndian(T& t) const
{
	in.read(reinterpret_cast<char*>(&t), sizeof(T));
#if defined(BOOST_LITTLE_ENDIAN)
#elif defined(BOOST_BIG_ENDIAN)
	ply::detail::swapBytes<sizeof(T)>(reinterpret_cast<char*>(&t));
#else
#error "can't determine system endianness"
#endif
}

template <typename T>
void istreamReader::readBigEndian(T& t) const
{
	in.read(reinterpret_cast<char*>(&t), sizeof(T));
#if defined(BOOST_LITTLE_ENDIAN)
	ply::detail::swapBytes<sizeof(T)>(reinterpret_cast<char*>(&t));
#elif defined(BOOST_BIG_ENDIAN)
#else
#error "can't determine system endianness"
#endif
}

template <typename T>
void istreamReader::readListText(T& t) const
{
	readText(t.first);
	t.second.resize(t.first);
	for(auto i = t.second.begin(); i != t.second.end(); i++) readText(*i);
}

template <typename T>
void istreamReader::readListLittleEndian(T& t) const
{
	readLittleEndian(t.first);
	t.second.resize(t.first);
	for(auto i = t.second.begin(); i != t.second.end(); i++) readLittleEndian(*i);
}

template <typename T>
void istreamReader::readListBigEndian(T& t) const
{
	readBigEndian(t.first);
	t.second.resize(t.first);
	for(auto i = t.second.begin(); i != t.second.end(); i++) readBigEndian(*i);
}

/*
 * auxiliary to operator ()
 */
template <typename T, const bool IsListType>
class istreamReaderAux;
/*
 * type isn't a vector: read one element
 */
template <typename T>
struct istreamReaderAux<T, false>
{
	void read(T& t, const istreamReader& reader)
	{
		switch(reader.getFormat())
		{
			case FMT_ASCII: reader.readText(t); break;
			case FMT_LITTLE_ENDIAN: reader.readLittleEndian(t); break;
			case FMT_BIG_ENDIAN: reader.readBigEndian(t); break;
		}
	}
};
/*
 * type is a vector: read length, then elements
 */
template <typename T>
struct istreamReaderAux<T, true>
{
	void read(T& t, const istreamReader& reader)
	{
		switch(reader.getFormat())
		{
			case FMT_ASCII: reader.readListText(t); break;
			case FMT_LITTLE_ENDIAN: reader.readListLittleEndian(t); break;
			case FMT_BIG_ENDIAN: reader.readListBigEndian(t); break;
		}
	}
};

/*
 * read into t
 */
template <typename T, const bool IsListType>
void istreamReader::operator () (T& t) const
{
	//need indirection through an aux class to get partial specialization
	istreamReaderAux<T, IsListType> reader;
	reader.read(t, *this);
}

} //namespace
