/*
 * ostreamWriter: generically write to a stream of text or binary format
 *
 * Evan Herbst
 * 6 / 26 / 10
 */

#include <array>
#include <boost/detail/endian.hpp>

namespace ply
{

template <typename T>
void ostreamWriter::writeText(const T& t) const
{
	out << t;
}

template <typename T>
void ostreamWriter::writeLittleEndian(const T& t) const
{
	std::array<char, sizeof(T)> buf;
	*reinterpret_cast<T*>(buf.data()) = t;
#if defined(BOOST_LITTLE_ENDIAN)
#elif defined(BOOST_BIG_ENDIAN)
	ply::detail::swapBytes<sizeof(T)>(buf.data());
#else
#error "can't determine system endianness"
#endif
	out.write(buf.data(), buf.size());
}

template <typename T>
void ostreamWriter::writeBigEndian(const T& t) const
{
	std::array<char, sizeof(T)> buf;
	*reinterpret_cast<T*>(buf.data()) = t;
#if defined(BOOST_LITTLE_ENDIAN)
	ply::detail::swapBytes<sizeof(T)>(buf.data());
#elif defined(BOOST_BIG_ENDIAN)
#else
#error "can't determine system endianness"
#endif
	out.write(buf.data(), buf.size());
}

template <typename T>
void ostreamWriter::writeListText(const T& t) const
{
	writeText(t.first);
	for(auto i = t.second.begin(); i != t.second.end(); i++) writeText(*i);
}

template <typename T>
void ostreamWriter::writeListLittleEndian(const T& t) const
{
	writeLittleEndian(t.first);
	for(auto i = t.second.begin(); i != t.second.end(); i++) writeLittleEndian(*i);
}

template <typename T>
void ostreamWriter::writeListBigEndian(const T& t) const
{
	writeBigEndian(t.first);
	for(auto i = t.second.begin(); i != t.second.end(); i++) writeBigEndian(*i);
}

/*
 * auxiliary to operator ()
 */
template <typename T, const bool IsListType>
class ostreamWriterAux;
/*
 * type isn't a vector: write one element
 */
template <typename T>
struct ostreamWriterAux<T, false>
{
	void write(const T& t, const ostreamWriter& writer)
	{
		switch(writer.getFormat())
		{
			case FMT_ASCII: writer.writeText(t); break;
			case FMT_LITTLE_ENDIAN: writer.writeLittleEndian(t); break;
			case FMT_BIG_ENDIAN: writer.writeBigEndian(t); break;
		}
	}
};
/*
 * type is a vector: write length, then elements
 */
template <typename T>
struct ostreamWriterAux<T, true>
{
	void write(const T& t, const ostreamWriter& writer)
	{
		switch(writer.getFormat())
		{
			case FMT_ASCII: writer.writeListText(t); break;
			case FMT_LITTLE_ENDIAN: writer.writeListLittleEndian(t); break;
			case FMT_BIG_ENDIAN: writer.writeListBigEndian(t); break;
		}
	}
};

/*
 * write t
 */
template <typename T, const bool IsListType>
void ostreamWriter::operator () (const T& t) const
{
	//need indirection through an aux class to get partial specialization
	ostreamWriterAux<T, IsListType> writer;
	writer.write(t, *this);
}

} //namespace
