/*
 * ply: read & write PLY files in a completely runtime manner (not type-safe, and no schema checking)
 *
 * Evan Herbst
 * 2 / 9 / 10
 */

#include <stdexcept>
#include <boost/function.hpp>
#include <boost/ref.hpp>
#include <boost/fusion/include/for_each.hpp>

namespace ply
{

/*
 * read numValues element values of the requested type; return them as tuples
 *
 * wherever espec has a property list, FusionVectorT should have a std::pair<length type, std::vector<value type>>
 *
 * throw on any parse error
 */
template <typename FusionVectorT>
std::vector<FusionVectorT> readValues(std::istream& in, const ply::format fmt, const element& espec)
{
	std::vector<FusionVectorT> values(espec.count);
	istreamReader reader(fmt, in);
	for(unsigned int i = 0; i < espec.count; i++) fusion::for_each(values[i], reader);
	return values;
}

/*
 * auxiliary to readElements(): read a SrcT and store it to a DestT
 */
template <typename SrcT>
struct readOneFrom
{
	template <typename DestT>
	void operator () (istreamReader& reader, DestT& v)
	{
		SrcT s;
		reader(s);
		v = s;
	}
};

/*
 * read e.count values of the requested number of properties, casting every value to ValueT
 *
 * the returned array has one element per row
 *
 * throw on any parse error
 */
template <typename ValueT>
boost::multi_array<ValueT, 2> readElements(std::istream& in, const ply::format fmt, const element& e)
{
	boost::multi_array<ValueT, 2> values(boost::extents[e.count][e.props.size()]);
	istreamReader reader(fmt, in);
	std::vector<boost::function<void (istreamReader&, ValueT&)> > readFuncs(e.props.size());
	for(unsigned int j = 0; j < e.props.size(); j++)
		if(e.props[j].type == "float") readFuncs[j] = readOneFrom<float>();
		else if(e.props[j].type == "uchar") readFuncs[j] = readOneFrom<unsigned char>();
		else throw std::runtime_error("unknown datatype string");
	for(unsigned int i = 0; i < e.count; i++)
		for(unsigned int j = 0; j < e.props.size(); j++)
			readFuncs[j](reader, values[i][j]);
	return values;
}

/*
 * write all the elements in v
 *
 * wherever an element of FusionVectorT has the form pair<X, vector<Y>>, we'll write a property list with count type X and element type Y
 *
 * throw on any error
 */
template <typename FusionVectorT>
void writeValues(std::ostream& out, const ply::format fmt, const std::vector<FusionVectorT>& v)
{
	ostreamWriter writer(fmt, out);
	for(unsigned int i = 0; i < v.size(); i++) fusion::for_each(v[i], writer);
}

} //namespace
