/*
 * icpCovariance: functions to compute covariance of ICP transform (eg for toro)
 *
 * Evan Herbst
 * 6 / 22 / 10
 */

#include <unistd.h> //getcwd()
#include <cassert>
#include <cmath> //rint()
#include <string>
#include <fstream>
#include <algorithm> //lower_bound()
#include <iterator> //distance()
#include <stdexcept>
#include "rgbd_util/eigen/LU" //inverse()
#include "rgbd_util/eigen/Sparse"
#include "rgbd_util/mathUtils.h" //sqr(), dabs()
#include "pcl_rgbd/cloudNormals.h"
#include "primesensor_tools/primeSense_quantization_v5.h"

namespace registration
{
using std::vector;
using std::cout;
using std::endl;
using rgbd::eigen::Vector3d;
using rgbd::eigen::Matrix3d;
using rgbd::eigen::MatrixXd;
using rgbd::eigen::Matrix;
using rgbd::eigen::SparseMatrix;

namespace aux
{

template <typename MatrixT>
void writeMATLAB(const MatrixT& m, const std::string& filepath)
{
	std::ofstream outfile(filepath.c_str());
	assert(outfile);
	for(unsigned int i = 0; i < m.rows(); i++)
	{
		for(unsigned int j = 0; j < m.cols(); j++) outfile << m.coeff(i, j) << ' ';
		outfile << std::endl;
	}
}

template <typename MatrixT>
void writeCovsMATLAB(const MatrixT& m, const std::string& filepath)
{
	char cwd[100000];
	assert(getcwd(cwd, sizeof(cwd)));

	cout << "writing " << filepath << " from " << cwd << endl;
	std::ofstream outfile(filepath.c_str());
	assert(outfile);
	for(const double* i = m._valuePtr(); i != m._valuePtr() + 3 * m.rows(); i += 3)
	{
		for(unsigned int j = 0; j < 3; j++) outfile << *i++ << ' ';
		outfile << std::endl;
	}
}

template <typename MatrixT>
bool anyNaN(const MatrixT& m)
{
	for(unsigned int i = 0; i < m.rows(); i++)
		for(unsigned int j = 0; j < m.cols(); j++)
			if(isnan(m(i, j)))
				return true;
	return false;
}

template <typename MatrixT>
void testNaN(const MatrixT& m)
{
	for(unsigned int i = 0; i < m.rows(); i++)
		for(unsigned int j = 0; j < m.cols(); j++)
			if(isnan(m(i, j)))
				assert(false);
}

} //namespace

template <typename PointT>
SparseMatrix<double, rgbd::eigen::RowMajor> getPointCovariance(const pcl::PointCloud<PointT>& cloud)
{
#if 1
	/*
	 * camera params (TODO parameterize)
	 */
	const double fovX = 50 * M_PI / 180, fovY = 50 * M_PI / 180; //fov is 70 degrees measured diagonally
	const unsigned int xResolution = 640, yResolution = 480;
	const double dthetaX = fovX / xResolution, dthetaY = fovY / yResolution;

	/*
	 * covariance scaling (TODO parameterize)
	 */
	const double xySigmaInPixels = 2;
	const double zSigmaInDepthBins = 1;

	SparseMatrix<double, rgbd::eigen::RowMajor> covX((signed)(3 * cloud.points.size()), (signed)(3 * cloud.points.size()));
	covX.startFill(9 * cloud.points.size());
	const vector<Matrix3d> indepPtCovs = getIndependentPointCovariances(cloud, dthetaX, dthetaY, xySigmaInPixels, zSigmaInDepthBins);
	for(unsigned int i = 0; i < cloud.points.size(); i++)
	{
		aux::testNaN(indepPtCovs[i]);
		for(unsigned int j = 0; j < 3; j++)
			for(unsigned int k = 0; k < 3; k++)
				covX.fill(3 * i + j, 3 * i + k) = indepPtCovs[i](j, k);
	}
	covX.endFill();
#else
	/*
	 * identity (useful for comparing output to Xiaofeng's octave script)
	 */
	SparseMatrix<double, rgbd::eigen::RowMajor> covX((signed)(3 * cloud.points.size()), (signed)(3 * cloud.points.size()));
	covX.startFill(covX.rows());
	for(unsigned int i = 0; i < (unsigned)covX.rows(); i++) covX.fill(i, i) = 1;
	covX.endFill();
#endif

	return covX;
}

/*
 * return an empirical covariance matrix for the three translation components and roll, pitch, yaw as used by TORO
 *
 * pre: each ith point in alignee corresponds to the ith point in alignand; alignand has normals
 */
template <typename PointT, typename PointT2>
Matrix<double, 6, 6> getPointToPlaneICPXformCovarianceXlateRPY(const pcl::PointCloud<PointT>& alignee, const pcl::PointCloud<PointT2>& alignand)
{
	/*
	 * see A. Censi, "An Accurate Closed-Form Estimate of ICP's Covariance", ICRA07, especially equation 6, for math explanation
	 */

	assert(alignee.points.size() == alignand.points.size());
	//TODO assert alignand has normals

	/*
	 * for making things well-conditioned by removing unfortunately placed zeros
	 */
	static const double EPSILON = 1e-7;

	/*
	 * covariance of pt positions
	 */
	const SparseMatrix<double, rgbd::eigen::RowMajor> covX = getPointCovariance(alignee);

	/*
	 * second derivatives of J
	 */
	/*
	 * express Euler angles as a rotation matrix: see the xyz-order matrix in the wikipedia "euler angles" "table of matrices"
	 *
	 * (si, ci the sine/cosine of Euler angle i)
	 * r00 = c1*c2;
	 * r01 = c2*s0*s1 - c0*s2;
	 * r02 = c0*c2*s1 + s0*s2;
	 * r10 = c1*s2;
	 * r11 = c0*c2 + s0*s1*s2;
	 * r12 = c0*s1*s2 - c2*s0;
	 * r20 = -s1;
	 * r21 = c1*s0;
	 * r22 = c0*c1;
	 */
	/*
	 * (easy here since we're evaluating at the zero transformation)
	 */
	const double s0 = 0, s1 = 0, s2 = 0, c0 = 1, c1 = 1, c2 = 1; //sines/cosines
	Matrix3d R; R.setIdentity();
	const Vector3d t(0, 0, 0); //translation

	/*
	 * rows and cols of R, named using the conventions I use below
	 */
	const Vector3d r0_ = R.row(0), r1_ = R.row(1), r2_ = R.row(2), r_0 = R.col(0), r_1 = R.col(1), r_2 = R.col(2);

	MatrixXd d2Jdt2(6, 6);
	d2Jdt2.setZero();
	for(unsigned int i = 0; i < alignee.points.size(); i++)
	{
		const Vector3d x(alignee.points[i].x, alignee.points[i].y, alignee.points[i].z); //alignee pt
		const Vector3d y(alignand.points[i].x, alignand.points[i].y, alignand.points[i].z); //alignand pt
		const Vector3d n(alignand.points[i].normal[0], alignand.points[i].normal[1], alignand.points[i].normal[2]); //alignand normal

		/*
		 * d/dt dJ/dt = 2 n n^T
		 */
		d2Jdt2.block<3, 3>(0, 0) += MatrixXd(2 * n * n.transpose());

		/*
		 * for matrix m, m0_ the 0th row, m_0 the 0th col
		 *
		 * d/dt dJ/dr =
		 * (
		 *  [
		 *   (r_2*x1 - r_1*x2)' ;
		 *   [(c2*(r2_*x)) (s2*(r2_*x)) -[c1 s0*s1 c0*s1]*x] ;
		 *   [(-r1_*x) (r0_*x) 0]
		 *  ]
		 * *n) * 2n'
		 */
		Matrix3d tmp;
		tmp.row(0) = (r_2 * x[1] - r_1 * x[2]).transpose();
		tmp(1, 0) = c2 * r2_.dot(x);
		tmp(1, 1) = s2 * r2_.dot(x);
		tmp(1, 2) = -Vector3d(c1, s0 * s1, c0 * s1).dot(x);
		tmp(2, 0) = -r1_.dot(x);
		tmp(2, 1) = r0_.dot(x);
		tmp(2, 2) = 0;
		d2Jdt2.block<3, 3>(0, 3) += 2 * (tmp * n) * n.transpose();

		/*
		 * d/dr dJ/dr =
		 * 2*
		 * [
		 * //first row
		 * (n*(r_2*x1-r_1*x2))^2+2*(n*(R*x+t-y))*(n*(-r_1*x1-r_2*x2))
		 * (n0*(c2*(r2_*x))+n1*(s2*(r2_*x))+n2*([-c1 -s1*s0 -c0*s1]*x))*(n*(r_2*x1-r_1*x2))+2*(n*(R*x+t-y))*(n0*(c2*(r22*x1-r21*x2))+n1*(s2*(r22*x1-r21*x2))+n2*(s1*(-c0*x1+s0*x2)))
		 * (n0*(-r1_*x)+n1*(r0_*x))*(n*(r_2*x1-r_1*x2))+2*(n*(R*x+t-y))*(n0*(-r12*x1+r11*x2)+n1*(r02*x1-r01*x2))
		 * //second row
		 * //(n0*(c2*(r2_*x))+n1*(s2*(r2_*x))+n2*(-x0*c1-s1*s0*x1-c0*s1*x2))*(n*(r_2*x1-r_1*x2))+2*(n*(R*x+t-y))*(n0*(c2*(r22*x1-r21*x2))+n1*(s2*(r22*x1-r21*x2))+n2*(s1*(-c0*x1+s0*x2)))
		 * (n0*(c2*(r2_*x))+n1*(s2*(r2_*x))+n2*([-c1 -s1*s0 -c0*s1]*x))^2+2*(n*(R*x+t-y))*(n0*(c2*([-c1 -s0*s1 -c0*s1]*x))+n1*(s2*([-c1 -s0*s1 -c0*s1]*x))+n2*(-r2_*x))
		 * (n0*(-r1_*x)+n1*(r0_*x))*(n0*(c2*(r2_*x))+n1*(s2*(r2_*x))+n2*([-c1 -s1*s0 -c0*s1]*x))+2*(n*(R*x+t-y))*(n0*(s2*(-r2_*x))+n1*(c2*(r2_*x)))
		 * //third row
		 * //(n0*(-r1_*x)+n1*(r0_*x))*(n*(r_2*x1-r_1*x2))+2*(n*(R*x+t-y))*(n0*(-r12*x1+r11*x2)+n1*(r02*x1-r01*x2))
		 * //(n0*(-r1_*x)+n1*(r0_*x))*(n0*(c2*(r2_*x))+n1*(s2*(r2_*x))+n2*(-x0*c1-s1*s0*x1-c0*s1*x2))+2*(n*(R*x+t-y))*(n0*(s2*(-r2_*x))+n1*(c2*(r2_*x)))
		 * (n0*(-r1_*x)+n1*(r0_*x))^2+2*(n*(R*x+t-y))*(n0*(-r0_*x)+n1*(-r1_*x))
		 * ]
		 */
		tmp(0, 0) = sqr(n.dot(r_2*x[1] - r_1*x[2])) + 2 * n.dot(R*x+t-y) * n.dot(-r_1*x[1] - r_2*x[2]);
		tmp(0, 1) = (n[0]*c2*r2_.dot(x) + n[1]*s2*r2_.dot(x) + n[2]*Vector3d(-c1, -s1*s0, -c0*s1).dot(x)) * n.dot(r_2*x[1] - r_1*x[2]) + 2 * n.dot(R*x+t-y) * (n[0]*c2*(R(2, 2)*x[1] - R(2, 1)*x[2]) + n[1]*s2*(R(2, 2)*x[1] - R(2, 1)*x[2]) + n[2]*s1*(-c0*x[1] + s0*x[2]));
		tmp(0, 2) = (-n[0]*r1_.dot(x) + n[1]*r0_.dot(x)) * n.dot(r_2*x[1] - r_1*x[2]) + 2 * n.dot(R*x+t-y) * (n[0] * (-R(1, 2)*x[1] + R(1, 1)*x[2]) + n[1] * (R(0, 2)*x[1] - R(0, 1)*x[2]));
		tmp(1, 0) = tmp(0, 1);
		tmp(1, 1) = sqr(n[0]*c2*r2_.dot(x) + n[1]*s2*r2_.dot(x) + n[2]*Vector3d(-c1, -s1*s0, -c0*s1).dot(x)) + 2 * n.dot(R*x+t-y) * (n[0]*c2*Vector3d(-c1, -s0*s1, -c0*s1).dot(x) + n[1]*s2*Vector3d(-c1, -s0*s1, -c0*s1).dot(x) - n[2]*r2_.dot(x));
		tmp(1, 2) = (-n[0]*r1_.dot(x) + n[1]*r0_.dot(x)) * (n[0]*c2*r2_.dot(x) + n[1]*s2*r2_.dot(x) + n[2]*Vector3d(-c1, -s1*s0, -c0*s1).dot(x)) + 2 * n.dot(R*x+t-y) * (-n[0]*s2*r2_.dot(x) + n[1]*c2*r2_.dot(x));
		tmp(2, 0) = tmp(0, 2);
		tmp(2, 1) = tmp(1, 2);
		tmp(2, 2) = sqr(-n[0]*r1_.dot(x) + n[1]*r0_.dot(x)) + 2 * n.dot(R*x+t-y) * (-n[0]*r0_.dot(x) - n[1]*r1_.dot(x));
		d2Jdt2.block<3, 3>(3, 3) += 2 * tmp;
	}
	d2Jdt2.block<3, 3>(3, 0) = d2Jdt2.block<3, 3>(0, 3).transpose();
	aux::testNaN(d2Jdt2);
//	aux::writeMATLAB(d2Jdt2, "d2jdt2.dat");
	/*
	 * ensure it's invertible
	 */
	for(unsigned int i = 0; i < 6; i++)
		if(dabs(d2Jdt2(i, i)) < EPSILON)
			d2Jdt2(i, i) = d2Jdt2(i, i) > 0 ? EPSILON : -EPSILON;

	/*
	 * d/dt dJ/dx for one x =
	 * 2*
	 * [
	 * //first row
	 * n0*(n*r_0)
	 * n1*(n*r_0)
	 * n2*(n*r_0)
	 * (n*(r_2*x1 - r_1*x2))*(n*r_0)
	 * (n*[c2*(r2_*x) s2*(r2_*x) (-[c1 s0*s1 c0*s1]*x)])*(n*r_0)+2*(n*(R*x+t-y))*(-n*[s1*c2 s1*s2 c1])
	 * (n*[-r1_*x r0_*x 0])*(n*r_0)+2*(n*(R*x+t-y))*(-n0*r10+n1*r00)
	 * //second row
	 * n0*(n*r_1)
	 * n1*(n*r_1)
	 * n2*(n*r_1)
	 * (n*(r_2*x1 - r_1*x2))*(n*r_1)+2*(n*(R*x+t-y))*(n*r_2)
	 * (n*[c2*(r2_*x) s2*(r2_*x) (-[c1 s0*s1 c0*s1]*x)])*(n*r_1)+2*(n*(R*x+t-y))*s0*(n*r_0)
	 * (n*[-r1_*x r0_*x 0])*(n*r_1)+2*(n*(R*x+t-y))*(-n0*r11+n1*r01)
	 * //third row
	 * n0*(n*r_2)
	 * n1*(n*r_2)
	 * n2*(n*r_2)
	 * (n*(r_2*x1 - r_1*x2))*(n*r_2)+2*(n*(R*x+t-y))*(-n*r_1)
	 * (n*[c2*(r2_*x) s2*(r2_*x) (-[c1 s0*s1 c0*s1]*x)])*(n*r_2)+2*(n*(R*x+t-y))*c0*(n*r_0)
	 * (n*[-r1_*x r0_*x 0])*(n*r_2)+2*(n*(R*x+t-y))*(-n0*r12+n1*r02)
	 * ]
	 */
	MatrixXd d2Jdtdx(3 * alignee.points.size(), 6);
	for(unsigned int i = 0; i < alignee.points.size(); i++)
	{
		const Vector3d x(alignee.points[i].x, alignee.points[i].y, alignee.points[i].z); //alignee pt
		const Vector3d y(alignand.points[i].x, alignand.points[i].y, alignand.points[i].z); //alignand pt
		const Vector3d n(alignand.points[i].normal[0], alignand.points[i].normal[1], alignand.points[i].normal[2]); //alignand normal

		d2Jdtdx.block<3, 3>(3 * i + 0, 0) = (R * n) * n.transpose();

		d2Jdtdx(3 * i + 0, 3) = n.dot(r_2*x[1] - r_1*x[2])*n.dot(r_0);
		d2Jdtdx(3 * i + 0, 4) = n.dot(Vector3d(c2*r2_.dot(x), s2*r2_.dot(x), -Vector3d(c1, s0*s1, c0*s1).dot(x)))*n.dot(r_0) + 2 * n.dot(R*x+t-y)*n.dot(-Vector3d(s1*c2, s1*s2, c1));
		d2Jdtdx(3 * i + 0, 5) = n.dot(Vector3d(-r1_.dot(x), r0_.dot(x), 0))*n.dot(r_0) + 2 * n.dot(R*x+t-y)*(-n[0]*R(1, 0)+n[1]*R(0, 0));

		d2Jdtdx(3 * i + 1, 3) = n.dot(r_2*x[1] - r_1*x[2])*n.dot(r_1) + 2 * n.dot(R*x+t-y)*n.dot(r_2);
		d2Jdtdx(3 * i + 1, 4) = n.dot(Vector3d(c2*r2_.dot(x), s2*r2_.dot(x), -Vector3d(c1, s0*s1, c0*s1).dot(x)))*n.dot(r_1) + 2 * n.dot(R*x+t-y)*n.dot(s0*r_0);
		d2Jdtdx(3 * i + 1, 5) = n.dot(Vector3d(-r1_.dot(x), r0_.dot(x), 0))*n.dot(r_1) + 2 * n.dot(R*x+t-y)*(-n[0]*R(1, 1)+n[1]*R(0, 1));

		d2Jdtdx(3 * i + 2, 3) = n.dot(r_2*x[1] - r_1*x[2])*n.dot(r_2) + 2 * n.dot(R*x+t-y)*n.dot(-r_1);
		d2Jdtdx(3 * i + 2, 4) = n.dot(Vector3d(c2*r2_.dot(x), s2*r2_.dot(x), -Vector3d(c1, s0*s1, c0*s1).dot(x)))*n.dot(r_2) + 2 * n.dot(R*x+t-y)*n.dot(c0*r_0);
		d2Jdtdx(3 * i + 2, 5) = n.dot(Vector3d(-r1_.dot(x), r0_.dot(x), 0))*n.dot(r_2) + 2 * n.dot(R*x+t-y)*(-n[0]*R(1, 2)+n[1]*R(0, 2));
	}
	d2Jdtdx *= 2;
	aux::testNaN(d2Jdtdx);
//	aux::writeMATLAB(d2Jdtdx, "d2jdtdx.dat");
	/*
	 * make it better conditioned so we don't end up with rows or cols of zeros in the result
	 */
	for(unsigned int i = 0; i < (unsigned)d2Jdtdx.rows(); i++)
		for(unsigned int j = 0; j < (unsigned)d2Jdtdx.cols(); j++)
			if(dabs(d2Jdtdx(i, j)) < EPSILON)
				d2Jdtdx(i, j) = d2Jdtdx(i, j) > 0 ? EPSILON : -EPSILON;

	const MatrixXd d2Jdt2Inv = d2Jdt2.inverse();
	aux::testNaN(d2Jdt2Inv);
	const MatrixXd S = d2Jdt2Inv * d2Jdtdx.transpose() * covX/*TODO but doesn't work 20110303: .selfadjointView<>()/* tell eigen this sparse mtx is symmetric */ * d2Jdtdx * d2Jdt2Inv;
	if(aux::anyNaN(S)) //debugging
	{
		aux::writeMATLAB(d2Jdt2, "d2Jdt2.dat");
		aux::writeMATLAB(d2Jdt2Inv, "d2Jdt2inv.dat");
		aux::writeMATLAB(d2Jdtdx, "d2Jdtdx.dat");
		aux::writeCovsMATLAB(covX, "covx.dat");
	}
	aux::testNaN(S);
	return S;
}

/*
 * return an empirical covariance matrix for the three translation components and roll, pitch, yaw as used by TORO
 *
 * pre: each ith point in alignee corresponds to the ith point in alignand; alignand has normals
 */
template <typename PointT, typename PointT2>
Matrix<double, 6, 6> getPointToPointICPXformCovarianceXlateRPY(const pcl::PointCloud<PointT>& alignee, const pcl::PointCloud<PointT2>& alignand)
{
	/*
	 * see A. Censi, "An Accurate Closed-Form Estimate of ICP's Covariance", ICRA07, especially equation 6, for math explanation
	 */

	assert(alignee.points.size() == alignand.points.size());

	/*
	 * for making things well-conditioned by removing unfortunately placed zeros
	 */
	static const double EPSILON = 1e-7;

	/*
	 * covariance of pt positions
	 */
	const SparseMatrix<double, rgbd::eigen::RowMajor> covX = getPointCovariance(alignee);

	/*
	 * second derivatives of J
	 */
	/*
	 * express Euler angles as a rotation matrix: see the xyz-order matrix in the wikipedia "euler angles" "table of matrices"
	 *
	 * (si, ci the sine/cosine of Euler angle i)
	 * r00 = c1*c2;
	 * r01 = c2*s0*s1 - c0*s2;
	 * r02 = c0*c2*s1 + s0*s2;
	 * r10 = c1*s2;
	 * r11 = c0*c2 + s0*s1*s2;
	 * r12 = c0*s1*s2 - c2*s0;
	 * r20 = -s1;
	 * r21 = c1*s0;
	 * r22 = c0*c1;
	 */
	/*
	 * (easy here since we're evaluating at the zero transformation)
	 */
	const double s1 = 0, s2 = 0, s3 = 0, c1 = 1, c2 = 1, c3 = 1; //sines/cosines
	Matrix3d R; R.setIdentity();
	const Vector3d t(0, 0, 0); //translation

	/*
	 * rows and cols of R, named using the conventions I use below
	 */
	const Vector3d r0_ = R.row(0), r1_ = R.row(1), r2_ = R.row(2), r_0 = R.col(0), r_1 = R.col(1), r_2 = R.col(2);

	MatrixXd d2Jdt2(6, 6);
	d2Jdt2.setZero();
	for(unsigned int i = 0; i < alignee.points.size(); i++)
	{
		const Vector3d x(alignee.points[i].x, alignee.points[i].y, alignee.points[i].z); //alignee pt
		const Vector3d y(alignand.points[i].x, alignand.points[i].y, alignand.points[i].z); //alignand pt

		/*
		 * d/dt dJ/dt = 2I
		 */
		d2Jdt2.block<3, 3>(0, 0) += 2 * Matrix3d::Identity();

		/*
		 * for matrix m, m0_ the 0th row, m_0 the 0th col
		 *
		 * d/dt dJ/dr =
		 *
		 * 2*
		 * //first col
		 * (r_2*x1-r_1*x2)`
		 * //second col
		 * c3*r2_.x
		 * s3*r2_.x
		 * -(r0_*c3+r1_*s3).x
		 * //third col
		 * -r1_.x
		 * r0_.x
		 * 0
		 */
		Matrix3d tmp;
		tmp.block<3, 1>(0, 0) = (r_2 * x[1] - r_1 * x[2]).transpose();
		tmp(0, 1) = c3 * r2_.dot(x);
		tmp(1, 1) = s3 * r2_.dot(x);
		tmp(2, 1) = -(r0_ * c3 + r1_ * s3).dot(x);
		tmp(0, 2) = -r1_.dot(x);
		tmp(1, 2) = r0_.dot(x);
		tmp(2, 2) = 0;
		d2Jdt2.block<3, 3>(0, 3) += 2 * tmp;

		/*
		 * d/dr dJ/dr =
		 *
		 * 2*
		 * //first row
		 * (r_2*x1-r_1*x2).(r_2*x1-r_1*x2) +W.(-r_1*x1-r_2*x2)
		 * //c3*r2_.x*(r02*x1-r01*x2) +s3*r2_.x*(r12*x1-r11*x2) +(r0_.x+t0-y0)*c3*(r22*x1-r21*x2) +(r1_.x+t1-y1)*s3*(r22*x1-r21*x2) +(-(r0_*c3+r1_*s3).x)*(r22*x1-r21*x2) +(r2_.x+t2-y2)*(-s2*c1*x1+s1*s2*x2)
		 * //(t0-y0)*(-r12*x1+r11*x2)+(t1-y1)*(r02*x1-r01*x2)
		 * //second row
		 * ([c3*r2_; s3*r2_; -(r0_*c3+r1_*s3)]*x).(r_2*x1-r_1*x2) +W.[c3*(r22*x1-r21*x2) s3*(r22*x1-r21*x2) (-c2*s1*x1+s1*s2*x2)]
		 * (r2_.x)^2+(-(r0_*c3+r1_*s3).x)^2+W.(-[c3*(r0_*c3+r1_*s3); s3*(r0_*c3+r1_*s3); r2_]*x)
		 * //(t0-y0)*s3*-r2_.x+(t1-y1)*c3*r2_.x
		 * //third row
		 * (t0-y0)*(-r12*x1+r11*x2)+(t1-y1)*(r02*x1-r01*x2)
		 * (t0-y0)*s3*-r2_.x+(t1-y1)*c3*r2_.x
		 * (t0-y0)*-r0_.x+(t1-y1)*-r1_.x
		 *
		 * where W = Rx + t - y
		 */
		const Vector3d W = R * x + t - y;
		tmp(0, 0) = (r_2*x[1]-r_1*x[2]).dot(r_2*x[1]-r_1*x[2]) + W.dot(-r_1*x[1]-r_2*x[2]);
		Matrix3d tmp2;
		tmp2.row(0) = c3*r2_;
		tmp2.row(1) = s3*r2_;
		tmp2.row(2) = -(r0_*c3+r1_*s3);
		tmp(0, 1) = tmp(1, 0) = (tmp2 * x).dot(r_2*x[1]-r_1*x[2]) + W.dot(Vector3d(c3*(R(2, 2)*x[1]-R(2, 1)*x[2]), s3*(R(2, 2)*x[1]-R(2, 1)*x[2]), (-c2*s1*x[1]+s1*s2*x[2])));
		tmp(0, 2) = tmp(2, 0) = (t[0]-y[0])*(-R(1, 2)*x[1]+R(1, 1)*x[2])+(t[1]-y[1])*(R(0, 2)*x[1]-R(0, 1)*x[2]);
		tmp2.row(0) = c3*(r0_*c3+r1_*s3);
		tmp2.row(1) = s3*(r0_*c3+r1_*s3);
		tmp2.row(2) = r2_;
		tmp(1, 1) = sqr(r2_.dot(x)) + sqr(-(r0_*c3+r1_*s3).dot(x)) - W.dot(tmp2 * x);
		tmp(1, 2) = tmp(2, 1) = (t[0]-y[0])*s3*-r2_.dot(x)+(t[1]-y[1])*c3*r2_.dot(x);
		tmp(2, 2) = (t[0]-y[0])*-r0_.dot(x)+(t[1]-y[1])*-r1_.dot(x);
		d2Jdt2.block<3, 3>(3, 3) += 2 * tmp;
	}
	d2Jdt2.block<3, 3>(3, 0) = d2Jdt2.block<3, 3>(0, 3).transpose();
//	aux::writeMATLAB(d2Jdt2, "d2jdt2.dat");
	/*
	 * ensure it's invertible
	 */
	for(unsigned int i = 0; i < 6; i++)
		if(dabs(d2Jdt2(i, i)) < EPSILON)
			d2Jdt2(i, i) = d2Jdt2(i, i) > 0 ? EPSILON : -EPSILON;

	/*
	 * d/dt dJ/dx for one x =
	 *
	 * 2*
	 * //first row
	 * r00
	 * r10
	 * r20
	 * (r_2*x1-r_1*x2).r_0
	 * -(Rx+t-y).[s2*c3 s2*s3 c2] +(r2_.x)*c2 -(-[c2 s2*s1 c1*s2].x)*s2
	 * -(t0-y0)*r10+(t1-y1)*r00
	 * //second row
	 * r01
	 * r11
	 * r21
	 * (r_2*x1-r_1*x2).r_1 +(Rx+t-y).r_2
	 * ([c3*r2_; s3*r2_; -[c2 s2*s1 c1*s2]]*x).r_1 +s1*(Rx+t-y).r_0
	 * -(t0-y0)*r11+(t1-y1)*r01
	 * //third row
	 * r02
	 * r12
	 * r22
	 * (r_2*x1-r_1*x2).r_2 -(Rx+t-y).r_1
	 * ([c3*r2_; s3*r2_; -[c2 s2*s1 c1*s2]]*x).r_2 +c1*(Rx+t-y).r_0
	 * -(t0-y0)*r12+(t1-y1)*r02
	 */
	MatrixXd d2Jdtdx(3 * alignee.points.size(), 6);
	for(unsigned int i = 0; i < alignee.points.size(); i++)
	{
		const Vector3d x(alignee.points[i].x, alignee.points[i].y, alignee.points[i].z); //alignee pt
		const Vector3d y(alignand.points[i].x, alignand.points[i].y, alignand.points[i].z); //alignand pt

		d2Jdtdx.block<3, 3>(3 * i, 0) = R.transpose();
		d2Jdtdx(3 * i + 0, 3) = (r_2 * x[1] - r_1 * x[2]).dot(r_0);
		d2Jdtdx(3 * i + 1, 3) = (r_2 * x[1] - r_1 * x[2]).dot(r_1) + (R * x + t - y).dot(r_2);
		d2Jdtdx(3 * i + 2, 3) = (r_2 * x[1] - r_1 * x[2]).dot(r_2) - (R * x + t - y).dot(r_1);
		d2Jdtdx(3 * i + 0, 4) = -(R * x + t - y).dot(Vector3d(s2*c3, s2*s3, c2)) + (r2_.dot(x)) * c2 + Vector3d(c2, s2*s1, c1*s2).dot(x)*s2;
		Matrix3d tmp;
		tmp.row(0) = c3 * r2_;
		tmp.row(1) = s3 * r2_;
		tmp.row(2) = -Vector3d(c2, s2*s1, c1*s2).transpose();
		d2Jdtdx(3 * i + 1, 4) = (tmp * x).dot(r_1) + s1 * (R * x + t - y).dot(r_0);
		d2Jdtdx(3 * i + 2, 4) = (tmp * x).dot(r_2) + c1 * (R * x + t - y).dot(r_0);
		d2Jdtdx.block<3, 1>(3 * i + 0, 5) = (-(t[0] - y[0]) * r1_ + (t[1] - y[1]) * r0_).transpose();
	}
	d2Jdtdx *= 2;
//	aux::writeMATLAB(d2Jdtdx, "d2jdtdx.dat");
	/*
	 * make it better conditioned so we don't end up with rows or cols of zeros in the result
	 */
	for(unsigned int i = 0; i < (unsigned)d2Jdtdx.rows(); i++)
		for(unsigned int j = 0; j < (unsigned)d2Jdtdx.cols(); j++)
			if(dabs(d2Jdtdx(i, j)) < EPSILON)
				d2Jdtdx(i, j) = d2Jdtdx(i, j) > 0 ? EPSILON : -EPSILON;

	const MatrixXd d2Jdt2Inv = d2Jdt2.inverse();
	return d2Jdt2Inv * d2Jdtdx.transpose() * covX * d2Jdtdx * d2Jdt2Inv;
}

} //namespace
