/*
 * icpPoint2Plane: ICP point-to-plane error function
 *
 * Evan Herbst
 * 7 / 10 / 10
 */

#include <cmath> //sqrt()
#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>

#define POINT_VECTOR_STUFF \
	BOOST_STATIC_ASSERT((boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector3f> >::value || boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector4f, rgbd::eigen::aligned_allocator<rgbd::eigen::Vector4f> > >::value));\
	typedef typename PointVectorT::value_type PointT;

namespace icp
{
namespace point2plane
{

template <typename PointVectorT>
void getSourceErrorVector(const PointVectorT& sourcePoints, const PointVectorT& targetNormals, const std::vector<float>& pointWeights,
									rgbd::eigen::Affine3f const& transform, const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(point_count);

	// point-to-plane
	// Need "n_t * T(s)" in (n_t*(T(s) - t))^2
	for (unsigned int i = 0; i < point_count; ++i) {
		if (correspondence_indices[i] >= 0) {
			PointT transformed_source = transform * sourcePoints[i];
			float dot_product = targetNormals[correspondence_indices[i]].dot(transformed_source);
			float new_value = pointWeights[i] * dot_product;
			result.push_back(new_value);
		}
	}
}
template <typename PointVectorT>
void getTargetErrorVector(const PointVectorT& targetPoints, const PointVectorT& targetNormals, const std::vector<float>& pointWeights,
									const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(point_count);

	// point-to-plane
	// Need "n_t * t" in (n_t*(T(s) - t))^2
	for (unsigned int i = 0; i < point_count; ++i) {
		if (correspondence_indices[i] >= 0) {
			PointT target = targetPoints[correspondence_indices[i]];
			float dot_product = targetNormals[correspondence_indices[i]].dot(target);
			float new_value = pointWeights[i] * dot_product;
			result.push_back(new_value);
		}
	}
}

template <typename PointVectorT>
void getSourceErrorVector(const PointVectorT& sourcePoints, const PointVectorT& targetPoints, const PointVectorT& targetNormals, const std::vector<rgbd::eigen::Matrix4f>& covs,
									rgbd::eigen::Affine3f const& transform, const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(3 * point_count);

	/*
	 * for covariance-weighted error we'll put all the info in the "source" vec since the per-point error can't be written (a - b)^2
	 *
	 * expected error = E((n_t'(x_s - x_t))^2) = n'(S+uu')n - 2(u - x_t)'nn'x_t
	 */
	for (unsigned int i = 0; i < point_count; ++i) {
		if (correspondence_indices[i] >= 0) {
			const PointT source_mu = transform * sourcePoints[i], target = targetPoints[correspondence_indices[i]], normal = targetNormals[correspondence_indices[i]];
			result.push_back(sqrt((normal.transpose() * (covs[i] + source_mu * source_mu.transpose()) * normal - 2 * (source_mu - target).transpose() * normal * normal.transpose() * target)(0, 0)));
		}
	}
}
template <typename PointVectorT>
void getTargetErrorVector(const PointVectorT& targetPoints, const PointVectorT& targetNormals, const std::vector<rgbd::eigen::Matrix4f>& covs,
									const std::vector<int>& correspondence_indices, std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(3 * point_count);

	/*
	 * for covariance-weighted error we'll put all the info in the "source" vec since the per-point error can't be written (a - b)^2
	 */
	for (unsigned int i = 0; i < point_count; ++i) {
		if (correspondence_indices[i] >= 0) {
			result.push_back(0);
		}
	}
}

} //namespace
} //namespace

#undef POINT_VECTOR_STUFF
