/*
 * icpPoint2Point
 *
 * Provides source and target error vectors for ICP
 *
 * Evan Herbst
 * 7 / 10 / 10
 */

#include <boost/static_assert.hpp>
#include <boost/type_traits/is_same.hpp>

#define POINT_VECTOR_STUFF \
	BOOST_STATIC_ASSERT((boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector3f> >::value || boost::is_same<PointVectorT, std::vector<rgbd::eigen::Vector4f, rgbd::eigen::aligned_allocator<rgbd::eigen::Vector4f> > >::value));\
	typedef typename PointVectorT::value_type PointT;

namespace icp
{
namespace point2pointReprojection
{

template <typename PointVectorT>
void getSourceErrorVector(
		boost::shared_ptr<rgbd::CameraParams> const& rgbd_camera_params_ptr,
		float stereo_baseline,
		const PointVectorT& sourcePoints,
		const std::vector<float>& pointWeights,
		rgbd::eigen::Affine3f const& transform,
		const std::vector<int>& correspondence_indices,
		std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(3 * point_count);

	// point to point.  Need a vector of "T(s)" in: (T(s) - t)^2
	for (unsigned int i = 0; i < point_count; i++) {
		if (correspondence_indices[i] >= 0) {
			PointT point_3d = transform * sourcePoints[i];

			// get u, v, d(isparity)
			float inverse_depth = 1.0 / point_3d[2];
			float u = inverse_depth * rgbd_camera_params_ptr->focalLength * point_3d[0] + rgbd_camera_params_ptr->centerX;
			float v = inverse_depth * rgbd_camera_params_ptr->focalLength * point_3d[1] + rgbd_camera_params_ptr->centerY;
			float d = inverse_depth * rgbd_camera_params_ptr->focalLength * stereo_baseline;

			// debug
			//cout << "Source u,v,d: " << u << ", " << v << ", " << d << endl;

			// debug
			if (isnan(u) || isnan(v) || isnan(d)) {
				std::cout << "source icp reprojection nan: " << u << ", " << v << ", " << d << std::endl;
				std::cout << "...from point:" << point_3d.transpose() << std::endl;
			}

			result.push_back(pointWeights[i] * u);
			result.push_back(pointWeights[i] * v);
			result.push_back(pointWeights[i] * d);
#if 0
			for (unsigned int j = 0; j < 3; j++) {
				float new_value = pointWeights[i] * transformed_source[j];
				result.push_back(new_value);
			}
#endif
		}
	}
}

template <typename PointVectorT>
void getTargetErrorVector(
		boost::shared_ptr<rgbd::CameraParams> const& rgbd_camera_params_ptr,
		float stereo_baseline,
		const PointVectorT& targetPoints,
		const std::vector<float>& pointWeights,
		const std::vector<int>& correspondence_indices,
		std::vector<float> & result)
{
	POINT_VECTOR_STUFF

	// this may include -1 indices
	const unsigned int point_count = correspondence_indices.size();
	result.clear();
	result.reserve(3 * point_count);

	// point to point.  Need a vector of "t" in: (T(s) - t)^2
	for (unsigned int i = 0; i < point_count; i++) {
		if (correspondence_indices[i] >= 0) {
			PointT point_3d = targetPoints[correspondence_indices[i]];

			// get u, v, d(isparity)
			float inverse_depth = 1.0 / point_3d[2];
			float u = inverse_depth * rgbd_camera_params_ptr->focalLength * point_3d[0] + rgbd_camera_params_ptr->centerX;
			float v = inverse_depth * rgbd_camera_params_ptr->focalLength * point_3d[1] + rgbd_camera_params_ptr->centerY;
			float d = inverse_depth * rgbd_camera_params_ptr->focalLength * stereo_baseline;

			// debug
			//cout << "Target u,v,d: " << u << ", " << v << ", " << d << endl;

			// debug
			if (isnan(u) || isnan(v) || isnan(d)) {
				std::cout << "target icp reprojection nan: " << u << ", " << v << ", " << d << std::endl;
				std::cout << "...from point:" << point_3d.transpose() << std::endl;
				std::cout << "i: " << i << ", correspondence_indices[i]: " << correspondence_indices[i] << std::endl;
			}

			result.push_back(pointWeights[i] * u);
			result.push_back(pointWeights[i] * v);
			result.push_back(pointWeights[i] * d);
#if 0
			for (unsigned int j = 0; j < 3; j++) {
				float new_value = pointWeights[i] * target[j];
				result.push_back(new_value);
			}
#endif
		}
	}
}

} // ns
} // ns

#undef POINT_VECTOR_STUFF
