/*
 * icp_combined: cleaner implementation of combined icp (multiple clouds, multiple error functions, priors, and fixed correspondences)
 *
 * Evan Herbst
 * 6 / 17 / 10
 */

#include <opencv2/imgproc/imgproc.hpp>
#include <cv_bridge/cv_bridge.h>
#include "rgbd_util/timer.h"
#include "rgbd_util/colorConversion.h" //rgb2hsv()
#include "pcl_rgbd/pointTypes.h"
#include "pcl_rgbd/cloudUtils.h"
#include "pcl_rgbd/cloudNormals.h"
#include "pcl_rgbd/cloudSearchTrees.h"

namespace registration
{

template <typename PointT1, typename PointT2>
ICPCloudPair::ICPCloudPair(const ICPCloudPairParams& params, const pcl::PointCloud<PointT1>& sourceCloud, const pcl::PointCloud<PointT2>& targetCloud, const boost::optional<colorMatchingParams>& cmparams)
: m_params(params)
{
	if(sourceCloud.points.empty()) ROS_FATAL("source cloud is empty");
	if(targetCloud.points.empty()) ROS_FATAL("target cloud is empty");

	rgbd::convertPointCloudToVector(sourceCloud, m_source_eigen_points);
	rgbd::convertPointCloudToVector(targetCloud, m_target_eigen_points);
	extractSrcNormalsIfNeeded(sourceCloud);
	extractTgtNormalsIfNeeded(targetCloud);

	// don't need any search structure if fixed
	if (!params.fixed_correspondence) {
		switch(m_params.corrType)
		{
			case ICP_CORRS_SINGLE_TREE:
				initSingleTreeCorrs(targetCloud);
				break;
			case ICP_CORRS_COLOR_BINNING:
				initColorBinning(sourceCloud, targetCloud);
				break;
			case ICP_CORRS_COLOR_6D_SEARCH:
				initColor6DSearchCorrs(targetCloud);
				break;
			case ICP_CORRS_COLOR_5D_LAB_SEARCH:
				assert(false && "not implemented");
				initColor5DLABSearchCorrs(targetCloud);
				break;
			case ICP_CORRS_COLOR_5D_HSV_SEARCH:
				initColor5DHSVSearchCorrs(targetCloud);
				break;
			case ICP_CORRS_REPROJECTION:
				// no tree needed
				break;
			default: assert(false);
		}
	}

	setDefaultWeights();

	if(cmparams) enableColorMatching(sourceCloud, targetCloud, *cmparams);
	else m_color_error_term_enabled = false;
}

/*
 * for running with many source clouds against the same target, to minimize recomputation of indices
 *
 * you'll also have to reset point weights and whatever other src-dependent things you've set
 */
template <typename PointT1>
void ICPCloudPair::replaceSourceCloud(const pcl::PointCloud<PointT1>& newSourceCloud)
{
	if(newSourceCloud.points.empty()) ROS_FATAL("source cloud is empty");
	rgbd::convertPointCloudToVector(newSourceCloud, m_source_eigen_points);
	extractSrcNormalsIfNeeded(newSourceCloud);

	switch(m_params.corrType)
	{
		case ICP_CORRS_SINGLE_TREE:
			//no-op
			break;
		case ICP_CORRS_COLOR_BINNING:
		{
			/*
			 * set which k-d tree each source point will use for lookup
			 */
			sourcePtKDTreeIndices.resize(newSourceCloud.points.size());
			for(unsigned int i = 0; i < newSourceCloud.points.size(); i++)
			{
				const boost::array<unsigned char, 3> hsv = rgbd::rgb2hsv(rgbd::unpackRGB<unsigned char>(newSourceCloud.points[i].rgb));
				const unsigned int hueIndex = std::max(0, std::min((signed)numHueBins - 1, ((signed)hsv[0] - (signed)hueBinSize / 4) / ((signed)hueBinSize / 2))),
					valueIndex = std::max(0, std::min((signed)numValueBins - 1, ((signed)hsv[2] - (signed)valueBinSize / 4) / ((signed)valueBinSize / 2)));
				sourcePtKDTreeIndices[i][0] = hueIndex;
				sourcePtKDTreeIndices[i][1] = valueIndex;
			}
			break;
		}
		default: assert(false && "unimplemented");
	}

	assert(!m_color_error_term_enabled); //if color matching is enabled I have no clue what needs updating -- EVH
}

/*
 * EXPERIMENTAL
 * for when you have a good estimate of the Targets's position and don't want the match influenced by distance points in the source
 * Added by Robin Held 12/7/11
 *
 * you'll also have to reset point weights and whatever other src-dependent things you've set
 */

template <typename PointT1>
void ICPCloudPair::replaceTargetCloud(const pcl::PointCloud<PointT1>& newTargetCloud)
{
	if(newTargetCloud.points.empty()) ROS_FATAL("target cloud is empty");
	rgbd::convertPointCloudToVector(newTargetCloud, m_target_eigen_points);
	extractTgtNormalsIfNeeded(newTargetCloud);

	if (!m_params.fixed_correspondence)
	{	
		switch(m_params.corrType)
		{
			case ICP_CORRS_SINGLE_TREE:
				initSingleTreeCorrs(newTargetCloud);
				break;
			default: assert(false);
		}
	}

	setDefaultWeights();

	assert(!m_color_error_term_enabled); //if color matching is enabled I have no clue what needs updating -- EVH ( & RTH)
}



template <typename PointT1, typename PointT2>
void ICPCloudPair::enableColorMatching(const pcl::PointCloud<PointT1>& sourceCloud, const pcl::PointCloud<PointT2>& targetCloud, const colorMatchingParams& cmparams,
													const typename boost::enable_if<boost::mpl::not_<boost::mpl::and_<rgbd::pointHasImgXY<PointT1>, rgbd::pointHasImgXY<PointT2> > > >::type* ignored)
{
	/*
	 * Duplicated from the other enableColorMatching:
	 */
	if(cmparams.errorTermWeight <= 0 && cmparams.colorDimensionScaling <= 0 && cmparams.colorCompatibilityMaxDist <= 0)
		return;

	m_color_error_term_enabled = (cmparams.errorTermWeight > 0);
	m_color_matching_weight = cmparams.errorTermWeight;
	m_color_matching_max_intensity = cmparams.maxIntensity;
	m_color_dimension_scaling = cmparams.colorDimensionScaling;
	m_color_compatibility_max_dist = cmparams.colorCompatibilityMaxDist;
	m_camera_params = cmparams.camParams;

	m_source_intensities.resize(sourceCloud.points.size());
	m_source_colors.resize(sourceCloud.points.size());
	for(unsigned int i=0; i<sourceCloud.points.size(); i++){
		const boost::array<float, 3> rgb = rgbd::unpackRGB<float>(sourceCloud.points[i].rgb);
		//conversion from http://opencv.willowgarage.com/documentation/cpp/
		// miscellaneous_image_transformations.html?highlight=cvtcolor#cvtColor
		m_source_intensities[i] = 0.299*rgb[0] + 0.587*rgb[1] + 0.114*rgb[2];
		m_source_colors[i] = rgbd::eigen::Vector3f(rgb[0],rgb[1],rgb[2]);
	}

	m_target_colors.resize(targetCloud.points.size());
	for(unsigned int i=0; i<targetCloud.points.size(); i++){
		const boost::array<float, 3> rgb = rgbd::unpackRGB<float>(targetCloud.points[i].rgb);
		m_target_colors[i] = rgbd::eigen::Vector3f(rgb[0],rgb[1],rgb[2]);
	}

	// debug
#if 0
	cout << "sourceCloud.size()" << sourceCloud.points.size() << endl;
	cout << "targetCloud.size()" << targetCloud.points.size() << endl;
	cout << "m_source_colors.size()" << m_source_colors.size() << endl;
	cout << "m_target_colors.size()" << m_target_colors.size() << endl;
#endif

	assert(!m_color_error_term_enabled && "Needs imgX and imgY");

	//this version, for if the point types don't have img{X,Y}, is necessary unless we also want enable_ifs on the ctor that calls this
	return;
}

template <typename PointT1, typename PointT2>
void ICPCloudPair::enableColorMatching(const pcl::PointCloud<PointT1>& sourceCloud, const pcl::PointCloud<PointT2>& targetCloud, const colorMatchingParams& cmparams,
													const typename boost::enable_if<boost::mpl::and_<rgbd::pointHasImgXY<PointT1>, rgbd::pointHasImgXY<PointT2> > >::type* ignored)
{
	if(cmparams.errorTermWeight <= 0 && cmparams.colorDimensionScaling <= 0 && cmparams.colorCompatibilityMaxDist <= 0)
		return;

	m_color_error_term_enabled = (cmparams.errorTermWeight > 0);
	m_color_matching_weight = cmparams.errorTermWeight;
	m_color_matching_max_intensity = cmparams.maxIntensity;
	m_color_dimension_scaling = cmparams.colorDimensionScaling;
	m_color_compatibility_max_dist = cmparams.colorCompatibilityMaxDist;
	m_camera_params = cmparams.camParams;

	m_source_intensities.resize(sourceCloud.points.size());
	m_source_colors.resize(sourceCloud.points.size());
	for(unsigned int i=0; i<sourceCloud.points.size(); i++){
		const boost::array<float, 3> rgb = rgbd::unpackRGB<float>(sourceCloud.points[i].rgb);
		//conversion from http://opencv.willowgarage.com/documentation/cpp/
		// miscellaneous_image_transformations.html?highlight=cvtcolor#cvtColor
		m_source_intensities[i] = 0.299*rgb[0] + 0.587*rgb[1] + 0.114*rgb[2];
		m_source_colors[i] = rgbd::eigen::Vector3f(rgb[0],rgb[1],rgb[2]);
	}

	m_target_colors.resize(targetCloud.points.size());
	for(unsigned int i=0; i<targetCloud.points.size(); i++){
		const boost::array<float, 3> rgb = rgbd::unpackRGB<float>(targetCloud.points[i].rgb);
		m_target_colors[i] = rgbd::eigen::Vector3f(rgb[0],rgb[1],rgb[2]);
	}

	//projections and pixelwise intensities and gradients only needed for color error term
	if(m_color_error_term_enabled){

		m_target_intensity_grid.resize(m_camera_params.xRes);
		m_target_gradient_grid.resize(m_camera_params.xRes);
		for(unsigned int x=0; x<m_camera_params.xRes; x++){
			m_target_intensity_grid[x].resize(m_camera_params.yRes);
			m_target_gradient_grid[x].resize(m_camera_params.yRes);
		}
		m_targetImgCoords.resize(targetCloud.points.size());

		for(unsigned int i = 0; i < targetCloud.points.size(); i++)
		{
			m_targetImgCoords[i].first = targetCloud.points[i].imgX;
			m_targetImgCoords[i].second = targetCloud.points[i].imgY;
		}

		const cv_bridge::CvImagePtr ciMsg = cv_bridge::toCvCopy(cmparams.targetImage, "mono8");
		cv::Mat greyscale = ciMsg->image;

		// try normalizing contrast
		cv::equalizeHist(greyscale, greyscale);

		//blur the image
		unsigned int kernelSize = 5; //TODO: PARAM
		cv::GaussianBlur(greyscale, greyscale, cv::Size(kernelSize, kernelSize), 0);

		//get the derivatives
		cv::Mat xDerivImage,yDerivImage;
		cv::Scharr(greyscale,xDerivImage,CV_32F,1,0);
		cv::Scharr(greyscale,yDerivImage,CV_32F,0,1);

		//store the target intensity value and the target gradient
		for(unsigned int x=0; x<m_camera_params.xRes; x++){
			for(unsigned int y=0; y<m_camera_params.yRes; y++){
				m_target_intensity_grid[x][y] = greyscale.at<unsigned char>(y,x)/255.0;
				m_target_gradient_grid[x][y].x() = xDerivImage.at<float>(y,x)/255.0;
				m_target_gradient_grid[x][y].y() = yDerivImage.at<float>(y,x)/255.0;
			}
		}
	}

}

template <typename PointT1>
void ICPCloudPair::extractSrcNormalsIfNeeded(const pcl::PointCloud<PointT1>& sourceCloud)
{
	const std::pair<bool, bool> normalsNeeded = areNormalsNeeded();
	if (normalsNeeded.first) rgbd::pointCloudNormalsToEigenNormals(sourceCloud, m_source_eigen_normals);
}
template <typename PointT2>
void ICPCloudPair::extractTgtNormalsIfNeeded(const pcl::PointCloud<PointT2>& targetCloud)
{
	const std::pair<bool, bool> normalsNeeded = areNormalsNeeded();
	if (normalsNeeded.second) rgbd::pointCloudNormalsToEigenNormals(targetCloud, m_target_eigen_normals);
}

template <typename PointT>
void ICPCloudPair::initSingleTreeCorrs(const pcl::PointCloud<PointT>& targetCloud)
{
	m_target_kdtree_ptr = rgbd::createKDTree2(targetCloud);
}

template <typename PointT1, typename PointT2>
void ICPCloudPair::initColorBinning(const pcl::PointCloud<PointT1>& sourceCloud, const pcl::PointCloud<PointT2>& targetCloud)
{
	/*
	 * create 2n - 1 bins in each dim, n - 1 of which halfway overlap the "base" bins
	 * (create extra overlapping bins to reduce quantization effects)
	 */
	targetKDTreesByColor.resize(boost::extents[numHueBins][numValueBins]);
	targetKDTreeIndices.resize(boost::extents[numHueBins][numValueBins]);
	for(unsigned int i = 0; i < numHueBins; i++)
		for(unsigned int j = 0; j < numValueBins; j++)
		{
			pcl::PointCloud<PointT2> subcloud;
			std::vector<unsigned int>& indices = targetKDTreeIndices[i][j];
			for(unsigned int k = 0; k < targetCloud.points.size(); k++)
			{
				const boost::array<unsigned char, 3> hsv = rgbd::rgb2hsv(rgbd::unpackRGB<unsigned char>(targetCloud.points[k].rgb));
				const unsigned int hueBaseIndex = 2 * (hsv[0] / hueBinSize), valueBaseIndex = 2 * (hsv[2] / valueBinSize); //base (non-overlapping) bin indices
				if(abs((signed)hueBaseIndex - (signed)i) <= 1 && abs((signed)valueBaseIndex - (signed)j) <= 1)
				{
					subcloud.points.push_back(targetCloud.points[k]);
					indices.push_back(k);
				}
			}
//			if(subcloud.points.empty()) throw runtime_error("empty subcloud! k-nn will break depending on query data");       instead we'll handle this case later

			targetKDTreesByColor[i][j] = rgbd::createKDTree2(subcloud);
		}

	/*
	 * set which k-d tree each source point will use for lookup
	 */
	sourcePtKDTreeIndices.resize(sourceCloud.points.size());
	for(unsigned int i = 0; i < sourceCloud.points.size(); i++)
	{
		const boost::array<unsigned char, 3> hsv = rgbd::rgb2hsv(rgbd::unpackRGB<unsigned char>(sourceCloud.points[i].rgb));
		const unsigned int hueIndex = std::max(0, std::min((signed)numHueBins - 1, ((signed)hsv[0] - (signed)hueBinSize / 4) / ((signed)hueBinSize / 2))),
			valueIndex = std::max(0, std::min((signed)numValueBins - 1, ((signed)hsv[2] - (signed)valueBinSize / 4) / ((signed)valueBinSize / 2)));
		sourcePtKDTreeIndices[i][0] = hueIndex;
		sourcePtKDTreeIndices[i][1] = valueIndex;
	}
}

template <typename PointT>
void ICPCloudPair::initColor6DSearchCorrs(const pcl::PointCloud<PointT>& targetCloud)
{
	boost::multi_array<float, 2> data(boost::extents[targetCloud.points.size()][6]);
	float* j = data.data();
	for (unsigned int i = 0; i < targetCloud.points.size(); i++)
	{
		*j++ = targetCloud.points[i].x;
		*j++ = targetCloud.points[i].y;
		*j++ = targetCloud.points[i].z;
		const boost::array<float, 3> rgb = rgbd::unpackRGB<float>(targetCloud.points[i].rgb);
		*j++ = m_color_dimension_scaling * rgb[0];
		*j++ = m_color_dimension_scaling * rgb[1];
		*j++ = m_color_dimension_scaling * rgb[2];
	}
	m_6dsearch_target_kdtree_ptr = boost::shared_ptr<kdtree2>(new kdtree2(data, true/* rearrange input pts */)); //if we rearrange input pts the tree doesn't need our multi_array to stay in scope (it has a copy)
}

template <typename PointT>
void ICPCloudPair::initColor5DLABSearchCorrs(const pcl::PointCloud<PointT>& targetCloud)
{
	boost::multi_array<float, 2> data(boost::extents[targetCloud.points.size()][5]);
	float* j = data.data();
	for (unsigned int i = 0; i < targetCloud.points.size(); i++)
	{
		*j++ = targetCloud.points[i].x;
		*j++ = targetCloud.points[i].y;
		*j++ = targetCloud.points[i].z;
		const boost::array<float, 3> rgb = rgbd::unpackRGB<float>(targetCloud.points[i].rgb);
		// here lab color
		assert (false && "to implement");

		*j++ = m_color_dimension_scaling * rgb[0];
		*j++ = m_color_dimension_scaling * rgb[1];
		*j++ = m_color_dimension_scaling * rgb[2];
	}
	m_5dsearch_target_kdtree_ptr = boost::shared_ptr<kdtree2>(new kdtree2(data, true/* rearrange input pts */)); //if we rearrange input pts the tree doesn't need our multi_array to stay in scope (it has a copy)
}

template <typename PointT>
void ICPCloudPair::initColor5DHSVSearchCorrs(const pcl::PointCloud<PointT>& targetCloud)
{
	boost::multi_array<float, 2> data(boost::extents[targetCloud.points.size()][5]);
	float* j = data.data();
	for (unsigned int i = 0; i < targetCloud.points.size(); i++)
	{
		*j++ = targetCloud.points[i].x;
		*j++ = targetCloud.points[i].y;
		*j++ = targetCloud.points[i].z;
		const boost::array<float, 2> cos_sin = rgbd::rgb2huecossin(rgbd::unpackRGB<float>(targetCloud.points[i].rgb));
		*j++ = m_color_dimension_scaling * cos_sin[0];
		*j++ = m_color_dimension_scaling * cos_sin[1];
	}
	m_5dsearch_target_kdtree_ptr = boost::shared_ptr<kdtree2>(new kdtree2(data, true/* rearrange input pts */)); //if we rearrange input pts the tree doesn't need our multi_array to stay in scope (it has a copy)
}

template <typename PointT1, typename PointT2>
void debugICPPublish(ICPCombined & icp_combined,
		unsigned int which_cloud_pair,
		pcl::PointCloud<PointT1> & source_cloud,
		pcl::PointCloud<PointT2> & target_cloud,
		const rgbd::eigen::Affine3f & transform_to_world,
		float animation_predelay,
		float animation_delay,
		float correspondence_line_width,
		ros::Publisher & source_cloud_publisher,
		ros::Publisher & target_cloud_publisher,
		ros::Publisher & marker_publisher,
		const std::string & frame_id)
{
	sensor_msgs::PointCloud2 cloud_msg;

	const std::deque<rgbd::eigen::Affine3f,rgbd::eigen::aligned_allocator<rgbd::eigen::Affine3f> > transform_history = icp_combined.getTransformHistory();
	const std::vector<std::vector<std::vector<int> > > corr_history = icp_combined.getCorrespondencesHistory();

	if (target_cloud_publisher.getNumSubscribers() == 0) return; // required

	pcl::PointCloud<PointT2> target_cloud_in_world;
	rgbd::transform_point_cloud(transform_to_world, target_cloud, target_cloud_in_world, true);
	target_cloud_in_world.header.frame_id = frame_id;
	pcl::toROSMsg(target_cloud_in_world, cloud_msg);
	target_cloud_publisher.publish(cloud_msg);

#if 0
	float normal_length = 0.05;
	float normal_width = 0.001;
	std::vector<unsigned int> empty_indices;
	visualization_msgs::Marker normal_marker = cloud_normal_marker(target_cloud_in_world, empty_indices, normal_length, normal_width, 1.0, 1.0, 1.0);
	normal_marker.ns = "target_normals";
	marker_publisher.publish(normal_marker);
#endif

	for (unsigned int i = 0 ; i < transform_history.size(); i++) {
		if (target_cloud_publisher.getNumSubscribers() > 0 && source_cloud_publisher.getNumSubscribers() > 0) {
			const rgbd::eigen::Affine3f & t = transform_history[i];
			pcl::PointCloud<PointT1> source_cloud_in_world;
			rgbd::transform_point_cloud(transform_to_world * t, source_cloud, source_cloud_in_world, false);
			source_cloud_in_world.header.frame_id = frame_id;
			pcl::toROSMsg(source_cloud_in_world, cloud_msg);
			source_cloud_publisher.publish(cloud_msg);

			if (marker_publisher.getNumSubscribers() > 0) {
				std::vector<std::pair<unsigned int, unsigned int> > corr_pairs;
				const std::vector<int> & this_step_pairs = corr_history[i][which_cloud_pair];
				for (unsigned int j = 0; j < this_step_pairs.size() ; j++) {
					if (this_step_pairs[j] >= 0) {
						corr_pairs.push_back(std::make_pair(j, (unsigned int) this_step_pairs[j]));
					}
				}
				visualization_msgs::Marker marker;
				marker = rgbd::cloud_correspondence_marker(
							source_cloud_in_world,
							target_cloud_in_world,
							corr_pairs, correspondence_line_width, 1, 0, 0);
				marker.ns = "icp_corrs";
				marker_publisher.publish(marker);
			}

			if (i == 0) {
				ros::Duration(animation_predelay).sleep();
			}

			ros::Duration(animation_delay).sleep();
		}
	}
}





} //namespace
