/*
 * icp_utility
 *
 * Evan Herbst
 * 6 / 17 / 10
 */

#include <boost/concept/assert.hpp>
#include <boost/concept_check.hpp>
#include <boost/range.hpp>
#include <boost/range/concepts.hpp>
#include <geometry_msgs/Point32.h>
#include <pcl/features/feature.h>

#include <iostream>

namespace registration
{

template <typename Real>
void transform_points(rgbd::eigen::Transform<Real,3,rgbd::eigen::Affine> const& t,
		std::vector<rgbd::eigen::Matrix<Real,4,1>,
			rgbd::eigen::aligned_allocator<rgbd::eigen::Matrix<Real,4,1> > > const& points,
		std::vector<unsigned int> const& to_transform,
		rgbd::eigen::Matrix<Real,4,1> *transformed_points)
{
	//indicates transform all points
	if(to_transform.empty()){
		for(unsigned int i=0; i<points.size(); i++){
			transformed_points[i] = t*points[i];
		}
	}
	//just those indicated in to_transform
	else{
		for(unsigned int i=0; i<to_transform.size(); i++){
			unsigned int j=to_transform[i];
			transformed_points[j] = t*points[j];
		}
	}
}


template <typename Real>
void transform_points(rgbd::eigen::Transform<Real,3,rgbd::eigen::Affine> const& t,
		std::vector<rgbd::eigen::Matrix<Real,3,1> > const& points,
		std::vector<bool> const& transform_point,
		std::vector<rgbd::eigen::Matrix<Real,3,1> > & transformed_points)
{
	transformed_points = points;
	if(points.empty())
		return;

	bool useNaive = true;
	if(useNaive){
		rgbd::eigen::Matrix<float,4,4> transform = t.matrix();
	    float rotation[9];
	    for(unsigned int row=0; row<3; row++)
	        for(unsigned int col=0; col<3; col++)
	            rotation[row*3+col] = transform(row, col);
	    float translation[3];
	    translation[0] = transform(0, 3);
	    translation[1] = transform(1, 3);
	    translation[2] = transform(2, 3);

	    for(unsigned int i=0; i<points.size(); i++) {
	    	if(transform_point[i]){
				transformed_points[i].x() = rotation[0] * points[i].x() + rotation[1] * points[i].y() + rotation[2] * points[i].z() + translation[0];
				transformed_points[i].y() = rotation[3] * points[i].x() + rotation[4] * points[i].y() + rotation[5] * points[i].z() + translation[1];
				transformed_points[i].z() = rotation[6] * points[i].x() + rotation[7] * points[i].y() + rotation[8] * points[i].z() + translation[2];
	    	}
	    }
	}
	else{
		for(unsigned int i=0; i<points.size(); i++){
			if(transform_point[i])
				transformed_points[i] = t*points[i];
		}
	}
}


template <typename PointT>
rgbd::eigen::Vector3f getEdgeDirection(const pcl::PointCloud<PointT>& edgeCloud, pcl::KdTreeFLANN<PointT> &edgeTree, unsigned int edgePtIndex, int k)
{
	std::vector<int> k_indices;
	std::vector<float> k_distances;

	edgeTree.nearestKSearch(edgePtIndex,k,k_indices,k_distances);
	geometry_msgs::Point32 centroid;
	// Compute the 3x3 covariance matrix
	rgbd::eigen::Matrix3d covariance_matrix;
	pcl::computeCovarianceMatrix (edgeCloud, k_indices, centroid, covariance_matrix);

	// Extract the eigenvalues and eigenvectors
	rgbd::eigen::SelfAdjointEigenSolver<rgbd::eigen::Matrix3d> ei_symm (covariance_matrix);
	rgbd::eigen::Matrix3d eigen_vectors = ei_symm.eigenvectors ();

	//XXX: not actually sure of the order. just following nearest.cpp's
	// indication that eigenvalues are in increasing order
	// Indeed, they are in increasing order (checked Eigen code) - Peter
	const rgbd::eigen::Vector3f edgeDir = eigen_vectors.col(2).normalized().cast<float>();
	return edgeDir;
}

template <typename PointT>
void getPtWeightsByDistFromCam(const pcl::PointCloud<PointT>& cloud, std::vector<float>& result_vector)
{
	result_vector.resize(cloud.points.size());
	for(unsigned int i=0; i < cloud.points.size(); i++) {
		const float z_value = cloud.points[i].z;
		result_vector[i] = 1.0 / (z_value * z_value * z_value * z_value);
	}
}

template <typename PointT>
int getBoundaryPoints(pcl::PointCloud<PointT> &organized_cloud, float max_depth_difference, std::vector<bool> & target_is_boundary)
{
	assert(isOrganizedCloud(organized_cloud));
	int boundary_count = 0;

	//figure out which of the target pixels are boundary pixels
	target_is_boundary.assign(organized_cloud.points.size(), false);
	for (unsigned int x = 0; x < organized_cloud.width; x++) {
		for (unsigned int y = 0; y < organized_cloud.height; y++) {
			unsigned int i = y * organized_cloud.width + x;
			float z = organized_cloud.points[i].z;

			//if it's at the edge of the image, it's definitely boundary
			if (x == 0 || x == organized_cloud.width - 1 || y == 0 || y == organized_cloud.height - 1) {
				target_is_boundary[i] = true;
				boundary_count++;
				continue;
			}

			// not at edge means these will always be valid
			if (organized_cloud.at(x - 1, y).z <= 0 ||
					organized_cloud.at(x + 1, y).z <= 0 ||
					organized_cloud.at(x, y - 1).z <= 0 ||
					organized_cloud.at(x, y + 1).z <= 0) {
				target_is_boundary[i] = true;
				boundary_count++;
				continue;
			}

			//check for depth discontinuities
			if (max_depth_difference > 0) {
				if (fabs(organized_cloud.at(x - 1, y).z - z) > max_depth_difference ||
						fabs(organized_cloud.at(x + 1, y).z - z) > max_depth_difference ||
						fabs(organized_cloud.at(x, y - 1).z - z) > max_depth_difference ||
						fabs(organized_cloud.at(x, y + 1).z - z) > max_depth_difference) {
					target_is_boundary[i] = true;
					boundary_count++;
					continue;
				}
			}
		}
	}
	return boundary_count;
}


template <typename PointT>
void getBoundaryPoints(pcl::PointCloud<PointT> &cloud, float max_depth_difference, unsigned int width, unsigned int height, std::vector<bool> & target_is_boundary)
{
	assert(!isOrganizedCloud(cloud)); // you should use the version for organized clouds
	target_is_boundary.assign(cloud.points.size(), false);
	//std::vector<std::vector<bool> > pixelOccupied(width, std::vector<bool>(height,false));
	std::vector<std::vector<float> > pixelDepth(width, std::vector<float>(height,-1.0));
	for(unsigned int i=0; i<cloud.points.size(); i++){
		unsigned int x = cloud.points[i].imgX, y = cloud.points[i].imgY;
		//pixelOccupied[x][y] = true;
		pixelDepth[x][y] = cloud.points[i].z;
	}
	for(unsigned int i=0; i<cloud.points.size(); i++){
		unsigned int x = cloud.points[i].imgX, y = cloud.points[i].imgY;
		const float z = cloud.points[i].z;

		//if it's at the edge of the image, it's definitely boundary
		if(x==0 || x==width-1 || y == 0 || y==height-1) {
			target_is_boundary[i] = true;
		}
		//otherwise, check neighbors
//		else if(!pixelOccupied[x-1][y] || !pixelOccupied[x+1][y] ||
//				!pixelOccupied[x][y-1] || !pixelOccupied[x][y+1]) {
		else if(pixelDepth[x-1][y] < 0 || pixelDepth[x+1][y] < 0 ||
				pixelDepth[x][y-1] < 0 || pixelDepth[x][y+1] < 0) {
			target_is_boundary[i] = true;
		}
		else if (max_depth_difference > 0) {
			if(fabs(pixelDepth[x-1][y] - z) > max_depth_difference ||
			   fabs(pixelDepth[x+1][y] - z) > max_depth_difference ||
			   fabs(pixelDepth[x][y-1] - z) > max_depth_difference ||
			   fabs(pixelDepth[x][y+1] - z) > max_depth_difference) {
				target_is_boundary[i] = true;
			}
		}

	}
}


template <typename PointT>
typename pcl::PointCloud<PointT>::Ptr prepare_general_icp_cloud(
		typename pcl::PointCloud<PointT>::ConstPtr input_cloud_ptr,
		float random_downsample_rate,
		int min_random_downsample_points,
		float max_distance,
		bool sanity_check)
{
	typename pcl::PointCloud<PointT>::Ptr result_cloud_ptr(new pcl::PointCloud<PointT>);
	result_cloud_ptr->header = input_cloud_ptr->header; // probably unnecessary but very fast
	// also need a temporary location
	typename pcl::PointCloud<PointT>::Ptr temp_cloud_ptr(new pcl::PointCloud<PointT>);
	temp_cloud_ptr->header = input_cloud_ptr->header; // probably unnecessary but very fast

	// downsample randomly
	if (random_downsample_rate > 0.0) {
		std::vector<unsigned int> random_downsample_indices = getRandomDownsamplingIndices(input_cloud_ptr->points.size(), random_downsample_rate, min_random_downsample_points);
		downsampleFromIndices(*input_cloud_ptr, *temp_cloud_ptr, random_downsample_indices);
		temp_cloud_ptr.swap(result_cloud_ptr);
	}
	else {
		*result_cloud_ptr = *input_cloud_ptr; // must put something in result_cloud_ptr
	}

	// remove points beyond distance
	if (max_distance > 0.0) {
		rgbd::eigen::Vector3f center(0,0,0);
		const std::vector<unsigned int> within_max_distance_indices = getRadiusDownsamplingIndices(*result_cloud_ptr, center, max_distance, true);
		downsampleFromIndices(*result_cloud_ptr, *temp_cloud_ptr, within_max_distance_indices);
		temp_cloud_ptr.swap(result_cloud_ptr);
	}

	if (sanity_check) {
		using std::cout;
		using std::endl;
		cout << "---prepare_general_icp_cloud sanity check:" << endl;
		cout << "Size: " << result_cloud_ptr->points.size() << endl;
		for (unsigned int i = 0; i < result_cloud_ptr->points.size(); i++) {
			const PointT& p = result_cloud_ptr->points[i];
			float epsilon = 0.001;
			bool bad = false;
			bad = bad || (p.z <= epsilon && p.z >= -epsilon);
			bad = bad || isnan(p.x) || isnan(p.y) || isnan(p.z);
			if (bad) {
				cout << "bad point at index " << i << " : " << result_cloud_ptr->points[i] << endl;
			}
		}
	}

	return result_cloud_ptr;
}

} //namespace
