#include <pcl_rgbd/cloudSampling.h>
#include <pcl_rgbd/cloudNormals.h>

#include <boost/make_shared.hpp>

namespace segmentation{

template <typename PointT>
std::vector<typename pcl::PointCloud<PointT>::Ptr >
PointCloudSegmenter::segmentCameraCloud(
		pcl::PointCloud<PointT> const& sourceCloud,
		float maxEdgeDist,unsigned int minSegmentSize, unsigned int gridIncrement)
{
	assert(sourceCloud.height == 1); //assumes unorganized

	unsigned int increment = gridIncrement;
	unsigned int xPoints = (m_xRes-1)/increment + 1;
	unsigned int yPoints = (m_yRes-1)/increment + 1;

	//reconstruct 2D grid structure
	std::vector<std::vector<rgbd::eigen::Vector3f> > points3D(xPoints);
	std::vector<std::vector<unsigned int> > pointIndices(xPoints);
	std::vector<std::vector<bool> > validity(xPoints);

	//link structures
	std::vector<std::vector<bool> > hasXPlusLink(xPoints-1);
	std::vector<std::vector<bool> > hasYPlusLink(xPoints);

	//set up the data structures
	for(unsigned int x=0; x<xPoints; x++){
		points3D[x] = std::vector<rgbd::eigen::Vector3f>(yPoints);
		pointIndices[x] = std::vector<unsigned int>(yPoints);
		validity[x] = std::vector<bool>(yPoints,false);

		if(x < (xPoints-1))
			hasXPlusLink[x] = std::vector<bool>(yPoints,false);
		hasYPlusLink[x] = std::vector<bool>(yPoints-1,false);
	}

	//fill in the points and validity
	for(unsigned int i=0; i<sourceCloud.points.size(); i++){
		PointT const& pt = sourceCloud.points[i];

		if(pt.imgX % increment == 0 && pt.imgY % increment == 0){
			points3D[pt.imgX/increment][pt.imgY/increment] = rgbd::eigen::Vector3f(pt.x,pt.y,pt.z);
			pointIndices[pt.imgX/increment][pt.imgY/increment] = i;
			validity[pt.imgX/increment][pt.imgY/increment] = true;
		}
	}

	//fill in the edge maps
	for(unsigned int x=0; x<xPoints; x++){
		for(unsigned int y=0; y<yPoints; y++){
			//no edges from invalid points
			if(!validity[x][y])
				continue;

			//check for x-edge
			if(x < (xPoints-1) && validity[x+1][y]){
				rgbd::eigen::Vector3f diff = points3D[x+1][y]-points3D[x][y];
				if(diff.norm() <= maxEdgeDist)
					hasXPlusLink[x][y] = true;
			}
			//check for y-edge
			if(y < (yPoints-1) && validity[x][y+1]){
				rgbd::eigen::Vector3f diff = points3D[x][y+1]-points3D[x][y];
				if(diff.norm() <= maxEdgeDist)
					hasYPlusLink[x][y] = true;
			}
		}
	}

	//construct the segments
	std::vector<std::vector<unsigned int> > segments;
	std::vector<bool> containedInSegment(sourceCloud.points.size(),false);
	for(unsigned int i=0; i<sourceCloud.points.size(); i++){
		unsigned int x=sourceCloud.points[i].imgX;
		unsigned int y=sourceCloud.points[i].imgY;
		//if the point is valid and not already in a segment, start
		//a new segment as the connected component of this point
		if(!containedInSegment[pointIndices[x][y]]){
			std::list<unsigned int> component = computeConnectedComponentDFS(
					x,y,hasXPlusLink,hasYPlusLink,pointIndices,containedInSegment);
			std::vector<unsigned int> componentVector(component.begin(),component.end());
			segments.push_back(componentVector);
		}
	}

	//put the largest segments first
	std::sort(segments.begin(),segments.end(),PointCloudSegmenter::segmentSizeComparator);

//	std::cout<<"Segmentation produced "<<segments.size()<<" segments out of "<<sourceCloud.points.size()<<" points"<<std::endl;
//	std::cout<<"Largest is "<<segments.back().size()<<", smallest is "<<segments.front().size()<<std::endl;

	//construct the output
	std::vector<typename pcl::PointCloud<PointT>::Ptr > toReturn;
	int currentIndex = segments.size()-1;
	while(currentIndex >= 0 && segments[currentIndex].size() >= minSegmentSize){
		std::vector<unsigned int> &indices = segments[currentIndex];

		//add the point-cloud to the vector
		typename pcl::PointCloud<PointT>::Ptr cloud(new pcl::PointCloud<PointT>());
		rgbd::downsampleFromIndices(sourceCloud,*cloud,indices);
		toReturn.push_back(cloud);

		currentIndex--;
	}

	return toReturn;

}

template <typename PointT, typename Real>
std::vector<std::pair<unsigned int, unsigned int> >
PointCloudSegmenter::segmentUsingHandInformation(
		pcl::PointCloud<PointT> const& cloud,
		typename pcl::KdTree<PointT>::Ptr kdTree,
		std::vector<rgbd::eigen::Matrix<Real,3,1> > const& handPointLocations,
		rgbd::eigen::Vector3f const& centerOfFingers, rgbd::eigen::Vector3f const& wristLocation,
		HandSegmentationParams const& handSegParams)
{
	ros::Time startTime = ros::Time::now();

	assert(cloud.height == 1); //assumes unorganized

	std::vector<std::pair<unsigned int, unsigned int> > pixelsToUse;

	//if any part of a segment is below this point in the wrist direction (starting from wrist COM)
	//then the segment isn't added
	float handPointDownsample = 0.4; //TODO: PARAM
	rgbd::eigen::Vector3f wristDir = (centerOfFingers-wristLocation).normalized();


	//perform hand-removal downsampling
	ros::Time startRemovalTime = ros::Time::now();
	std::vector<bool> usePoint(cloud.points.size(),true);

	//remove near each of the finger tips to cause a disconnect
	for(unsigned int i=0; i<handPointLocations.size(); i++){

		//don't use all the hand points -- the radius search is slow
		float randVal = (rand()%10000)/10000.0;
		if(randVal > handPointDownsample){
			continue;
		}

		PointT center;
		center.x = handPointLocations[i].x();
		center.y = handPointLocations[i].y();
		center.z = handPointLocations[i].z();

		std::vector<int> k_indices;
		std::vector<float> k_distances;
		kdTree->radiusSearch(center,handSegParams.handRemovalDist,k_indices,k_distances,INT_MAX);

		for(unsigned int j=0; j<k_indices.size(); j++){
			usePoint[k_indices[j]] = false;
		}
	}

	//build a point cloud
	pcl::PointCloud<PointT> downsampledCloud;
	std::vector<unsigned int> acceptedIndices;
	rgbd::boolVectorToIndices(usePoint,acceptedIndices);
	rgbd::downsampleFromIndices(cloud,downsampledCloud,acceptedIndices);
	ros::Time endRemovalTime = ros::Time::now();
	ros::Duration removalDur = endRemovalTime-startRemovalTime;

//	std::cout<<"Removed "<<(cloud.points.size()-acceptedIndices.size())<<" points out of "<<
//		cloud.points.size()<<" during hand removal for segmentation"<<std::endl;
//	std::cout<<"Downsampled cloud has "<<downsampledCloud.points.size()<<" points"<<std::endl;

//	std::cout<<"Hand removal took "<<removalDur.toSec()<<" seconds"<<std::endl;

	//segment the downsampled cloud
	std::vector<typename pcl::PointCloud<PointT>::Ptr > segments = this->segmentCameraCloud(
			downsampledCloud,handSegParams.maxEdgeDist,handSegParams.minSegmentSize,1);

//	std::cout<<"Segmenting result gave "<<segments.size()<<" segments"<<std::endl;
//	if(segments.size() > 0)
//		std::cout<<"Largest has "<<segments[0]->points.size()<<" points"<<std::endl;

	//find the segments close enough to the reference point
	m_segmentationMap = boost::make_shared<boost::multi_array<bool,2> >(boost::extents[m_xRes][m_yRes]);
	std::fill(m_segmentationMap->data(),m_segmentationMap->data()+m_segmentationMap->num_elements(),false);
	unsigned int minX=0,maxX=0,minY=0,maxY=0;
	bool firstAdded = false;
	for(unsigned int seg=0; seg<segments.size(); seg++){
		//find the minimum distance for the segment
		float segmentMinDist=0;
		float maxSegmentWristDist=0;
		float minSegmentWristDist=0;
		for(unsigned i=0; i<segments[seg]->points.size(); i++){
			PointT const& pt = segments[seg]->points[i];
			rgbd::eigen::Vector3f eigenPt(pt.x,pt.y,pt.z);
			rgbd::eigen::Vector3f refDiff = eigenPt - centerOfFingers;
			rgbd::eigen::Vector3f wristDiff = eigenPt - wristLocation;
			float dist = refDiff.norm();
			float wristDist = wristDiff.dot(wristDir);

			if(i==0 || dist < segmentMinDist){
				segmentMinDist = dist;
			}
			if(i==0 || wristDist > maxSegmentWristDist){
				maxSegmentWristDist = wristDist;
			}
			if(i==0 || wristDist < minSegmentWristDist){
				minSegmentWristDist = wristDist;
			}

			if(minSegmentWristDist < handSegParams.wristLowerLimit)
				break;

//			if(segmentMinDist < proximityForAdding &&
//					maxSegmentWristDist > minDistAlongWrist)
//				break;
		}
		bool add = (segmentMinDist < handSegParams.proximityForAdding &&
				maxSegmentWristDist > handSegParams.minDistAlongWrist &&
				minSegmentWristDist > handSegParams.wristLowerLimit);

		//add to the pixel list
		if(add){
			for(unsigned int i=0; i<segments[seg]->points.size(); i++){
				unsigned int x = segments[seg]->points[i].imgX;
				unsigned int y = segments[seg]->points[i].imgY;
				pixelsToUse.push_back(std::pair<unsigned int, unsigned int>(x,y));
				(*m_segmentationMap)[x][y] = true;
				if(x < minX || !firstAdded)
					minX = x;
				if(x > maxX || !firstAdded)
					maxX = x;
				if(y < minY || !firstAdded)
					minY = y;
				if(y > maxY || !firstAdded)
					maxY = y;
				firstAdded = true;
			}
		}
	}
	//store the bounding box of the segments for later use
	m_segmentationMapMinX = minX;
	m_segmentationMapMaxX = maxX;
	m_segmentationMapMinY = minY;
	m_segmentationMapMaxY = maxY;

	ros::Time endTime = ros::Time::now();
	ros::Duration dur = endTime-startTime;
//	std::cout<<"Segmentation took "<<dur.toSec()<<" seconds"<<std::endl;

	m_segmentationDone = true;
	m_segmentationMapSet = true;
	m_gridIncrementUsed = 1;

	return pixelsToUse;

}

}//namespace
