/*
 * segmentation_utility.ipp
 *
 *  Created on: 2010-07-20
 *      Author: mkrainin
 */

#include <algorithm>

#include <pcl_rgbd/cloudSampling.h>
#include <boost/iterator/counting_iterator.hpp>
#include <boost/make_shared.hpp>

#include <ros/console.h>

namespace segmentation
{

template <typename PointT>
bool fitSACPlane (typename pcl::PointCloud<PointT>::ConstPtr const& cloud,
		std::vector<int> &indices,
		std::vector<int> &inliers,
		std::vector<double> &coeff,
		double dist_thresh,
		int max_iterations)
{
	typename pcl::SampleConsensusModelPlane<PointT>::Ptr model(
			new pcl::SampleConsensusModelPlane<PointT>(cloud,indices));
	pcl::MEstimatorSampleConsensus<PointT> sac(model,dist_thresh);
	sac.setMaxIterations(max_iterations);
	bool success = sac.computeModel();
	if(success){
		rgbd::eigen::VectorXf coeffVec;
		sac.getModelCoefficients(coeffVec);
		sac.getInliers(inliers);

		coeff.resize(coeffVec.size());
		for(unsigned int i=0; i<coeff.size(); i++)
			coeff[i] = coeffVec[i];
	}
	else
	{
		ROS_ERROR ("Could not compute a plane model.");
	}

	return success;
}

#if 0
template <typename PointT>
bool computePlaneWithNormal(
		typename pcl::PointCloud<PointT>::ConstPtr cloud_ptr,
		const std::vector<int> & all_indices,
		const std::vector<int> & sample_indices,
		double inlier_distance,
		const rgbd::eigen::Vector3f & normal_prior,
		double max_angle_difference,
		double displacement_from_origin,
		double max_displacement_difference,
		double normal_distance_weight,
		rgbd::eigen::VectorXf& plane_coefficients,
		std::vector<int>& plane_inliers,
		typename pcl::PointCloud<PointT>::Ptr projected_inliers,
		typename pcl::PointCloud<PointT>::Ptr convex_hull,
		bool do_euclidean_clustering,
		double euclidean_tolerance)
{
	typename pcl::SampleConsensusModelNormalPlane<PointT, PointT>::Ptr sac_model = boost::make_shared<pcl::SampleConsensusModelNormalPlane<PointT, PointT> >(cloud_ptr);
	sac_model->setInputNormals(cloud_ptr);
	sac_model->setAxis(normal_prior);
	if (max_angle_difference > 0) {
		sac_model->setEpsAngle(max_angle_difference);
	}

	if (max_displacement_difference > 0) {
		sac_model->setDistanceFromOrigin(displacement_from_origin);
		sac_model->setEpsDist(max_displacement_difference);
	}
	sac_model->setNormalDistanceWeight(normal_distance_weight);
	typename pcl::RandomSampleConsensus<PointT>::Ptr sac_method = boost::make_shared<pcl::RandomSampleConsensus<PointT> > (sac_model, inlier_distance);

	return computePlaneSACModel<PointT>(
			sac_model,
			sac_method,
			all_indices,
			sample_indices,
			inlier_distance,
			plane_coefficients,
			plane_inliers,
			projected_inliers,
			convex_hull,

			do_euclidean_clustering,
			euclidean_tolerance);
}
#endif

template <typename PointT>
bool computePlane(
		typename pcl::PointCloud<PointT>::ConstPtr cloud_ptr,
		const std::vector<int> & all_indices,
		const std::vector<int> & sample_indices,
		double inlier_distance,
		rgbd::eigen::VectorXf& plane_coefficients,
		std::vector<int>& plane_inliers,
		typename pcl::PointCloud<PointT>::Ptr projected_inliers,
		typename pcl::PointCloud<PointT>::Ptr convex_hull,
		bool do_euclidean_clustering,
		double euclidean_tolerance)
{
	typename pcl::SampleConsensusModelPlane<PointT>::Ptr sac_model = boost::make_shared<pcl::SampleConsensusModelPlane<PointT> >(cloud_ptr);
	typename pcl::RandomSampleConsensus<PointT>::Ptr sac_method = boost::make_shared<pcl::RandomSampleConsensus<PointT> > (sac_model, inlier_distance);

	return computePlaneSACModel<PointT>(
			sac_model,
			sac_method,
			all_indices,
			sample_indices,
			inlier_distance,
			plane_coefficients,
			plane_inliers,
			projected_inliers,
			convex_hull,
			do_euclidean_clustering,
			euclidean_tolerance);
}

template <typename PointT>
bool computePlaneSACModel(
		typename pcl::SampleConsensusModel<PointT>::Ptr sac_model,
		typename pcl::SampleConsensus<PointT>::Ptr sac_method,
		const std::vector<int> & all_indices,
		const std::vector<int> & sample_indices,
		double inlier_distance,
		rgbd::eigen::VectorXf& plane_coefficients,
		std::vector<int>& plane_inliers,
		typename pcl::PointCloud<PointT>::Ptr projected_inliers,
		typename pcl::PointCloud<PointT>::Ptr convex_hull,
		bool do_euclidean_clustering,
		double euclidean_tolerance)
{
	assert(sac_model);
	assert(sac_method);
	using std::vector;

	boost::shared_ptr<vector<int> > all_indices_ptr;
	if (!all_indices.empty()) {
		sac_model->setIndices(boost::make_shared<vector<int> > (all_indices));
	}
	all_indices_ptr = sac_model->getIndices();

	boost::shared_ptr<vector<int> > sample_indices_ptr;
	if (!sample_indices.empty()) {
		sac_model->setIndices(boost::make_shared<vector<int> > (sample_indices));
	}
	sample_indices_ptr = sac_model->getIndices();

	// avoid an infinite loop in PCL
	if (all_indices_ptr->size() < 3 || sample_indices_ptr->size() < 3)
		return false;

	// actually compute the plane
	if (!sac_method->computeModel (0)) return false;
	vector<int> downsampled_inliers;
	sac_method->getInliers (downsampled_inliers);
	rgbd::eigen::VectorXf original_coefficients;
	sac_method->getModelCoefficients (original_coefficients);

	// could optimize based on just downsampled inliers:
	//sac_model->optimizeModelCoefficients (downsampled_inliers, coeff, plane_coefficients);

	// find all inliers in original cloud
	sac_model->setIndices(*all_indices_ptr);
	sac_model->selectWithinDistance (original_coefficients, inlier_distance, plane_inliers);

	// refine based on ALL inliers
	sac_model->optimizeModelCoefficients(plane_inliers, original_coefficients, plane_coefficients);

	if (do_euclidean_clustering) {
		ROS_INFO_STREAM("Starting clustering...");
		// todo: replace with raw call to function instead of class (use normals)
		typename pcl::EuclideanClusterExtraction<PointT> euclidean_cluster;
		euclidean_cluster.setInputCloud(sac_model->getInputCloud());
		euclidean_cluster.setIndices(boost::make_shared<vector<int> >(plane_inliers));
		euclidean_cluster.setClusterTolerance(euclidean_tolerance);
		vector<pcl::PointIndices> point_indices_vector;
		euclidean_cluster.extract(point_indices_vector);
		ROS_INFO_STREAM("Euclidean clusters: " << point_indices_vector.size());
		if (!point_indices_vector.empty()) {
			ROS_INFO_STREAM("Largest cluster: " << point_indices_vector[0].indices.size());
		}
		// set inliers to largest cluster and refine again
		plane_inliers.assign(point_indices_vector[0].indices.begin(), point_indices_vector[0].indices.end());
		sac_model->optimizeModelCoefficients(plane_inliers, original_coefficients, plane_coefficients);
	}

	if (projected_inliers && convex_hull) {
		typename pcl::PointCloud<PointT>::Ptr cloud_of_inliers = boost::make_shared<pcl::PointCloud<PointT> >();
		rgbd::downsampleFromIndices(*sac_model->getInputCloud(), *cloud_of_inliers, plane_inliers);
		sac_model->setInputCloud(cloud_of_inliers);
		boost::shared_ptr<vector<int> > cloud_of_inliers_indices = boost::make_shared<vector<int> > (boost::counting_iterator<int>(0), boost::counting_iterator<int>(cloud_of_inliers->points.size()));
		sac_model->setIndices(cloud_of_inliers_indices);
		sac_model->projectPoints(*cloud_of_inliers_indices, plane_coefficients, *projected_inliers, true);

		// convex hull of projectect points
		// TODO: Just use xyz for output?
		pcl::ConvexHull<PointT> chull;

		chull.setInputCloud (projected_inliers);
		chull.reconstruct (*convex_hull);
	}

	return true;
}

template <typename PointT>
visualization_msgs::Marker getConvexHullMarker(
		typename pcl::PointCloud<PointT>::ConstPtr convex_hull,
		const std::string & ns,
		int id,
		const std_msgs::ColorRGBA & color)
{
	// convert convex hull into a marker
	visualization_msgs::Marker convex_hull_marker;
	convex_hull_marker.header = convex_hull->header;
	convex_hull_marker.ns = ns;
	convex_hull_marker.id = id;
	convex_hull_marker.type = visualization_msgs::Marker::TRIANGLE_LIST;
	convex_hull_marker.color = color;
	convex_hull_marker.scale.x = convex_hull_marker.scale.y = convex_hull_marker.scale.z = 1.0;
	// pose default

	for (unsigned int j = 1; j < convex_hull->points.size() - 1; j++) {
		// 0,j,j+1
		convex_hull_marker.points.push_back(geometry_msgs::Point());
		convex_hull_marker.points.back().x = convex_hull->points[0].x;
		convex_hull_marker.points.back().y = convex_hull->points[0].y;
		convex_hull_marker.points.back().z = convex_hull->points[0].z;

		convex_hull_marker.points.push_back(geometry_msgs::Point());
		convex_hull_marker.points.back().x = convex_hull->points[j].x;
		convex_hull_marker.points.back().y = convex_hull->points[j].y;
		convex_hull_marker.points.back().z = convex_hull->points[j].z;

		convex_hull_marker.points.push_back(geometry_msgs::Point());
		convex_hull_marker.points.back().x = convex_hull->points[j+1].x;
		convex_hull_marker.points.back().y = convex_hull->points[j+1].y;
		convex_hull_marker.points.back().z = convex_hull->points[j+1].z;
	}

	return convex_hull_marker;
}

} //namespace
