/*
 * readOneCloud: get a cloud from some file format based on a spec
 *
 * Evan Herbst
 * 6 / 16 / 10
 */

#include <cassert>
#include <iostream>
#include <stdexcept>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/algorithm/string/split.hpp>
#include "rgbd_bag_utils/contents.h"
#include "rgbd_bag_utils/rgbdBagReader.h"
#include "pcl_rgbd/cloudUtils.h"
#include "pcl_rgbd/depth_to_cloud_lib.h"
#include "pcl_rgbd/cloudTofroPLY.h"
using std::cout;
using std::endl;
namespace fs = boost::filesystem;
namespace alg = boost::algorithm;

/*
 * possible formats for framespec:
 * - ply_filepath
 * - filepath for colored depth map
 * - bag_filepath:frame_index
 * - bag_filepath:time_sec:time_nsec
 */
template <typename PointT>
void readCloud(pcl::PointCloud<PointT>& cloud, const std::string& framespec)
{
	if(alg::ends_with(framespec, ".ply"))
	{
		cout << "reading ply " << framespec << endl;
		sensor_msgs::PointCloud2 cloud2, cloud3;
		rgbd::read_ply_file(cloud2, framespec);
		rgbd::packColorChannels("diffuse_red", "diffuse_green", "diffuse_blue", cloud2, cloud3);
		pcl::fromROSMsg(cloud3, cloud);
	}
	else
	{
		std::vector<std::string> tokens;
		alg::split(tokens, framespec, alg::is_any_of(":"));
		if(tokens.size() == 2)
		{
			const fs::path bagFilepath(tokens[0]);
			const unsigned int frameIndex = boost::lexical_cast<unsigned int>(tokens[1]);
			std::cout << "reading bag " << bagFilepath << ", frame " << frameIndex << std::endl;

			std::string depthTopic, imgTopic;
			bool depthCompressed, imgCompressed;
			rgbd::determineRGBDBagSchema(bagFilepath, depthTopic, imgTopic, depthCompressed, imgCompressed);
//			imgTopic = "/rgbd/image";
			rgbd::rgbdBagReader<PointT> frameReader(bagFilepath, depthTopic, imgTopic, frameIndex, 1000000/* max end frame index */, 0/* frameskip */, 0u/* num prev frames to keep */,
				primesensor::getColorCamParams(rgbd::KINECT_640_DEFAULT), primesensor::getColorCamParams(rgbd::KINECT_640_DEFAULT));
			frameReader.readOne();
			cloud = frameReader.getLastCloud();
		}
		else if(tokens.size() == 3)
		{
			const fs::path bagFilepath(tokens[0]);
			const unsigned long sec = boost::lexical_cast<unsigned long>(tokens[1]), nsec = boost::lexical_cast<unsigned long>(tokens[2]);
			const ros::Time t(sec, nsec);
			std::cout << "reading bag " << bagFilepath << ", time " << sec << ":" << nsec << std::endl;

			std::string depthTopic, imgTopic;
			bool depthCompressed, imgCompressed;
			rgbd::determineRGBDBagSchema(bagFilepath, depthTopic, imgTopic, depthCompressed, imgCompressed);
//			imgTopic = "/rgbd/image";
			rgbd::rgbdBagReader<PointT> frameReader(bagFilepath, depthTopic, imgTopic, 0, 1000000/* max end frame index */, 0/* frameskip */, 0u/* num prev frames to keep */,
				primesensor::getColorCamParams(rgbd::KINECT_640_DEFAULT), primesensor::getColorCamParams(rgbd::KINECT_640_DEFAULT));
			frameReader.readThroughTime(t);
			if(frameReader.getLastFrameTime() != t) throw std::runtime_error("no frame at specified time found in bag");
			cloud = frameReader.getLastCloud();
		}
		else throw std::invalid_argument("invalid framespec");
	}
}
