/*
 * rgbdBagReader: read rgbd frames from a bag using the ros::Bag interface
 *
 * Evan Herbst
 * 7 / 27 / 10
 */

#include <cassert>
#include <iostream>
#include <stdexcept>
#include <boost/make_shared.hpp>
#include <boost/variant/apply_visitor.hpp>
#include <boost/variant/get.hpp>
#include <pcl/ros/conversions.h>
#include "rgbd_util/ros_utility.h"
#include "rgbd_msgs/imageCompression.h"
#include "pcl_rgbd/depth_to_cloud_lib.h"
#include "rgbd_bag_utils/rgbdCompressedImageSubscriber.h"
#include "rgbd_bag_utils/rgbdBagReader.h"

namespace rgbd
{

/*
 * return whether the pointer is non-null
 */
struct someImgDefinedVisitor : public boost::static_visitor<bool>
{
	bool operator () (const sensor_msgs::ImageConstPtr& i) const
	{
		return i;
	}
	bool operator () (const sensor_msgs::CompressedImageConstPtr& i) const
	{
		return i;
	}
};

/*
 * if maxFrameIndex is beyond the end of the bag, things will be ok as long as you use readOne() to tell you how big the bag is
 *
 * frameSkip: # frames to actually skip (eg 4 to use every 5th)
 *
 * minFrame, maxFrame and frameSkip define a "virtual" frame index starting at zero and incrementing with each batch of raw-bag frames skipped
 *
 * numKeptPrevFrames: conserve memory; if empty, no limit; 0 means delete frame i as soon as i + 1 is loaded
 *
 * depth2imgXform, camParams: describe how to sample colors for depth pts
 */
template <typename PointT>
rgbdBagReader<PointT>::rgbdBagReader(const fs::path& bagFilepath, const std::string& depthTopic, const std::string& imgTopic, const uint32_t minFrame, const uint32_t maxFrame, const uint32_t frameSkip,
		const boost::optional<unsigned int> numKeptPrevFrames, const rgbd::CameraParams& camparams, const rgbd::CameraParams& depthCamParams, const rgbd::eigen::Affine3f& depth2img)
: depth2imgXform(depth2img), camParams(camparams), depthCamParamsPtr(new rgbd::CameraParams(depthCamParams)), bag(bagFilepath.string(), rosbag::bagmode::Read), minFrameIndex(minFrame), maxFrameIndex(maxFrame),
  frameskip(frameSkip), lastFrameRead(-1), lastVirtualFrameRead(-1), lastFrameTimeRead(ros::Time(0)), numPrevFramesToKeep(numKeptPrevFrames)
{
	assert(maxFrameIndex >= minFrameIndex);

	std::vector<std::string> topicNames;
	topicNames.push_back(depthTopic);
	topicNames.push_back(imgTopic);
	view.reset(new rosbag::View(bag, rosbag::TopicQuery(topicNames)));
	iter = view->begin();
	endIter = view->end();

	while(lastFrameRead < (int32_t)minFrameIndex - 1)
	{
		const std::pair<rgbd_msgs::DepthMapConstPtr, rgbd::someImageConstPtr> frame = readOneFrame();
		if(!frame.first && !boost::apply_visitor(someImgDefinedVisitor(), frame.second)) throw std::runtime_error("rgbdreader failed before start frame (perhaps bag isn't as long as you thought?)");
		else lastFrameRead++;
	}
}

/*
 * read until we hit end of bag or get a complete frame
 *
 * return <empty, empty> if hit end of bag
 */
template <typename PointT>
std::pair<rgbd_msgs::DepthMapConstPtr, rgbd::someImageConstPtr> rgbdBagReader<PointT>::readOneFrame()
{
	std::pair<rgbd_msgs::DepthMapConstPtr, rgbd::someImageConstPtr> result;

	while(iter != endIter)
	{
		rosbag::MessageInstance& m = *iter;
		const rgbd_msgs::DepthMapConstPtr d = m.instantiate<rgbd_msgs::DepthMap>();
		if(d)
		{
//			cout << m.getDataType() << " @ " << d->header.stamp << endl;
			depthBuffer[d->header.stamp] = d;
		}
		else
		{
			const sensor_msgs::ImageConstPtr i = m.instantiate<sensor_msgs::Image>();
			if(i)
			{
//				cout << m.getDataType() << " @ " << i->header.stamp << endl;
				imgBuffer[i->header.stamp] = i;
			}
			else //the msg isn't a depth map or img
			{
				const sensor_msgs::CompressedImageConstPtr i = m.instantiate<sensor_msgs::CompressedImage>();
				if(i)
				{
//					cout << m.getDataType() << " @ " << i->header.stamp << endl;
					imgBuffer[i->header.stamp] = i;
				}
				else
				{
//					cout << m.getDataType() << endl;
					iter++;
					continue;
				}
			}
		}
		iter++;

		/*
		 * find the first paired times in the buffers (which hold everything since the last frame found)
		 */
		std::map<ros::Time, rgbd::someImageConstPtr, cmpRosTimes>::const_iterator ii = imgBuffer.begin();
		std::map<ros::Time, rgbd_msgs::DepthMapConstPtr, cmpRosTimes>::const_iterator id = depthBuffer.begin();
		ros::Time frameTime(0); //0 means none found
		while(ii != imgBuffer.end() && id != depthBuffer.end() && frameTime == ros::Time(0))
		{
			while(ii != imgBuffer.end() && (*ii).first < (*id).first) ii++;
			if(ii == imgBuffer.end()) break;
			while(id != depthBuffer.end() && (*id).first < (*ii).first) id++;
			if(id == depthBuffer.end()) break;
			/*
			* now *id >= *ii and neither is an end()
			*/
			if((*ii).first == (*id).first) frameTime = (*ii).first;
		}

		if(frameTime != ros::Time(0))
		{
			result.first = depthBuffer[frameTime];
			result.second = imgBuffer[frameTime];

			/*
			 * delete everything through frameTime
			 */
		{
			std::set<ros::Time> toErase;
			for(std::map<ros::Time, rgbd::someImageConstPtr, cmpRosTimes>::const_iterator j = imgBuffer.begin(); j != imgBuffer.end(); j++)
				if((*j).first <= frameTime)
					toErase.insert((*j).first);
			for(std::set<ros::Time>::const_iterator j = toErase.begin(); j != toErase.end(); j++)
				imgBuffer.erase(*j);
		}
		{
			std::set<ros::Time> toErase;
			for(std::map<ros::Time, rgbd_msgs::DepthMapConstPtr, cmpRosTimes>::const_iterator j = depthBuffer.begin(); j != depthBuffer.end(); j++)
				if((*j).first <= frameTime)
					toErase.insert((*j).first);
			for(std::set<ros::Time>::const_iterator j = toErase.begin(); j != toErase.end(); j++)
				depthBuffer.erase(*j);
		}

			lastFrameTimeRead = frameTime;
			return result;
		}
	}

	//at end of bag
	return std::make_pair(rgbd_msgs::DepthMapConstPtr(), sensor_msgs::ImageConstPtr());
}

/*
 * read one frame, then frame-skip as appropriate
 *
 * return false if we were already at the end of the bag or had reached the max requested index
 */
template <typename PointT>
bool rgbdBagReader<PointT>::readOne()
{
	if(lastFrameRead >= (int64_t)maxFrameIndex) return false;

	const std::pair<rgbd_msgs::DepthMapConstPtr, rgbd::someImageConstPtr> frame = readOneFrame();
	if(!frame.first && !boost::apply_visitor(someImgDefinedVisitor(), frame.second)) return false; //if at end of bag

	depthMaps[lastVirtualFrameRead + 1] = frame.first;
	imgs[lastVirtualFrameRead + 1] = frame.second;
	lastFrameRead++;
	lastVirtualFrameRead++;

	if(numPrevFramesToKeep) erase(lastVirtualFrameRead - 1 - *numPrevFramesToKeep);

	for(uint32_t i = 0; i < frameskip && lastFrameRead < (int64_t)maxFrameIndex; i++)
	{
		const std::pair<rgbd_msgs::DepthMapConstPtr, rgbd::someImageConstPtr> frame = readOneFrame();
		if(!frame.first && !boost::apply_visitor(someImgDefinedVisitor(), frame.second)) return true; //if at end of bag (we'll return false next time)
		else lastFrameRead++;
	}
	return true;
}
/*
 * this is the "virtual" frame index: start at zero and add one for each readOne()
 *
 * throw if the bag doesn't contain enough frames
 *
 * post: lastVirtualFrameRead == frameIndex
 */
template <typename PointT>
void rgbdBagReader<PointT>::readThroughFrame(const uint32_t frameIndex)
{
	while(lastVirtualFrameRead < (int64_t)frameIndex)
		if(!readOne())
			throw std::invalid_argument("not enough frames left in bag");
}

/*
 * read frames until the last frame read is at or after t
 *
 * throw if the bag doesn't go that far
 */
template <typename PointT>
void rgbdBagReader<PointT>::readThroughTime(const ros::Time t)
{
	while(lastFrameTimeRead < t)
		if(!readOne())
			throw std::invalid_argument("bag ends before specified timestamp");
}

/*
 * avoid duplicate definitions of templated functions 'get<>' in boost.variant and boost.graph
 */
sensor_msgs::ImageConstPtr getImgAux(const rgbd::someImageConstPtr& p);
sensor_msgs::CompressedImageConstPtr getCompressedImgAux(const rgbd::someImageConstPtr& p);
sensor_msgs::ImageConstPtr* getImgPtrAux(rgbd::someImageConstPtr* const p);
sensor_msgs::CompressedImageConstPtr* getCompressedImgPtrAux(rgbd::someImageConstPtr* const p);

/*
 * read frames as necessary
 *
 * this is the "virtual" frame index: start at zero and add one for each readOne()
 *
 * throw if we don't have the desired frame stored or you ask for the wrong img format
 */
template <typename PointT>
const sensor_msgs::Image& rgbdBagReader<PointT>::getImg(const uint32_t frameIndex)
{
	readThroughFrame(frameIndex);
	if(imgs.find(frameIndex) == imgs.end()) throw std::invalid_argument("bad index");
	return *getImgAux(imgs[frameIndex]);
}
template <typename PointT>
const sensor_msgs::CompressedImage& rgbdBagReader<PointT>::getCompressedImg(const uint32_t frameIndex)
{
	readThroughFrame(frameIndex);
	if(imgs.find(frameIndex) == imgs.end()) throw std::invalid_argument("bad index");
	return *getCompressedImgAux(imgs[frameIndex]);
}
/*
 * return uncompressed regardless of whether the bag contains compressed
 */
template <typename PointT>
const sensor_msgs::ImageConstPtr rgbdBagReader<PointT>::getUncompressedImgPtr(const uint32_t frameIndex)
{
	readThroughFrame(frameIndex);
	if(imgs.find(frameIndex) == imgs.end()) throw std::invalid_argument("bad index");
	try
	{
		return getImgAux(imgs[frameIndex]);
	}
	catch(const boost::bad_get& x)
	{
		rgbd::rgbdCompressedImageSubscriber decompressor;
		return decompressor.decompress(getCompressedImgAux(imgs[frameIndex]));
	}
}
template <typename PointT>
const rgbd_msgs::DepthMap& rgbdBagReader<PointT>::getDepthMap(const uint32_t frameIndex)
{
	readThroughFrame(frameIndex);
	if(depthMaps.find(frameIndex) == depthMaps.end()) throw std::invalid_argument("bad index");
	return *depthMaps[frameIndex];
}
template <typename PointT>
const pcl::PointCloud<PointT>& rgbdBagReader<PointT>::getCloud(const uint32_t frameIndex)
{
	readThroughFrame(frameIndex);
	/*
	 * compute clouds here rather than when reading a frame because (1) we might not want clouds for this bag, (2) we might have different-sized imgs and depth in the bag
	 */
	if(clouds.find(frameIndex) == clouds.end()) //compute the cloud if this is the first time it's been requested
	{
		pcl::PointCloud<PointT>& cloud = clouds[frameIndex];
		bool ok = false;
		// added extra pointer level to get boost::get to work -Peter
		if(sensor_msgs::ImageConstPtr *img = getImgPtrAux(&imgs[frameIndex]))
		{
			ok = image_and_depth_to_cloud(**img, *depthMaps[frameIndex], depth2imgXform, camParams, true, false, cloud, *depthCamParamsPtr);
		}
		else
		{
			sensor_msgs::ImagePtr img(new sensor_msgs::Image);
			rgbd::decompressImg(*getCompressedImgAux(imgs[frameIndex]), *img);
			ok = image_and_depth_to_cloud(*img, *depthMaps[frameIndex], depth2imgXform, camParams, true, false, cloud, *depthCamParamsPtr);
		}
		if(!ok) throw std::runtime_error("image_and_depth_to_cloud failed");
	}
	return clouds[frameIndex];
}

/*
 * return the last one read
 *
 * throw if we haven't read anything
 */
template <typename PointT>
const ros::Time rgbdBagReader<PointT>::getLastFrameTime()
{
	return getLastDepthMap().header.stamp; //20101203: currently all legal frame formats contain depth maps so this is ok
}
template <typename PointT>
const sensor_msgs::Image& rgbdBagReader<PointT>::getLastImg()
{
	return getImg(lastVirtualFrameRead);
}
template <typename PointT>
const sensor_msgs::CompressedImage& rgbdBagReader<PointT>::getLastCompressedImg()
{
	return getCompressedImg(lastVirtualFrameRead);
}
template <typename PointT>
const rgbd_msgs::DepthMap& rgbdBagReader<PointT>::getLastDepthMap()
{
	return getDepthMap(lastVirtualFrameRead);
}
template <typename PointT>
const pcl::PointCloud<PointT>& rgbdBagReader<PointT>::getLastCloud()
{
	return getCloud(lastVirtualFrameRead);
}

template <typename PointT>
const sensor_msgs::ImageConstPtr rgbdBagReader<PointT>::getLastImgPtr()
{
	const uint32_t frameIndex = lastVirtualFrameRead;
	readThroughFrame(frameIndex);
	if(imgs.find(frameIndex) == imgs.end()) throw std::invalid_argument("bad index");
	return getImgAux(imgs[frameIndex]);
}
template <typename PointT>
const sensor_msgs::CompressedImageConstPtr rgbdBagReader<PointT>::getLastCompressedImgPtr()
{
	const uint32_t frameIndex = lastVirtualFrameRead;
	readThroughFrame(frameIndex);
	if(imgs.find(frameIndex) == imgs.end()) throw std::invalid_argument("bad index");
	return getCompressedImgAux(imgs[frameIndex]);
}
/*
 * return uncompressed regardless of whether the bag contains compressed
 */
template <typename PointT>
const sensor_msgs::ImageConstPtr rgbdBagReader<PointT>::getLastUncompressedImgPtr()
{
	return getUncompressedImgPtr(lastVirtualFrameRead);
}
template <typename PointT>
const rgbd_msgs::DepthMapConstPtr rgbdBagReader<PointT>::getLastDepthMapPtr()
{
	const uint32_t frameIndex = lastVirtualFrameRead;
	readThroughFrame(frameIndex);
	if(depthMaps.find(frameIndex) == depthMaps.end()) throw std::invalid_argument("bad index");
	return depthMaps[frameIndex];
}

template <typename PointT>
void rgbdBagReader<PointT>::erase(const uint32_t frameIndex)
{
	imgs.erase(frameIndex);
	depthMaps.erase(frameIndex);
	clouds.erase(frameIndex);
}

} //namespace
