/*
 * mathUtils
 *
 * Evan Herbst
 * 3 / 3 / 10
 */

#include <cassert>
#include <cstdlib> //rand()
#include <vector>
#include <algorithm> //copy(), random_shuffle(), sort()
#include <iterator> //iterator_traits, distance()
#include <boost/range/concepts.hpp>
#include <boost/range.hpp>
#include <boost/concept/assert.hpp>
#include <boost/concept_check.hpp> //OutputIterator

/*
 * integer power of a float
 */
template <typename T>
T powi(T x, unsigned int n)
{
	T r = 1;
	while(n)
	{
		if(n & 1)
		{
			r *= x;
			--n;
		}
		x *= x;
		n /= 2;
	}
	return r;
}

template <typename T>
T clamp(const T x, const T min, const T max)
{
	return std::min(max, std::max(min, x));
}

/*
 * write sampleSize uniformly selected elements of the input range to the output range (without replacement), using rand()
 *
 * pre: the input range has size at least sampleSize; the output range has size at least sampleSize
 */
template <typename RandomAccessIteratorT, typename OutputIteratorT>
void randomSample(const RandomAccessIteratorT begin, const RandomAccessIteratorT end, const size_t sampleSize, OutputIteratorT obegin)
{
	const size_t inputSize = std::distance(begin, end);
	assert(inputSize >= sampleSize);
	/*
	 * select a sampling algorithm
	 *
	 * if you aren't choosing very many points, the naive algorithm beats the pants off shuffling the whole list (like .05s vs 1.7s -- EVH 20100308)
	 */
	if((double)sampleSize / inputSize < .5) //this threshold is an uneducated guess -- EVH 20100308
	{
		std::vector<bool> used(inputSize, false); //just don't try to iterate
		for(unsigned int i = 0; i < sampleSize; i++)
		{
			unsigned int j = rand() % inputSize;
			while(used[j]) j = rand() % inputSize;
			*obegin = *(begin + j);
			obegin++;
			used[j] = true;
		}
	}
	else
	{
		std::vector<typename std::iterator_traits<RandomAccessIteratorT>::value_type> shuffled(begin, end);
		std::random_shuffle(shuffled.begin(), shuffled.end());
		std::copy(shuffled.begin(), shuffled.begin() + sampleSize, obegin);
	}
}

/*
 * copy the elements of values specified by indices to the output range
 */
template <typename RandomAccessRangeT, typename SinglePassRangeT, typename OutputIteratorT>
void copySelected(const RandomAccessRangeT& values, const SinglePassRangeT& indices, OutputIteratorT obegin)
{
	BOOST_CONCEPT_ASSERT((boost::RandomAccessRangeConcept<RandomAccessRangeT>));
	BOOST_CONCEPT_ASSERT((boost::SinglePassRangeConcept<SinglePassRangeT>));
	BOOST_CONCEPT_ASSERT((boost::OutputIterator<OutputIteratorT, typename boost::range_value<RandomAccessRangeT>::type>));

	typename boost::range_iterator<const RandomAccessRangeT>::type j = boost::begin(values);
	for(typename boost::range_iterator<const SinglePassRangeT>::type i = boost::begin(indices); i != boost::end(indices); i++)
	{
		assert((int)*i < (int)boost::size(values));
		*obegin = *(j + *i);
		obegin++;
	}
}
/*
 * by-value version of copySelected()
 */
template <typename RandomAccessRangeT, typename ForwardRangeT>
std::vector<typename boost::range_value<RandomAccessRangeT>::type> selectByIndices(const RandomAccessRangeT& values, const ForwardRangeT& indices)
{
	std::vector<typename boost::range_value<RandomAccessRangeT>::type> result(indices.size());
	unsigned int j = 0;
	for(typename boost::range_iterator<const ForwardRangeT>::type i = boost::begin(indices); i != boost::end(indices); i++, j++)
	{
		assert((int)*i < (int)boost::size(values));
		result[j] = values[*i];
	}
	return result;
}

template <typename T>
struct itemWithIndex
{
	bool operator < (const itemWithIndex<T>& i)
	{
		return data < i.data;
	}

	unsigned int index;
	T data;
};

/*
 * like operator <
 *
 * CmpT compares Ts; we'll compare itemWithIndex<T>s by it
 */
template <typename T, typename CmpT>
class cmpItemsWithIndex
{
	public:

		cmpItemsWithIndex(const CmpT& c) : cmp(c) {}

		bool operator () (const itemWithIndex<T>& t1, const itemWithIndex<T>& t2)
		{
			return cmp(t1.data, t2.data);
		}

	private:

		CmpT cmp;
};

/*
 * return the sorted list of indices (in [0 .. n-1]) and edit the given range
 *
 * the given data are copied twice
 *
 * pre: the given range is writable
 */
template <typename RandomAccessIteratorT>
std::vector<unsigned int> sortWithIndices(const RandomAccessIteratorT begin, const RandomAccessIteratorT end)
{
	const unsigned int numItems = std::distance(begin, end);
	std::vector<itemWithIndex<typename std::iterator_traits<RandomAccessIteratorT>::value_type> > items(numItems);
	for(unsigned int i = 0; i < numItems; i++)
	{
		items[i].index = i;
		items[i].data = *(begin + i);
	}
	std::sort(items.begin(), items.end());
	std::vector<unsigned int> indices(numItems);
	for(unsigned int i = 0; i < numItems; i++)
	{
		indices[i] = items[i].index;
		*(begin + i) = items[i].data;
	}
	return indices;
}
template <typename RandomAccessIteratorT, typename CmpT>
std::vector<unsigned int> sortWithIndices(const RandomAccessIteratorT begin, const RandomAccessIteratorT end, CmpT cmp)
{
	const unsigned int numItems = std::distance(begin, end);
	std::vector<itemWithIndex<typename std::iterator_traits<RandomAccessIteratorT>::value_type> > items(numItems);
	for(unsigned int i = 0; i < numItems; i++)
	{
		items[i].index = i;
		items[i].data = *(begin + i);
	}
	std::sort(items.begin(), items.end(), cmpItemsWithIndex<typename std::iterator_traits<RandomAccessIteratorT>::value_type, CmpT>(cmp));
	std::vector<unsigned int> indices(numItems);
	for(unsigned int i = 0; i < numItems; i++)
	{
		indices[i] = items[i].index;
		*(begin + i) = items[i].data;
	}
	return indices;
}

/*
 * remove the specified elements from r: return an iterator to the end of the shortened part (but don't resize r, since we don't know how)
 *
 * post: relative order of elements may have been changed (deterministically, in the same way as by removeFromVector())
 */
template <typename RandomAccessRangeT>
typename boost::range_iterator<RandomAccessRangeT>::type removeFromRange(RandomAccessRangeT& r, const std::vector<unsigned int>& indices)
{
	BOOST_CONCEPT_ASSERT((boost::RandomAccessRangeConcept<RandomAccessRangeT>));
	typedef typename boost::range_value<RandomAccessRangeT>::type T;
	typedef typename boost::range_reference<RandomAccessRangeT>::type Tref;

	if(indices.empty()) return boost::end(r);

	for(unsigned int i = 0; i < indices.size() - 1; i++) assert(indices[i] < indices[i + 1]);
	assert(indices.back() < boost::size(r));

	unsigned int nextToReorder = boost::size(r) - 1, //index in v of next element to be copied backward
		lastOKComp = indices.size() - 1; //index in indices
	while(nextToReorder == indices[lastOKComp] && lastOKComp > 0)
	{
		nextToReorder--;
		lastOKComp--;
	}
	unsigned int nextToRemove = 0; //index into indices
	while(nextToRemove < indices.size() && indices[nextToRemove] < nextToReorder)
	{
		//std::swap(v[indices[nextToRemove]], v[nextToReorder]); breaks when T = bool
		Tref t1 = *(boost::begin(r) + nextToReorder), t2 = *(boost::begin(r) + indices[nextToRemove]);
		const T tmp = t1;
		t1 = t2;
		t2 = tmp;
		nextToRemove++;
		nextToReorder--;
		while(nextToReorder == indices[lastOKComp] && lastOKComp > 0)
		{
			nextToReorder--;
			lastOKComp--;
		}
	}
	return boost::end(r) - indices.size();
}
/*
 * remove the specified elements from r: return an iterator to the end of the shortened part (but don't resize r, since we don't know how)
 *
 * post: relative order of elements is unchanged
 */
template <typename RandomAccessRangeT>
typename boost::range_iterator<RandomAccessRangeT>::type removeFromRangeNoReorder(RandomAccessRangeT& r, const std::vector<size_t>& indices)
{
	BOOST_CONCEPT_ASSERT((boost::RandomAccessRangeConcept<RandomAccessRangeT>));

	if(indices.empty()) return boost::end(r);

	for(size_t i = 0; i < indices.size() - 1; i++)
		std::copy(boost::begin(r) + indices[i] + 1, boost::begin(r) + indices[i + 1], boost::begin(r) + indices[i] - i);
	std::copy(boost::begin(r) + indices.back() + 1, boost::end(r), boost::begin(r) + indices.back() + 1 - indices.size());

	return boost::end(r) - indices.size();
}
/*
 * remove the specified elements from v
 *
 * pre: indices is sorted ascending
 *
 * post: relative order of elements may have been changed (deterministically)
 */
template <typename T, typename Alloc>
void removeFromVector(std::vector<T, Alloc>& v, const std::vector<unsigned int>& indices)
{
	removeFromRange(v, indices);
	v.erase(v.end() - indices.size(), v.end());
}
/*
 * remove the specified elements from v
 *
 * post: relative order of elements is unchanged
 */
template<typename T>
void removeFromVector(std::vector<T> &vec, std::vector<bool> const& shouldRemove)
{
	std::vector<T> vecCopy = vec;
	vec.clear();
	for(unsigned int i=0; i<vecCopy.size(); i++){
		if(!shouldRemove[i]){
			vec.push_back(vecCopy[i]);
		}
	}
}

/*
 * remove the specified values from v
 *
 * pre: the order of values in valsToRemove is the same as their order in v; each value in valsToRemove occurs exactly once in v
 *
 * post: relative order of elements is unchanged
 */
template<typename T, typename ForwardRangeT>
void removeUniqueValuesFromOrderedVector(std::vector<T>& v, const ForwardRangeT& valsToRemove)
{
	BOOST_CONCEPT_ASSERT((boost::RandomAccessRangeConcept<ForwardRangeT>));
	BOOST_CONCEPT_ASSERT((boost::is_same<T, typename boost::range_value<ForwardRangeT>::type>));

	std::vector<size_t> indices(boost::size(valsToRemove)); //indices of vals from valsToRemove in v
	size_t i0 = 0, j0 = 0;
	for(typename boost::range_iterator<const ForwardRangeT>::type i = boost::begin(valsToRemove); i != boost::end(valsToRemove); i++, i0++)
	{
		while(v[j0] != *i) j0++; //don't have to check j0 against v.size() because we know *i is in v
		indices[i0] = j0;
	}

	removeFromRangeNoReorder(v, indices);
	v.resize(v.size() - indices.size());
}

/*
 * toRemove: indices into vec
 *
 * optimized for a relatively small number of removals
 *
 * pre: toRemove is sorted ascending
 */
template<typename RandomAccessRangeT>
std::vector<typename boost::range_value<RandomAccessRangeT>::type> selectAllBut(const RandomAccessRangeT& vec, std::vector<unsigned int> const& toRemove)
{
	BOOST_CONCEPT_ASSERT((boost::RandomAccessRangeConcept<RandomAccessRangeT>));

	std::vector<typename boost::range_value<RandomAccessRangeT>::type> result(boost::size(vec) - toRemove.size());
	if(toRemove.empty())
	{
		std::copy(boost::begin(vec), boost::end(vec), result.begin());
	}
	else
	{
		for(unsigned int i = 0; i < toRemove.size() - 1; i++) assert(toRemove[i] < toRemove[i + 1]);
		assert(toRemove.back() <= (unsigned int)boost::size(vec));

		typename std::vector<typename boost::range_value<RandomAccessRangeT>::type>::iterator j = result.begin();
		std::copy(boost::begin(vec), boost::begin(vec) + toRemove[0], j);
		j += toRemove[0];
		for(unsigned int i = 0; i < toRemove.size() - 1; i++)
		{
			std::copy(boost::begin(vec) + toRemove[i] + 1, boost::begin(vec) + toRemove[i + 1], j);
			j += toRemove[i + 1] - toRemove[i] - 1;
		}
		std::copy(boost::begin(vec) + toRemove.back() + 1, boost::end(vec), j);
	}
	return result;
}
